#!/usr/bin/env python

# Ubuntu Testing Automation Harness
# Copyright 2012 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"UTAH client used to run tests."""


import logging
import platform
import sys
import os

from utah import logger
from utah.url import url_argument
from utah.client.probe import probes
from utah.client.probe.battery import battery
from utah.client.runner import Runner
from utah.client.state_agent import StateAgentYAML
from utah.client.result import classes as result_classes

from utah.client.common import (
    DEFAULT_STATE_FILE,
    get_install_type,
    ReturnCodes,
)
from utah.client.exceptions import UTAHClientError


def get_parser():
    """Process the command line arguments."""
    # requires Python2.7+
    import argparse

    parser = argparse.ArgumentParser(
        description='''Run test cases locally and write results to a file.

Note: This is expected to be used only for development purposes on a machine
that has been provisioned using run_utah_tests.py''',
        formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('--resume', action='store_true',
                        help='Continue a previous run.  Used after a reboot')
    parser.add_argument('-s', '--state-file',
                        help=('File to use for storing state (default "{}")'
                              .format(DEFAULT_STATE_FILE)))
    parser.add_argument('-f', '--format',
                        choices=['text', 'yaml', 'json'],
                        default='yaml',
                        help='Output format (default "%(default)s")')
    parser.add_argument('-t', '--testdir',
                        default='/var/lib/utah', help='Main test directory')
    parser.add_argument('-i', '--install-type',
                        choices=['desktop', 'server', 'mini', 'alternate'],
                        default=get_install_type(),
                        help=('Installation Variant '
                              '(i.e. server, desktop, etc.)'))
    parser.add_argument('-r', '--runlist', type=url_argument,
                        # Having a default here means that if a resume happens
                        # post-reboot without a runlist specified, it will
                        # run the default runlist instead of erroring
                        #default='/usr/share/utah/client/examples/master.run',
                        help='runlist file name')
    parser.add_argument('-o', '--output', help='write output to this file')
    parser.add_argument('-a', '--append', action='store_true',
                        help='append to output')
    parser.add_argument('-d', '--debug', action='store_true',
                        help='Print debugging output')
    parser.add_argument('-b', '--battery',
                        default=battery.implementation,
                        choices=battery.implementations(),
                        help='Choose which battery implementation to use.')
    parser.add_argument('-p', '--probe-dir', default='/tmp',
                        help='Directory to store probe files.')
    parser.add_argument('--no-root', action='store_true',
                        help='Do not require root permissions')

    return parser


def setup_logging(debug=False):
    if debug:
        level = logging.DEBUG
    else:
        level = logging.WARNING

    logging.basicConfig(level=level,
                        format='%(asctime)s - %(levelname)s - %(message)s')


def setup_probes(args):
    os.environ['UTAH_PROBE_DIR'] = args.probe_dir
    battery.implementation = args.battery
    probes.output_dir = args.probe_dir


def main(argv=None):
    if argv is None:
        argv = sys.argv[1:]

    """Interpret command line arguments and run tests."""
    # TODO: write <2.7 optparse version and set based on version of python
    #       being used.
    parser = get_parser()

    # Capture parsing error and use the returncode defined
    # in utah.client.common, not the one that argparse uses by default
    try:
        args = parser.parse_args(argv)
    except SystemExit:
        sys.exit(ReturnCodes.CMD_PARSING_ERROR)

    if not args.no_root and not os.getuid() == 0:
        parser.print_usage()
        msg = ('{}: UTAH client is expected to be executed as root user\n'
               .format(parser.prog))
        sys.stderr.write(msg)
        logger.log_error(msg)
        sys.exit(ReturnCodes.INVALID_USER)

    setup_logging(args.debug)
    setup_probes(args)

    logging.debug(args)
    logging.info(platform.uname())

    resume = args.resume
    state_file = args.state_file
    testdir = args.testdir
    runlist = args.runlist
    output = args.output
    install_type = args.install_type

    # Read results before last reboot so that they can be properly merged
    if resume and output:
        with open(output) as f:
            old_results = f.read()
    else:
        old_results = None

    result = result_classes[args.format](output, args.append)
    try:
        state_agent = StateAgentYAML(state_file=state_file)
        runner = Runner(install_type=install_type, state_agent=state_agent,
                        result=result, runlist=runlist,
                        resume=resume, old_results=old_results,
                        testdir=testdir, output=args.output)
        if resume:
            runner.load_state()

        for x in xrange(runner.repeat_count + 1):
            returncode = runner.run(x)
            if returncode:
                break

        if returncode != ReturnCodes.REBOOT:
            returncode = runner.process_results()
    except UTAHClientError as e:
        sys.stderr.write(str(e))
        sys.stderr.write('\n')
        logger.log_error(str(e))
        sys.exit(ReturnCodes.EXCEPTION_ERROR)

    sys.exit(returncode)

if __name__ == "__main__":
    main()
