#!/usr/bin/env python
# Ubuntu Testing Automation Harness
# Copyright 2013 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Provisions and configures a phablet device for testing under UTAH."""

from __future__ import print_function

import argparse
import os
import shutil
import socket
import subprocess
import sys
import tempfile
import threading
import time

from SimpleXMLRPCServer import SimpleXMLRPCServer

import yaml

# ensure we use the proper logger for phablet *before* importing any UTAH libs
os.environ['UTAH_LOGGER'] = 'STDOUT'

from utah import template
from utah.provisioning import touch
from utah.retry import retry

CLIENT_PROBES_DIR = '/var/cache/utah-probes'


def phablet_tools_errors(f):
    """Decorating phablet-tools error management."""
    def _errors(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except subprocess.CalledProcessError:
            # phablet-flash prints a pretty good error, so we should just exit
            sys.exit(1)
    return _errors


@phablet_tools_errors
def _flash(args, device):
    cmd = ['phablet-flash']
    if args.ubuntu_bootstrap:
        cmd.append('ubuntu-system')
        cmd.extend(['--channel', args.channel])
    else:
        cmd.append('cdimage-touch')
        if not args.current:
            cmd.append('--pending')
    if args.serial:
        cmd.extend(['-s', args.serial])
    if args.revision:
        cmd.extend(['--revision', args.revision])
    if args.device:
        cmd.extend(['-d', args.device])
    if not args.preserve:
        cmd.append('--bootstrap')

    print('= flashing device...')
    subprocess.check_call(cmd)
    print()
    print('= waiting for install/boot...')
    device.wait_for_device()
    device.adb_root()
    device.wait_for_device()
    if args.ubuntu_bootstrap:
        # give the system some time. otherwise it seems like we start setting
        # up utah and network stuff too soon and the device hangs
        print(' install completing waiting 120s')
        time.sleep(120)
    print()


@phablet_tools_errors
def _network(args):
    cmd = ['phablet-network']
    if args.serial:
        cmd.extend(['-s', args.serial])
    if args.network_file:
        cmd.extend(['-n', args.network_file])

    print('= setting up network...')
    subprocess.check_call(cmd)
    print()


def _install_branch(device, branch):
    device.run_ubuntu('rm -rf /home/phablet/utah')
    device.run_ubuntu('bzr branch {} /home/phablet/utah'.format(branch))


def _assert_developer_mode(device):
    """Required for system-images until click-packages are complete."""
    try:
        device.run_ubuntu('touch /.test_dmode\\; rm /.test_dmode', False)
    except touch.ADBShellError:
        # the device is mounted r/o, put into developer mode
        print("= moving device into developer mode...")
        device.run_ubuntu('touch /userdata/.writable_image')
        device.reboot_and_wait()


def _install_utah(device, ppa, branch):
    print("= installing utah...")
    device.wait_for_device()
    device.wait_for_network()
    device.run_ubuntu('add-apt-repository -y {}'.format(ppa))
    device.run_ubuntu('apt-get update -qq')
    device.run_ubuntu('apt-get install -yq --force-yes utah-client')

    if branch:
        _install_branch(device, branch)


def _save_props(device, results_dir):
    def run():
        try:
            buf = device.run_ubuntu('getprop', False)
            ofile = os.path.join(results_dir, 'adb.props')
            with open(ofile, 'w') as f:
                f.write(buf)
        except touch.ADBShellError:
            print("Unable to get adb.props")

        try:
            buf = device.run_ubuntu('sudo dmesg', False)
            ofile = os.path.join(results_dir, 'dmesg.log')
            with open(ofile, 'w') as f:
                f.write(buf)
        except touch.ADBShellError:
            print("Unable to get dmesg.log")

        try:
            buf = device.run_ubuntu(
                'sudo /var/lib/lxc/android/rootfs/system/bin/logcat -d', False)
            ofile = os.path.join(results_dir, 'logcat.log')
            with open(ofile, 'w') as f:
                f.write(buf)
        except touch.ADBShellError:
            print("Unable to get logcat.log")

        try:
            buf = device.run_ubuntu(
                'sudo cat /var/log/upstart/whoopsie.log', False)
            ofile = os.path.join(results_dir, 'whoopsie.log')
            with open(ofile, 'w') as f:
                f.write(buf)
        except touch.ADBShellError:
            print("Unable to get whoopsie.log")

    retry(run)


def _configure_autorun(device, event_name, env):
    autorun = '/usr/local/bin/utah-autorun.sh'
    buf = template.as_buff('utah-autorun.sh.jinja2', log_file='/tmp/utah.log')
    device.create_file_ubuntu(autorun, buf, '755')

    buf = 'phablet    ALL = NOPASSWD: ALL'
    device.create_file_ubuntu('/etc/sudoers.d/phablet', buf, '440')

    buf = [
        'start on {} or started ubuntu-touch-session'.format(event_name),
        'task',
        'exec {} >/tmp/utah.log 2>&1'.format(autorun),
    ]
    device.create_file_ubuntu('/etc/init/run_utah.conf', '\n'.join(buf))

    p = '/etc/utah/autorun'
    device.run_ubuntu('sudo sh -c "[ -d {} ] || mkdir -p {}"'.format(p, p))

    if not env:
        env = ''
    buf = template.as_buff('utah-autopilot.jinja2', env=env)
    device.create_file_ubuntu('/usr/local/bin/utah-autopilot', buf, '755')


def _push_run_file(args, device, response_server=None):
    buf = template.as_buff('utah-autorun-phablet.sh.jinja2',
                           probes_dir=CLIENT_PROBES_DIR,
                           utah_cmd=args.utah_cmd,
                           result='/var/lib/utah/utah.yaml',
                           runlist=args.runlist)
    device.create_file_ubuntu('/etc/utah/autorun/01_run-utah', buf, '755')

    if response_server:
        buf = template.as_buff('utah-phablet-disconnected-resp.py.jinja2',
                               server=response_server)
        device.create_file_ubuntu(
            '/etc/utah/autorun/02_disconnected_response', buf, '755')


def _run_utah_connected(args, device, event_name):
    """run utah and assumes we can safely use ADB throughout the test."""
    _push_run_file(args, device)

    try:
        with device.tail_syslog():
            device.run_ubuntu('initctl emit {}'.format(event_name))
    except touch.ADBShellError as e:
        # print the error but keep going so we can try and collect the logs
        # which may include a utah.yaml
        print(e.message)
    except KeyboardInterrupt as e:
        print('command interrupted, shutting down utah on target...')
        device.run_ubuntu('initctl stop {}'.format(event_name))
        raise e


def _public_ip():
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.connect(('8.8.8.8', 80))
    ip = s.getsockname()[0]
    s.close()
    return ip


def _run_utah_disconnected(args, device, event_name):
    """run utah with the USB connection to the device disabled."""

    def _utah_cb():
        print('client response received')

    # determine our IP and ephermal port for the client
    server = SimpleXMLRPCServer(('0.0.0.0', 0), allow_none=True)
    port = server.socket.getsockname()[1]

    _push_run_file(args, device, 'http://{}:{}'.format(_public_ip(), port))

    server.register_function(_utah_cb, 'utah_done')

    t = threading.Thread(target=server.handle_request)
    t.daemon = True
    t.start()
    print('= waiting for response on port {}...'.format(port))
    device.run_ubuntu('initctl emit --no-wait {}'.format(event_name))
    subprocess.check_call([args.disconnected, 'off'])
    t.join()
    subprocess.check_call([args.disconnected, 'on'])
    device.wait_for_device()


def _combine_yamls(host, target):
    """Replaces host props like "x86" with the real ones from the target."""
    replace_props = [
        'arch',
        'build_number',
        'install_type',
        'media-info',
        'packages',
        'release',
        'uname',
    ]

    with open(host, 'r') as f:
        host = yaml.load(f.read())
    if target:
        with open(target, 'r') as f:
            target = yaml.load(f.read())
    else:
        target = {}
    for p in replace_props:
        host[p] = target.get(p, 'n/a')
    return host


def _run_from_host(args, device):
    """Executes a runlist from the host that will drive the target."""
    env = os.environ.copy()
    if args.serial:
        env['ANDROID_SERIAL'] = args.serial

    device.run_ubuntu('sudo mkdir {}'.format(CLIENT_PROBES_DIR))

    host = os.path.join(args.results_dir, 'utah-host.yml')
    tmpdir = tempfile.mkdtemp()
    cmd = '{} -d -o {} -r {} -p {} -t {} --no-root'.format(
        args.utah_cmd, host, args.runlist, args.results_dir, tmpdir)
    rc = subprocess.call(cmd, env=env, shell=True)
    print('utah runlist exited with rc={}'.format(rc))
    shutil.rmtree(tmpdir)

    # We've run the test, now do a run on the target so we can grab its YAML
    # and combine the proper parts of it with our results
    try:
        device.run_ubuntu('which utah', stream=False)
        target = os.path.join(args.results_dir, 'utah-tgt.yml')
        rl = '/usr/share/utah/client/examples/pass.run'
        device.run_ubuntu('utah -r {} -o /tmp/utah.yaml'.format(rl))
        device.get_file_ubuntu('/tmp/utah.yaml', target)
    except touch.ADBShellError:
        # 1 of 2 things happened
        # 1) utah wasn't installed so don't combine stuff. It will produce
        #    an odd .yaml file, but that's probably okay to the user.
        # 2) utah failed doing pass.run. This is unlikely, but still
        #    not fatal and we should give the user some results back
        target = None

    combined = _combine_yamls(host, target)
    dst = os.path.join(args.results_dir, 'utah.yaml')
    with open(dst, 'w') as f:
        yaml.dump(combined, f, default_flow_style=False, allow_unicode=True)


def _gather_results(args, device):
    if not args.from_host:
        dst = os.path.join(args.results_dir, 'utah.yaml')
        device.get_file_ubuntu('/var/lib/utah/utah.yaml', dst, False)

        dst = os.path.join(args.results_dir, 'utah.log')
        device.get_file_ubuntu('/tmp/utah.log', dst, False)

    if args.whoopsie:
        try:
            device.run_ubuntu('/usr/share/apport/whoopsie-upload-all -t 300')
        except touch.ADBShellError:
            print('WARNING: Could not confirm whoopsie-upload-all after '
                  '5 minutes')

    device.get_file_ubuntu(CLIENT_PROBES_DIR, args.results_dir, False)

    for f in args.pull:
        device.get_file_ubuntu(f, args.results_dir, False)


def _clean_up(device):
    """Ensure there aren't utah.log/yaml etc files laying around.

    Sometimes utah-client could fail, but we could gather old logs from a prior
    run. This helps prevent that.

    """
    device.run_ubuntu(
        'sudo rm -rf /var/lib/utah/utah.yaml /tmp/utah.log {}'.format(
            CLIENT_PROBES_DIR))


def _run_utah(args, device):
    _clean_up(device)
    if args.from_host:
        _run_from_host(args, device)
    else:
        event_name = 'run_utah'
        _configure_autorun(device, event_name, args.env)

        if args.disconnected:
            _run_utah_disconnected(args, device, event_name)
        else:
            _run_utah_connected(args, device, event_name)
    _gather_results(args, device)


def _get_parser():
    parser = argparse.ArgumentParser(
        description='Provisions a device for UTAH testing.')
    parser.add_argument('-s', '--serial',
                        help='Android serial ID of device')
    parser.add_argument('--skip-install', action='store_true',
                        help='Skip running the phablet-flash install.')
    parser.add_argument('-r', '--revision',
                        help='series/revision to install')
    parser.add_argument('--current', action='store_true',
                        help='Use the "current" image rather than "pending".')
    parser.add_argument('--ubuntu-bootstrap', action='store_true',
                        help='Use the system-images.ubuntu.com images.')
    parser.add_argument('--developer-mode', action='store_true',
                        help='Allow write access system-image partitions.')
    parser.add_argument('--channel',
                        help='Use an alternative channel for system-image.')
    parser.add_argument('--preserve', action='store_true',
                        help='Preserve the systems user data when flashing.')
    parser.add_argument('--skip-network', action='store_true',
                        help='Skip setting up wifi settings on device.')
    parser.add_argument('-n', '--network-file',
                        help='''Specify network manager config file for wifi.
                        If none is specified we'll try and use the default wifi
                        configuration found on host system.''')
    parser.add_argument('--skip-utah', action='store_true',
                        help='Skip install the utah-client on device.')
    parser.add_argument('-b', '--branch',
                        help='Install UTAH client on device from branch.')
    parser.add_argument('-e', '--env',
                        help='Environment variables to set for utah-client.')
    parser.add_argument('--ppa', default='ppa:utah/stable',
                        help='''Specify an alternative PPA for utah.
                        default=%(default)s''')
    parser.add_argument('--results-dir', default='/tmp',
                        help='''Directory to store results in on the host.
                        default=%(default)s''')
    parser.add_argument('--pull', default=[], action='append',
                        help='''A file or directory to copy into the hosts's
                        results-dir after UTAH has been run. This option can
                        be specified multiple times.''')
    parser.add_argument('-l', '--runlist',
                        help='The utah runlist to execute')
    parser.add_argument('-d', '--disconnected',
                        help='''Uses the supplied script to disconnect USB
                        while UTAH is executed.''')
    parser.add_argument('--device',
                        help='The Android device type.')
    parser.add_argument('--from-host', action='store_true',
                        help='''Executes from the host (the runlist must talk
                        to the target using ADB.''')
    parser.add_argument('--whoopsie', action='store_true',
                        help='''Run whoopsie-upload-all after the test to
                             to ensure complete .crash files are uploaded.''')
    return parser


def _check_args(args, device):
    args.utah_cmd = 'utah'
    if args.branch:
        path = '/home/phablet/utah'
        args.utah_cmd = 'PYTHONPATH={} {}/client.py'.format(path, path)

    if args.env:
        args.utah_cmd = '{} {}'.format(args.env, args.utah_cmd)

    if not args.skip_install and not args.device and not device.detect_type():
        print('ERROR: --device option required for provisioning')
        sys.exit(1)

    if not args.skip_utah:
        args.developer_mode = True  # installing utah will require write access

    if not os.path.exists(args.results_dir):
        print('ERROR: results-dir does not exist: {}'.format(args.results_dir))
        sys.exit(1)

    if not os.path.isdir(args.results_dir):
        print(
            'ERROR: results-dir not a directory: {}'.format(args.results_dir))
        sys.exit(1)

    if args.revision and not args.ubuntu_bootstrap:
        print('ERROR: --revision cannot be used with --ubuntu-bootstrap')
        sys.exit(1)

    if args.channel and not args.ubuntu_bootstrap:
        print('ERROR: --channel cannot be used with --ubuntu-bootstrap')
        sys.exit(1)

    if not args.channel:
        args.channel = 'devel-proposed'
        if args.current:
            args.channel = 'stable'


def main(argv=None):
    if argv:
        args = _get_parser().parse_args(argv)
    else:
        args = _get_parser().parse_args()

    device = touch.Device(print, args.serial)
    _check_args(args, device)

    if not args.skip_install:
        _flash(args, device)

    if not args.skip_network:
        _network(args)

    if args.developer_mode:
        _assert_developer_mode(device)

    if not args.skip_utah:
        _install_utah(device, args.ppa, args.branch)

    _save_props(device, args.results_dir)

    if args.runlist:
        _run_utah(args, device)

if __name__ == "__main__":
    try:
        main()
    except touch.ADBShellError as e:
        print(e.long_message())
        sys.exit(1)
