#!/usr/bin/env python

# Ubuntu Testing Automation Harness
# Copyright 2012 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Provision a machine a run a test."""

import logging
import sys

from traceback import format_exception

import utah

from utah.cleanup import cleanup
from utah.config import config
from utah.exceptions import UTAHException
from utah.group import check_user_group, print_group_error_message
from utah.iso import ISO
from utah.parser import get_parser, parse_args  # NOQA
from utah.provisioning.ssh import ProvisionedMachine
from utah.run import (
    configure_logging,
    run_tests,
    ReturnCodes,
)
from utah.timeout import timeout, UTAHTimeout


MISSING = []
try:
    from utah.provisioning.baremetal import get_baremetal
except ImportError:
    MISSING.append('baremetal')
try:
    from utah.provisioning.vm import get_vm
except ImportError:
    MISSING.append('vm')


def _get_unprovisioned_machine(args, image=None):
    if image is None:
        image = ISO(arch=args.arch, installtype=args.type, series=args.series)
    kw = {'clean': (not args.no_destroy),
          'poweroff': args.poweroff,
          'boot': args.boot,
          'debug': args.debug,
          'image': image,
          'initrd': args.initrd,
          'kernel': args.kernel,
          'name': args.name,
          'new': True,
          'preseed': args.preseed,
          'rewrite': args.rewrite,
          'xml': args.xml
          }
    if args.machinetype == 'physical':
        if 'baremetal' in MISSING:
            raise UTAHException(
                'utah-baremetal package needed for physical machines')
        return get_baremetal(args.arch, **kw)
    else:
        if args.machinetype == 'virtual-live-server':
            from utah.provisioning.live_server import LiveServerVM
            kw['machinetype'] = LiveServerVM
        if 'vm' in MISSING:
            raise UTAHException('Virtual Machine support is not available')
        # TODO: consider removing these as explicit command line arguments
        # and using config instead
        kw['diskbus'] = args.diskbus
        kw['emulator'] = args.emulator
        kw['disksizes'] = args.gigabytes
        return get_vm(**kw)


def _get_machine(args):
    image = args.image
    if image and image.endswith('.iso'):
        image = ISO(image=args.image)
    if args.skip_provisioning:
        # TBD: Inventory should be used to verify machine
        # is not running other tests
        return ProvisionedMachine(name=args.name)
    return _get_unprovisioned_machine(args, image)


def run_utah_tests(argv=None):
    if argv is None:
        argv = sys.argv[1:]

    # Capture parsing error and use the returncode defined in utah.run,
    # not the one that argparse uses by default
    try:
        args = parse_args(argv)
    except SystemExit:
        sys.exit(ReturnCodes.CMD_PARSING_ERROR)

    if not check_user_group():
        print_group_error_message(__file__)
        return ReturnCodes.GROUP_ERROR

    configure_logging(args.debug)
    logging.info('UTAH version: %s', utah.__version__)
    logging.debug('Configuration variables:\n%s', config.dump())

    exitstatus = ReturnCodes.SUCCESS
    locallogs = []

    exitstatus, locallogs = run_tests(args, _get_machine(args))
    if len(locallogs) > 0:
        print('Test logs copied to the following files:')
        print("\t" + "\n\t".join(locallogs))

    return(exitstatus)


if __name__ == '__main__':
    try:
        try:
            exitstatus = timeout(config.jobtimeout, run_utah_tests)
        finally:
            cleanup.run()
    except UTAHTimeout as exception:
        sys.stderr.write('UTAH timeout: {}\n'.format(exception.message))
        sys.exit(ReturnCodes.TIMEOUT_ERROR)
    except UTAHException as error:
        sys.stderr.write('UTAH exception: {}\n'.format(error))
        sys.exit(ReturnCodes.UTAH_EXCEPTION_ERROR)
    except Exception:
        sys.stderr.write('Unhandled error in UTAH:\n{}\n'
                         .format(''.join(format_exception(*sys.exc_info()))))
        sys.exit(ReturnCodes.UNHANDLED_ERROR)
    sys.exit(exitstatus)
