# Ubuntu Testing Automation Harness
# Copyright 2013 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for the utah-autorun.sh script."""

import os
import re
import shutil
import stat
import subprocess
import tempfile
import unittest

from utah import template


class TestAutoRun(unittest.TestCase):

    """Tests the functionality of the utah-autorun.sh script."""

    def setUp(self):
        """Set up a temporary work directory for the script."""
        self.workdir = tempfile.mkdtemp(prefix='utah-autotest-')
        # ensure we can safely ignore directories
        os.mkdir('{}/foo'.format(self.workdir))

        #piggy back the foo directory for the run script location
        self.script = os.path.join(self.workdir, 'foo', 'utah-autorun.sh')
        lf = '/dev/null'
        template.write('utah-autorun.sh.jinja2', self.script, log_file=lf)
        os.chmod(self.script, stat.S_IRWXU)

        self._write_script('01_foo', 'echo foo running')
        self._write_script('02_bar', 'echo bar running')

    def tearDown(self):
        """Clean up the work directory."""
        shutil.rmtree(self.workdir)

    def _write_script(self, name, contents):
        with open('{}/{}'.format(self.workdir, name), 'w') as f:
            f.write('#!/bin/sh\n')
            f.write(contents)
            os.fchmod(f.fileno(), stat.S_IRWXU)

    def test_autorun(self):
        """minimal test to make sure autorun works."""
        env = {'WORKDIR': self.workdir, 'RESULTDIR': self.workdir}
        output = subprocess.check_output([self.script], env=env)
        self.assertEqual(output, "foo running\nbar running\n")
        entries = sorted(os.listdir(os.path.join(self.workdir, 'complete')))
        self.assertIsNotNone(re.match(r'01_foo\.\d+', entries[0]))
        self.assertIsNotNone(re.match(r'02_bar\.\d+', entries[1]))
        self.assertEqual(len(entries), 2)
        entries = sorted(os.listdir(self.workdir))
        self.assertEqual(entries[0], 'complete')
        self.assertEqual(entries[1], 'foo')
        self.assertEqual(entries[2], 'inprogress')
        self.assertEqual(len(entries), 3)
