# Ubuntu Testing Automation Harness
# Copyright 2013 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test configuration module."""

import io
import unittest

import json
import jsonschema

from mock import patch

import utah.config


class TestConfig(unittest.TestCase):

    """Test configuration functionality."""

    def test_defaults_validate(self):
        """Verify that defaults validate against the schema."""
        config = utah.config._Config()
        jsonschema.validate(dict(config), config.SCHEMA)

    def test_private_variables_not_in_dict(self):
        """Verify there are no private variables when converting to a dict."""
        config = utah.config._Config()
        config._private = '<value>'
        dictionary = dict(config)
        self.assertNotIn('_private', dictionary)

    def test_invalid_json_file(self):
        """Verify error is printed to stderr on invalid json file."""
        config = utah.config._Config()

        config_file = '<config_file>'
        with patch('utah.config.os.path') as path, \
                patch('__builtin__.open') as open_, \
                patch('sys.stderr') as stderr:
            path.is_file.return_value = True
            path.getsize.return_value = 1
            open_.return_value = io.StringIO(u'not a valid json file')

            config._process_config_file(config_file)

            stderr.write.assert_called_with(
                'ERROR: {} is not a valid JSON file\n'.format(config_file))

    def test_invalid_config_file(self):
        """Verify invalid config files no longer generate errors."""
        config = utah.config._Config()

        config_file = '<config_file>'
        unknown_key = 'unknown_key'
        with patch('utah.config.os.path') as path, \
                patch('__builtin__.open') as open_, \
                patch('sys.stderr') as stderr:
            path.isfile.return_value = True
            path.getsize.return_value = 1
            contents = unicode(json.dumps({unknown_key: 'value'}))
            open_.return_value = io.StringIO(contents)
            config._process_config_file(config_file)

            self.assertFalse(stderr.write.called)

    def test_value_overwritten(self):
        """Verify info when configuration value is overwritten."""
        config = utah.config._Config()

        config_filenames = ['config_file_1', 'config_file_2']
        option_name = 'arch'
        option_value = 'amd64'
        with patch('utah.config.os.path') as path, \
                patch('__builtin__.open') as open_, \
                patch('sys.stderr') as stderr:
            path.isfile.return_value = True
            path.getsize.return_value = 1
            contents = unicode(json.dumps({option_name: option_value}))
            open_.side_effect = (
                io.StringIO(contents),
                io.StringIO(contents),
            )
            config._process_config_file(config_filenames[0])
            config._process_config_file(config_filenames[1])

            stderr.write.assert_called_with(
                'INFO: {} was already set by {}\n'
                .format(option_name, config_filenames[0]))
