# Ubuntu Testing Automation Harness
# Copyright 2013 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for the utah.parser module."""

import os
import sys
import unittest

from tempfile import mkstemp

from mock import patch, DEFAULT

from utah.config import config
from utah.parser import parse_args


class TestParser(unittest.TestCase):

    """Tests that command line parsing works as expected."""

    def setUp(self):
        """Create custom defaults."""
        self.preseed = config.preseed
        self.xml = config.xml
        config.preseed = mkstemp()[1]
        config.xml = mkstemp()[1]

    def tearDown(self):
        """Remove temporary resources."""
        os.unlink(config.preseed)
        os.unlink(config.xml)
        config.preseed = self.preseed
        config.xml = self.xml

    def parse_args_wrapper(self, argv):
        """A wrapper to ensure that runlist argument passes validation.

        :param argv: Command line arguments to be tested
        :type argv: list
        :returns: Parse arguments
        :rtype: `argparse.Namespace`

        """
        with patch.multiple('utah.run',
                            url_argument=DEFAULT,
                            parse_yaml_file=DEFAULT) as patches:
            patches['url_argument'].return_value = '<local_runlist_file>'
            patches['parse_yaml_file'].return_value = {
                'testsuites': [
                    {'name': 'test_suite_name',
                     'fetch_method': 'dev',
                     'fetch_location': '<fetch_location>',
                     },
                ]
            }
            return parse_args(argv)

    def test_gigabytes_default(self):
        """Verify that gigabytes is set to default value if not passed."""
        args = self.parse_args_wrapper(['<runlist_file>'])
        self.assertEqual(args.gigabytes, config.disksizes)

    def test_gigabytes_one_argument(self):
        """Verify that gigabytes is set to the argument passed."""
        gigabytes = '3'
        args = self.parse_args_wrapper(['--gigabytes', gigabytes,
                                        '<runlist_file>'])
        self.assertEqual(args.gigabytes, [gigabytes])

    def test_gigabytes_multiple_arguments(self):
        """Verify that gigabytes is set to the arguments passed."""
        gigabytes = ['3', '5', '7']
        args = self.parse_args_wrapper(['--gigabytes', gigabytes[0],
                                        '--gigabytes', gigabytes[1],
                                        '--gigabytes', gigabytes[2],
                                        '<runlist_file>'])
        self.assertEqual(args.gigabytes, gigabytes)

    def test_logpath(self):
        """Verify that logpath from command line updates log file names."""
        logpath = 'valid-logpath'
        hostname = config.hostname
        orig_lp = config.logpath
        with patch('utah.parser.directory_argument') as directory_argument:
            directory_argument.return_value = logpath
            args = self.parse_args_wrapper(['--logpath', logpath,
                                            '<runlist_file>'])
        # reset values so other tests after this don't get messed up
        newlp = config.logpath
        newlf = config.logfile
        newdl = config.debuglog
        newsl = config.ssh_logfile
        config.logpath = orig_lp
        config._set_log_files()

        self.assertEqual(args.logpath, logpath)
        self.assertEqual(newlp, logpath)
        self.assertEqual(newlf,
                         os.path.join(logpath,
                                      '{}.log'.format(hostname)))
        self.assertEqual(newdl,
                         os.path.join(logpath,
                                      '{}-debug.log'.format(hostname)))
        self.assertEqual(newsl,
                         os.path.join(logpath,
                                      '{}-ssh.log'.format(hostname)))

    @patch('argparse._sys.stderr')
    def test_invalid_logpath(self, stderr):
        """Verify parsing fails if logpath isn't a directory."""
        prog = 'run_utah_tests.py'
        sys.argv[0] = prog
        logpath = 'invalid-logpath'
        with self.assertRaises(SystemExit):
            self.parse_args_wrapper(['--logpath', logpath,
                                     '<runlist_file>'])

        stderr.write.assert_called_with(
            '{}: error: argument -l/--logpath: '
            'Invalid directory: {}\n'
            .format(prog, logpath))
