# Ubuntu Testing Automation Harness
# Copyright 2012 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Test utah.preseed module
"""
from utah.preseed import (
    Preseed,
    Section,
    BlankSection,
    CommentSection,
    ConfigurationSection,
    DuplicatedQuestionName,
    ParsingError,
)

import unittest
from mock import Mock


class TestPreseedLoadDump(unittest.TestCase):

    """Test preseed load/dump."""

    PRESEED = """# Comment

d-i passwd/username string utah
"""

    PRESEED_CONTINUATION_COMMENT = r"""# Comment\

d-i passwd/username string utah
"""

    MALFORMED_PRESEED = """not valid
"""

    DUPLICATED_LINE_PRESEED = """d-i passwd/username string utah
d-i passwd/username string utah
"""

    def test_load(self):
        """Load basic preseed."""
        preseed = Preseed(self.PRESEED.splitlines())

        self.assertEqual(len(preseed.sections), 3)
        section_types = (CommentSection, BlankSection, ConfigurationSection)
        for index, section_type in enumerate(section_types):
            self.assertIsInstance(preseed.sections[index], section_type)

    def test_load_with_continuation_in_comment(self):
        """Continuation character in a comment line is ignored."""
        preseed = Preseed(self.PRESEED_CONTINUATION_COMMENT.splitlines())

        self.assertEqual(len(preseed.sections), 3)
        section_types = (CommentSection, BlankSection, ConfigurationSection)
        for index, section_type in enumerate(section_types):
            self.assertIsInstance(preseed.sections[index], section_type)

    def test_load_with_malformed_input(self):
        """Raise exception with line number when preseed cannot be parsed."""
        with self.assertRaises(ParsingError) as context:
            preseed = Preseed(self.MALFORMED_PRESEED.splitlines())
            preseed.load()

        message = context.exception.message
        self.assertEqual(
            message,
            'Line 1: Unable to parse configuration lines: not valid')

    def test_duplicated_question_name(self):
        """Exception raised on duplicated question name."""
        with self.assertRaises(DuplicatedQuestionName):
            Preseed(self.DUPLICATED_LINE_PRESEED.splitlines())

    def test_dump(self):
        """Dump basic preseed."""
        preseed = Preseed()
        preseed.append(Section.new('# Comment\n'.splitlines()))
        preseed.append(Section.new('\n'.splitlines()))
        preseed.append(Section.new('d-i passwd/username string utah\n'
                                   .splitlines()))
        self.assertEqual(preseed.dump(), self.PRESEED)


class TestPreseedIndexing(unittest.TestCase):

    """Test preseed indexing."""

    PRESEED = """# Comment

d-i passwd/username string utah
"""

    def setUp(self):
        """Set preseed object for test cases."""
        self.preseed = Preseed(self.PRESEED.splitlines())

    def test_get_by_question_name(self):
        """Question can be retrieved by its name."""
        qname = 'passwd/username'
        question = self.preseed[qname]
        self.assertEqual(question.qname, qname)

    def test_key_error_unknown_question_name(self):
        """Key error raised when trying to access an unknown question name."""
        qname = 'unknown/unknwon'

        with self.assertRaises(KeyError):
            self.preseed[qname]

    def test_type_error_question_is_not_string(self):
        """Type error is raised when not using a string as key."""
        with self.assertRaises(TypeError):
            self.preseed[0]

    def test_index_updated_on_question_name_assignment(self):
        """Index is updated when a new value is assigned to question name."""
        qname = 'passwd/username'
        new_qname = 'new/new'
        question = self.preseed[qname]
        question.qname = new_qname

        # Question can no longer be accessed using old qname
        with self.assertRaises(KeyError):
            self.preseed[qname]

        # Question can now be accessed using new qname
        new_question = self.preseed[new_qname]
        self.assertEqual(new_question.qname, new_qname)
        self.assertIs(new_question, question)

    def test_index_updated_on_question_name_prepend(self):
        """Index is updated when some text is prepended to question name."""
        qname = 'passwd/username'
        text_to_prepend = 'prefix-'
        new_qname = text_to_prepend + qname
        question = self.preseed[qname]
        question.qname.prepend(text_to_prepend)

        # Question can no longer be accessed using old qname
        with self.assertRaises(KeyError):
            self.preseed[qname]

        # Question can now be accessed using new qname
        new_question = self.preseed[new_qname]
        self.assertEqual(new_question.qname, new_qname)

    def test_index_updated_on_question_name_append(self):
        """Index is updated when some text is appended to question name."""
        qname = 'passwd/username'
        text_to_append = '-suffix'
        new_qname = qname + text_to_append
        question = self.preseed[qname]
        question.qname.append(text_to_append)

        # Question can no longer be accessed using old qname
        with self.assertRaises(KeyError):
            self.preseed[qname]

        # Question can now be accessed using new qname
        new_question = self.preseed[new_qname]
        self.assertEqual(new_question.qname, new_qname)


class TestPreseedMembership(unittest.TestCase):

    """Test preseed membership."""

    PRESEED = """# Comment

d-i passwd/username string utah
"""

    def setUp(self):
        """Set preseed object for test cases."""
        self.preseed = Preseed(self.PRESEED.splitlines())

    def test_contains(self):
        """True is returned if question name is in preseed."""
        qname = 'passwd/username'
        self.assertIn(qname, self.preseed)

    def test_not_contains(self):
        """False is returned if question name is not in preseed."""
        qname = 'unknown/unknown'
        self.assertNotIn(qname, self.preseed)


class TestPreseedAppendPrepend(unittest.TestCase):

    """Test addition of new sections to a preseed."""

    TIMES = 10  # Used as loop counter when adding multiple sections

    def setUp(self):
        """Set preseed object for test cases and a few sections."""
        self.preseed = Preseed()

    def test_prepend_section(self):
        """Prepend section."""
        for _ in range(self.TIMES):
            self.preseed.prepend(Section())
        self.assertEqual(len(self.preseed.sections), self.TIMES)
        self.assertTrue(all([isinstance(section, Section)
                             for section in self.preseed.sections]))

    def test_prepend_with_ref_section(self):
        """Prepend before a given section."""
        sections = [Section()
                    for section in range(self.TIMES)]
        for section in sections:
            self.preseed.prepend(section)

        index = len(sections) / 2
        ref_section = sections[index]
        new_section = Section()
        self.preseed.prepend(new_section, ref_section)
        self.assertIs(self.preseed.sections[index - 1], new_section)

    def test_prepend_string(self):
        """Make a section from the given string and prepend it."""
        line = '# Comment\n'
        self.preseed.prepend(line)
        self.assertEqual(str(self.preseed.sections[0]), line)

    def test_prepend_with_unknown_ref_section(self):
        """Exception raised if prepending to a section not in the preseed."""
        ref_section = Section()
        new_section = Section()
        with self.assertRaises(ValueError):
            self.preseed.prepend(new_section, ref_section)

    def test_prepend_grouping_blank_section(self):
        """Prepend groups blank sections."""
        # Prepending many blank section results in a single blank section
        # with all the empty lines joined together
        for _ in range(self.TIMES):
            self.preseed.prepend(BlankSection(1))
        self.assertEqual(len(self.preseed.sections), 1)
        self.assertIsInstance(self.preseed.sections[0], Section)

    def test_prepend_grouping_comment_section(self):
        """Prepend groups comment sections."""
        # Prepending many comments section results in a single comment section
        # with all the comment lines joined together
        for _ in range(self.TIMES):
            self.preseed.prepend(CommentSection('# Comment\n'.splitlines()))
        self.assertEqual(len(self.preseed.sections), 1)
        self.assertIsInstance(self.preseed.sections[0], Section)

    def test_append_section(self):
        """Append section."""
        for _ in range(self.TIMES):
            self.preseed.append(Section())
        self.assertEqual(len(self.preseed.sections), self.TIMES)
        self.assertTrue(all([isinstance(section, Section)
                             for section in self.preseed.sections]))

    def test_append_with_ref_section(self):
        """Append after a given section."""
        sections = [Section()
                    for section in range(self.TIMES)]
        for section in sections:
            self.preseed.append(section)

        index = len(sections) / 2
        ref_section = sections[index]
        new_section = Section()
        self.preseed.append(new_section, ref_section)
        self.assertIs(self.preseed.sections[index + 1], new_section)

    def test_append_string(self):
        """Make a section from the given string and append it."""
        line = '# Comment\n'
        self.preseed.append(line)
        self.assertEqual(str(self.preseed.sections[0]), line)

    def test_append_with_unknown_ref_section(self):
        """Exception raised if appending to a section not in the preseed."""
        ref_section = Section()
        new_section = Section()
        with self.assertRaises(ValueError):
            self.preseed.append(new_section, ref_section)

    def test_append_grouping_blank_section(self):
        """Append groups blank sections."""
        # Appending many blank section results in a single blank section
        # with all the empty lines joined together
        for _ in range(self.TIMES):
            self.preseed.append(BlankSection(1))
        self.assertEqual(len(self.preseed.sections), 1)
        self.assertIsInstance(self.preseed.sections[0], Section)

    def test_prepend_append_comment_section(self):
        """Append groups comment sections."""
        # Appending many comments section results in a single comment section
        # with all the comment lines joined together
        for _ in range(self.TIMES):
            self.preseed.prepend(CommentSection('# Comment\n'.splitlines()))
        self.assertEqual(len(self.preseed.sections), 1)
        self.assertIsInstance(self.preseed.sections[0], Section)


class TestSection(unittest.TestCase):

    """Test section object instantiation."""

    def test_new_blank_section(self):
        """Convert empty lines to a blank section."""
        lines_count = 5
        lines = '\n' * lines_count
        section = Section.new(lines.splitlines())
        self.assertIsInstance(section, BlankSection)
        self.assertEqual(str(section), lines)

    def test_new_comment_section(self):
        """Convert lines that start with a hash char to a comment section."""
        lines = """# this is a comment
# this is another comment
"""
        section = Section.new(lines.splitlines())
        self.assertIsInstance(section, CommentSection)
        self.assertEqual(str(section), lines)

    def test_new_configuration_section_single_line(self):
        """Question line => configuration section."""
        lines = """d-i passwd/username string utah
"""
        section = Section.new(lines.splitlines())
        self.assertIsInstance(section, ConfigurationSection)
        self.assertEqual(str(section), lines)

    def test_new_configuration_section_multiple_lines(self):
        """Question lines => configuration section."""
        lines = """d-i preseed/late_command string first_command;\\
second_command;\\
third_command;
"""
        section = Section.new(lines.splitlines())
        self.assertIsInstance(section, ConfigurationSection)

        # Note that ConfigurationSection object doesn't dump the exat input
        # but a one line version of it
        self.assertEqual(str(section), lines.replace('\\\n', ' '))


class TestBlankSection(unittest.TestCase):

    """Test blank section."""

    def test_str(self):
        """Blank section string representation."""
        lines_count = 10
        lines = '\n' * lines_count
        section = BlankSection(lines_count)
        self.assertEqual(section.lines_count, lines_count)
        self.assertEqual(str(section), lines)

    def test_add(self):
        """Add two blank sections."""
        section1 = BlankSection(3)
        section2 = BlankSection(5)
        section = section1 + section2
        self.assertEqual(section.lines_count,
                         section1.lines_count + section2.lines_count)

    def test_iadd(self):
        """Add two blank sections with side effect."""
        lines1 = 3
        lines2 = 5
        section1 = BlankSection(lines1)
        section2 = BlankSection(lines2)
        section1 += section2
        self.assertEqual(section1.lines_count,
                         lines1 + lines2)


class TestCommentSection(unittest.TestCase):

    """Test comment section."""

    def test_str(self):
        """Comment section string representation."""
        comment = '# Comment\n'
        section = CommentSection(comment.splitlines())
        self.assertEqual(str(section), comment)

    def test_add(self):
        """Add two comment sections."""
        comment1 = '# Comment 1\n'
        comment2 = '# Comment 1\n'
        section1 = CommentSection(comment1.splitlines())
        section2 = CommentSection(comment2.splitlines())
        section = section1 + section2
        self.assertEqual(str(section), comment1 + comment2)

    def test_iadd(self):
        """Add two comment sections with side effect."""
        comment1 = '# Comment 1\n'
        comment2 = '# Comment 1\n'
        section1 = CommentSection(comment1.splitlines())
        section2 = CommentSection(comment2.splitlines())
        section1 += section2
        self.assertEqual(str(section1), comment1 + comment2)


class TestConfigurationSection(unittest.TestCase):

    """Test configuration section."""

    def setUp(self):
        """Set section and preseed for test cases."""
        self.line = 'd-i passwd/username string utah\n'
        self.preseed = Preseed()
        self.section = ConfigurationSection(self.line.splitlines())
        self.preseed.append(self.section)

    def test_str(self):
        """Configuration section string representation."""
        self.assertEqual(str(self.section), self.line)

    def test_malformed_input(self):
        """Raise exception when line cannot be parsed correctly."""
        with self.assertRaises(ParsingError):
            ConfigurationSection('not valid\n'.splitlines())

    def test_prepend(self):
        """Prepend new section to this one in the preseed."""
        self.preseed.prepend = Mock()
        new_section = Section()
        self.section.prepend(new_section)
        self.preseed.prepend.assert_called_once_with(new_section, self.section)

    def test_prepend_string(self):
        """Make a section from the given string and prepend it."""
        self.preseed.prepend = Mock()
        line = '# Comment\n'
        self.section.prepend(line)
        args, _kwargs = self.preseed.prepend.call_args
        self.assertEqual(str(args[0]), line)
        self.assertEqual(args[1], self.section)

    def test_append(self):
        """Append new section to this one in the preseed."""
        self.preseed.append = Mock()
        new_section = Section()
        self.section.append(new_section)
        self.preseed.append.assert_called_once_with(new_section, self.section)

    def test_append_string(self):
        """Make a section from the given string and append it."""
        self.preseed.append = Mock()
        line = '# Comment\n'
        self.section.append(line)
        args, _kwargs = self.preseed.append.call_args
        self.assertEqual(str(args[0]), line)
        self.assertEqual(args[1], self.section)

    def test_properties(self):
        """Configuration section splits line into properties."""
        self.assertEqual(self.section.owner, 'd-i')
        self.assertEqual(self.section.qname, 'passwd/username')
        self.assertEqual(self.section.qtype, 'string')
        self.assertEqual(self.section.value, 'utah')

    def test_section_updated(self):
        """Call section_updated in preseed when a propery is updated."""
        method_mock = Mock()
        self.preseed.section_updated = method_mock
        # qname is tested because is the property that the preseed
        # uses to provide index based access
        self.section.qname = 'new_qname'
        method_mock.assert_called_with(self.section,
                                       'qname',
                                       'passwd/username',
                                       'new_qname')
