# Ubuntu Testing Automation Harness
# Copyright 2013 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for the utah.process module."""

import os
import tempfile
import unittest

from utah.process import _get_process_children, pid_in_use, run


class TestProcess(unittest.TestCase):

    """Tests the functions of the utah.process module."""

    def setUp(self):
        """Create files needed for testing."""
        _, self.tmpfile = tempfile.mkstemp()

    def tearDown(self):
        """Delete files created for testing."""
        os.unlink(self.tmpfile)

    def test_is_running(self):
        """Ensure simple test of pid_in_use works for a pid running."""
        pid = os.getpid()
        self.assertIsNotNone(pid_in_use(pid))
        self.assertIsNotNone(pid_in_use(pid, ['test']))
        # make sure we don't return when the command line doesn't match
        self.assertIsNone(pid_in_use(pid, ['blah']))

    def test_is_not_running(self):
        """Ensure simple test of pid_in_use works for a pid not running."""
        pid = os.fork()
        if pid == 0:
            os.execv('/bin/true', ['/bin/true'])
        os.waitpid(pid, 0)
        # the child should be dead so this will return None
        self.assertIsNone(pid_in_use(pid))
        self.assertIsNone(pid_in_use(pid, ['true']))

    def test_run_basic(self):
        """Ensure the run API passes a basic test."""
        rc, timedout, out, err = run('echo err >&2; echo out', '/', 0, False)
        self.assertEqual(0, rc)
        self.assertFalse(timedout)
        self.assertEqual(out, 'out\n')
        self.assertEqual(err, 'err\n')

    def test_run_lots(self):
        """Ensure the run API can handle lots of stdout."""
        fname = __file__
        with open(fname, 'r') as f:
            contents = f.read()

        # run lots of times since this doesn't always fail
        for _ in range(30):
            stdout = run('cat {}'.format(fname), './', 0, False)[2]
            self.assertEqual(contents, stdout)

    def test_run_err(self):
        """Ensure the run API gets proper error code from command."""
        rc, timedout, out, err = run('echo out; exit 123', '/', 0, False)
        self.assertEqual(123, rc)
        self.assertFalse(timedout)
        self.assertEqual(out, 'out\n')
        self.assertEqual(err, '')

    def test_run_timeout(self):
        """Ensure the run API handles timeouts."""
        rc, timedout, out, err = run('echo out; sleep 2', '/', .5, False)
        self.assertTrue(timedout)
        self.assertEqual(15, rc)  # 15 is the rc for a killed process
        self.assertEqual(out, 'out\n')
        self.assertEqual(err, '')

    def test_run_kill(self):
        """Ensure the run API does a sigkill when needed."""
        with open(self.tmpfile, 'w') as f:
            f.write('''#!/bin/bash
trap "echo ignorning" TERM
sleep 2
exit 3
''')

        cmd = 'bash {}'.format(self.tmpfile)
        rc, timedout, out, err = run(cmd, '/', .5, False)
        self.assertEqual(15, rc)
        self.assertTrue(timedout)
        self.assertEqual(1, len(_get_process_children(os.getpid())))
