# Ubuntu Testing Automation Harness
# Copyright 2013 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for the utah.provisioning.data module."""

import os
import shutil
import subprocess
import tempfile
import unittest

from utah.provisioning.data import ProvisionData


class TestProvisionData(unittest.TestCase):

    """Tests the functions of the ProvisionData class."""

    def setUp(self):
        """Create temp files needed for testing."""
        self.files = []
        for x in xrange(4):
            fd, f = tempfile.mkstemp(prefix='utah_provision_data')
            self.files.append(f)
            os.write(fd, '{}'.format(x))
            os.close(fd)

        self.initrd = tempfile.mkdtemp(prefix='utah_initrd')

    def tearDown(self):
        """Clean up the artifacts create in setUp."""
        for f in self.files:
            os.unlink(f)
        shutil.rmtree(self.initrd)

    def test_add_files(self):
        """Ensure the add_files API works."""
        pd = ProvisionData()
        pd.add_files(self.files[0], '/foo')
        pd.add_files((self.files[1], self.files[2]), '/bar')

        pd.update_initrd(self.initrd)

        dstdir = os.path.join(self.initrd, 'utah-copy-files')
        for x in xrange(3):
            bname = os.path.basename(self.files[x])
            dstfile = os.path.join(dstdir, bname)
            self.assertTrue(os.path.isfile(dstfile))
            with open(dstfile) as f:
                content = f.read()
                self.assertEqual(content, str(x))

        script = os.path.join(self.initrd, 'utah-copy-files.sh')
        self.assertTrue(os.path.isfile(script))

        #update the copy script to copy to our tmpdir so we can run locally
        escaped_dir = self.initrd.replace('/', '\/')
        exp = 's/\/target/{}/g'.format(escaped_dir)
        subprocess.check_call(['sed', '-i', '-e', exp, script])

        escaped_dir = self.initrd.replace('/', '\/')
        exp = 's/\/utah-copy-files/{}\/utah-copy-files/g'.format(escaped_dir)
        subprocess.check_call(['sed', '-i', '-e', exp, script])

        subprocess.check_call(['sh', script])

    def test_add_cmds(self):
        """Ensure the add_autostart_cmd API works."""
        pd = ProvisionData()
        for x in xrange(3):
            pd.add_autostart_cmd('{}_foo'.format(x), str(x))

        pd.update_initrd(self.initrd)
        for x in xrange(3):
            fname = os.path.join(
                self.initrd, 'utah-autorun', '{}_foo'.format(x))
            self.assertTrue(os.path.isfile(fname))
            with open(fname) as f:
                cmd = f.readlines()[-1]
                self.assertEqual(cmd, '{}\n'.format(x))
