# Ubuntu Testing Automation Harness
# Copyright 2013 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test syslog processing functionality."""


import logging
import mock
import os
import shutil
import socket
import time
import tempfile
import threading
import unittest
import uuid

from mock import patch

from utah.config import config
from utah.exceptions import UTAHException
from utah.provisioning.exceptions import UTAHProvisioningException
from utah.provisioning.rsyslog import RSyslog
from utah.timeout import UTAHTimeout


class TestRsyslog(unittest.TestCase):

    """Test syslog processing code."""

    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)
    logging.getLogger('rsyslog').addHandler(logging.StreamHandler())

    @staticmethod
    def producer(port, messages):
        """Send messages to a port for testing."""
        print('sending fake messages to port: {}'.format(port))
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.connect(('localhost', port))
        for m in messages:
            time.sleep(.5)
            s.send(m)

    @staticmethod
    def file_producer(f, messages, truncate=False):
        """Produce a file for testing."""
        for m in messages:
            f.write(m)
            f.write('\n')
            f.flush()
            time.sleep(1)
            if truncate:
                f.truncate(0)
                f.flush()
                time.sleep(1)
                truncate = False

    def setUp(self):
        """Create a temporary directory for logging."""
        self.logdir = tempfile.mkdtemp()
        self.logfile = os.path.join(self.logdir, 'tmp.log')

    def tearDown(self):
        """Cleanup artifacts from setUp."""
        shutil.rmtree(self.logdir)

    def test_easy(self):
        """Make sure we can match a typical first message."""
        steps = [
            {
                "message": "test_easy",
                "pattern": ".*log source = /proc/kmsg started",
            },
        ]

        messages = [
            'test 1',
            '<13>Feb 20 13:38:55 host user: log source = /proc/kmsg started',
        ]
        r = RSyslog("utah-test", self.logdir)
        threading.Thread(target=self.producer, args=(r.port, messages)).start()
        r._wait_for_steps(steps, 5, self.logfile, {})

    def test_fail_pattern(self):
        """Exception is raised on fail_pattern match."""
        steps = [
            {
                "message": "test_late_command_failure",
                "pattern": "finished",
                "fail_pattern": "failure",
            },
        ]

        messages = [
            'started',
            'failure',
            'finished',
        ]

        r = RSyslog("utah-test", self.logdir)
        threading.Thread(target=self.producer, args=(r.port, messages)).start()
        with self.assertRaises(UTAHException):
            r._wait_for_steps(steps, 5, self.logfile, {})

    def test_future(self):
        """Test handling a missing message."""
        steps = [
            {
                "message": "test_future",
                "pattern": ".*log source = /proc/kmsg started",
                # set a really small timeout to prove we properly calculate
                # future timeout logic
                "booted": True,
            },
            {
                "message": "hit skipped",
                "pattern": "hit skipped",
            },
            {
                "message": "message 3",
                "pattern": "message 3",
            },
        ]

        messages = [
            'test 1',
            'hit skipped',
            'message 3',
        ]

        r = RSyslog("utah-test", self.logdir)
        threading.Thread(target=self.producer, args=(r.port, messages)).start()

        def booted_cb(match):
            self.test_future_booted = True
            self.assertEqual(match.string.rstrip(), steps[1]['message'])

        self.test_future_booted = False
        with patch.object(r.logger, 'warn') as mock_method:
            r._wait_for_steps(steps, 5, self.logfile, {'booted': booted_cb})
            mock_method.assert_called_with(
                'Expected pattern missed, matched future pattern: %s',
                'hit skipped')

        self.assertTrue(self.test_future_booted, 'booted callback not made')

    def test_future_array(self):
        """Test handling a missing message if the pattern is an array."""
        steps = [
            {
                "message": "test_future",
                "pattern": [".*"],
                "booted": True,
            },
            {
                "message": "hit skipped",
                "pattern": ["hit skipped"],
            },
            {
                "message": "this will hit step three",
                "pattern": [".*this will hit step three"],
            },
        ]

        messages = [
            'this will hit step one',
            'this will hit step three',
        ]

        r = RSyslog("utah-test", self.logdir)
        threading.Thread(target=self.producer, args=(r.port, messages)).start()

        def booted_cb(match):
            self.test_future_booted = True

        self.test_future_booted = False
        with patch.object(r.logger, 'warn') as mock_method:
            r._wait_for_steps(steps, 5, self.logfile, {'booted': booted_cb})
            mock_method.assert_called_with(
                'Expected pattern missed, matched future pattern: %s',
                steps[2]['pattern'][0])

        self.assertTrue(self.test_future_booted, 'booted callback not made')

    def test_callbacks(self):
        """Make sure wait_for_steps works."""
        steps = [
            {
                "message": "test callbacks",
                "pattern": ".*log source = /proc/kmsg started",
                "booted": True,
            },
            {
                "message": "hit skipped",
                "pattern": "hit skipped",
            },
            {
                "message": "message 3",
                "pattern": "message 3",
                "blah": True,
            },
        ]

        messages = [
            'test 1',
            'hit skipped',
            'message 3',
        ]

        r = RSyslog("utah-test", self.logdir)
        threading.Thread(target=self.producer, args=(r.port, messages)).start()

        def booted_cb(match):
            self.test_callbacks_booted = True
            self.assertEqual(match.string.rstrip(), messages[1])

        def blah_cb(match):
            self.test_callbacks_blah = True
            self.assertEqual(match.string.rstrip(), messages[-1])

        self.test_callbacks_booted = self.test_callbacks_blah = False
        callbacks = {
            'booted': booted_cb,
            'blah': blah_cb,
        }
        r._wait_for_steps(steps, 5, "/dev/null", callbacks)
        self.assertTrue(self.test_callbacks_booted, 'booted callback not made')
        self.assertTrue(self.test_callbacks_blah, 'blah callback not made')

    def test_install(self):
        """Make sure wait_for_install works."""
        messages = [
            'blah',
            'ubiquity[11]: Ubiquity 1',
            'finish-install: umount',
        ]

        def cb(match):
            self.test_install_cb = True

        self.test_install_cb = False
        r = RSyslog("utah-test", self.logdir)
        r.logger = mock.Mock()
        threading.Thread(target=self.producer, args=(r.port, messages)).start()
        r.wait_for_install(cb)
        self.assertFalse(r.logger.error.called)
        self.assertTrue(self.test_install_cb)

        r = RSyslog("utah-test", self.logdir)
        messages[-1] = 'utah: Late command failure detected'
        threading.Thread(target=self.producer, args=(r.port, messages)).start()
        with self.assertRaises(UTAHException):
            r.wait_for_install()

    def test_booted(self):
        """Make sure wait_for_booted works."""
        tst_uuid = str(uuid.uuid4())
        messages = [
            'test 1',
            '<13>Feb 20 13:38:55 host user: log source = /proc/kmsg started',
            '<13>Feb 21 13:40:55 host user: Machine UUID: {}'.format(tst_uuid),
        ]

        r = RSyslog("utah-test", self.logdir)
        r.logger = mock.Mock()
        threading.Thread(target=self.producer, args=(r.port, messages)).start()
        r.wait_for_booted(tst_uuid)
        self.assertFalse(r.logger.error.called)

        threading.Thread(target=self.producer, args=(r.port, messages)).start()
        with self.assertRaises(UTAHProvisioningException):
            r.wait_for_booted('bad')

    def test_booted_workaround(self):
        """Test that wait_for_booted doesn't raise an exception on timeout."""
        orig = config.boottimeout
        try:
            config.boottimeout = 0.5
            r = RSyslog("utah-test", self.logdir)
            r.wait_for_booted(None)
        finally:
            config.boottimeout = orig

    def test_wait_for_run(self):
        """Test that wait_for_utah_run works."""
        messages = [
            'test 1',
            'utah: completed /etc/utah/autorun/01_run-utah rc=7',
        ]

        def cb(match):
            self.utah_run_rc = int(match.group(1))

        r = RSyslog("utah-test", self.logdir)
        threading.Thread(target=self.producer, args=(r.port, messages)).start()
        r.wait_for_utah_run(cb)
        self.assertEqual(7, self.utah_run_rc)

    def usefile(self, truncate):
        """Make sure wait_for_booted works when tailing a file."""
        steps = [
            {
                "message": "truncate test",
                "pattern": ".*log source = /proc/kmsg started",
            },
        ]

        messages = [
            'test 1',
            'lllllllllllllllllllllllllllllllllllllll',
            'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa',
            '<13>Feb 20 13:38:55 host user: log source = /proc/kmsg started',
        ]

        with tempfile.NamedTemporaryFile() as f:
            r = RSyslog("utah-test", self.logdir, f.name)
            t = threading.Thread(
                target=self.file_producer,
                args=(f, messages, truncate))
            t.start()
            r._wait_for_steps(steps, 5, self.logfile, {})
            t.join()

    def test_usefile(self):
        """Test tailing a file."""
        self.usefile(False)

    def test_usefile_truncation(self):
        """Test tailing a file with truncation."""
        self.usefile(True)

    def test_timeout(self):
        """Verify UTAHTimeout is raised on timeout."""
        steps = [
            {
                "message": "message_1",
                "pattern": "finished",
            },
            {
                "message": "message_2",
                "pattern": "pattern",
            },
            {
                "message": "message_3",
                "pattern": "pattern",
            },
        ]

        messages = ['message_1', 'message_2', 'message_4']
        pattern = (r'Timeout \(\d+\) occurred for {} message.\n'
                   r'Remaining messages: {}'
                   .format(steps[0]['message'],
                           ', '.join([step['message']
                                     for step in steps[1:]])))

        r = RSyslog('utah-test', self.logdir)
        thread = threading.Thread(target=self.producer,
                                  args=(r.port, messages))
        thread.daemon = True
        thread.start()
        start = time.time()
        with self.assertRaisesRegexp(UTAHTimeout, pattern):
            r._wait_for_steps(steps, 5, self.logfile, {})
        # assert the timeout was within 2 seconds of what we expect
        self.assertLess(abs(time.time() - start - 5), 2)
