# Ubuntu Testing Automation Harness
# Copyright 2013 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for the phablet script."""

from __future__ import print_function

import mock
import os
import shutil
import subprocess
import tempfile
import unittest

from run_utah_phablet import (
    _check_args,
    _flash,
    _get_parser,
    _install_utah,
    _network,
    _save_props,
    main,
)

import utah.provisioning.touch as touch


class TestPhabletScript(unittest.TestCase):

    """Test the phablet script."""

    def setUp(self):
        """Create things needed for testing."""
        self.tmpdir = tempfile.mkdtemp()
        self.args = _get_parser().parse_args(['--results-dir', self.tmpdir])
        self.device = touch.Device(print)
        self.device._adb_operation = mock.Mock()
        self.device.detect_type = mock.Mock()

    def tearDown(self):
        """Cleanup artifacts from the setUp method."""
        shutil.rmtree(self.tmpdir)

    @mock.patch('utah.process.run')
    def test_adb_shell(self, run):
        """Ensure the adb_shell command can track failed commands."""
        run.return_value = [0, 0, 'stdout\nADB_RC=0']
        self.device._adb_shell(self.args, 'true')

        run.return_value = [0, 0, 'stdout\nADB_RC=1']
        with self.assertRaises(touch.ADBShellError):
            self.device._adb_shell(self.args, 'false')

    @mock.patch('utah.provisioning.touch.Device._adb_shell')
    def test_save_props(self, adb):
        """Make sure save_props works."""
        adb.return_value = 'foobar'
        _save_props(self.device, self.tmpdir)
        fname = os.path.join(self.tmpdir, 'adb.props')
        with open(fname) as f:
            self.assertEqual(adb.return_value, f.read())

    @mock.patch('subprocess.check_call')
    @mock.patch('sys.exit')
    def _test_phablet_cmd(self, func, args, exit, check_call):
        """Ensure functions calling phablet commands handle errors.

        :return: the mock object used for the call

        """
        func(*args)
        self.assertFalse(exit.called)

        check_call.side_effect = subprocess.CalledProcessError(1, 'blah')
        func(*args)
        self.assertTrue(exit.called)

        return check_call

    def test_flash(self):
        """ensure the flash function handles errors properly."""
        device = mock.Mock()
        self._test_phablet_cmd(_flash, (self.args, device))

    def test_flash_uri_revision(self):
        """ensure the flash function handles revision arg."""
        args = _get_parser().parse_args(['-r', 'bar'])
        device = mock.Mock()
        check_call = self._test_phablet_cmd(_flash, (args, device))
        self.assertIn('bar', check_call.call_args[0][0])

    @mock.patch('time.sleep')
    def test_network(self, sleep):
        """ensure the network commands behaves as expected."""
        self._test_phablet_cmd(_network, (self.args,))
        args = _get_parser().parse_args(['--network-file', 'blah'])
        check_call = self._test_phablet_cmd(_network, (args,))
        self.assertEqual('-n', check_call.call_args[0][0][-2])
        self.assertEqual('blah', check_call.call_args[0][0][-1])

    @mock.patch('utah.provisioning.touch.Device._adb_shell')
    def test_install_utah(self, adbshell):
        """ensure utah install logic is okay by ensure proper PPA is used."""
        _install_utah(self.device, self.args.ppa, self.args.branch)
        self.assertIn(self.args.ppa, adbshell.call_args_list[1][0][0])

        adbshell.reset_mock()
        args = _get_parser().parse_args(['--ppa', 'blah', '-b', 'blah'])
        _install_utah(self.device, args.ppa, args.branch)
        self.assertIn(args.ppa, adbshell.call_args_list[1][0][0])
        self.assertIn('bzr branch blah', adbshell.call_args_list[-1][0][0])

    @mock.patch('run_utah_phablet._flash')
    @mock.patch('run_utah_phablet._network')
    @mock.patch('run_utah_phablet._install_utah')
    @mock.patch('run_utah_phablet._run_utah')
    @mock.patch('run_utah_phablet._save_props')
    @mock.patch('utah.provisioning.touch.Device')
    def _test_main(self, args, device, sprops, rutah, iutah, net, flash):
        main(args)
        return (flash.called, net.called, iutah.called, rutah.called,
                sprops.called)

    def test_main_all(self):
        """ensure we handle the combination of options correctly."""
        r = self._test_main(['--runlist', 'blah'])
        self.assertEqual(r, (True, True, True, True, True))

    def test_main_skip_install(self):
        """ensure we handle the skip-install correctly."""
        r = self._test_main(['--skip-install'])
        self.assertEqual(r, (False, True, True, False, True))

    def test_main_skip_network(self):
        """ensure we handle the skip-network correctly."""
        r = self._test_main(['--skip-network'])
        self.assertEqual(r, (True, False, True, False, True))

    def test_main_skip_utah(self):
        """ensure we handle the skip-utah correctly."""
        r = self._test_main(['--skip-utah'])
        self.assertEqual(r, (True, True, False, False, True))

    def test_main_skip_all(self):
        """ensure we handle skipping everything correctly."""
        r = self._test_main(['--skip-utah', '--skip-network', '--skip-ins'])
        self.assertEqual(r, (False, False, False, False, True))

    @mock.patch('sys.exit')
    def test_main_bad_dir(self, exit):
        """ensure we handle a bad results-dir."""
        args = ['--results-dir', '/does not exist']
        _check_args(_get_parser().parse_args(args), self.device)
        self.assertTrue(exit.called)

        exit.reset_mock()
        args = ['--results-dir', '/bin/true']  # a file, not a directory
        _check_args(_get_parser().parse_args(args), self.device)
        self.assertTrue(exit.called)
