# Ubuntu Testing Automation Harness
# Copyright 2013 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Probe API for the client"""

import contextlib
import importlib
import logging

PROBES = {
    # key = object in module that implements a probe's start/stop functions
    # val = module to import
    'battery': 'utah.client.probe.battery',
    'eventstat': 'utah.client.probe.eventstat',
    'contextswitch': 'utah.client.probe.contextswitch',
    'interrupts': 'utah.client.probe.interrupts',
}


class ProbeManager(object):

    """Provides a simple class to interact with probes."""

    def __init__(self):
        self._available = {}
        for probe, module in PROBES.iteritems():
            m = importlib.import_module(module)
            self._available[probe] = getattr(m, probe)
        self._enabled = None
        self.output_dir = '/tmp'

    def enable_probes(self, probes):
        """Enable each probe in the list to be run."""
        self._enabled = []
        for p in probes:
            if p in self._available:
                probe = self._available[p]
                probe.name = p
                self._enabled.append(probe)
            else:
                logging.warn('Probe "%s" not known to utah client', p)

    def start(self):
        """Start executing all configured probes.

        :returns: returns each probe that was started and its corresponding
                  probe data.
        :rtype: list(tuple)

        """
        probes_data = []
        for p in self._enabled:
            start_data = p.start(self.output_dir)
            probes_data.append((p, start_data))
        return probes_data

    def stop(self, probes_data):
        """Stop executing all configured probes.

        :return: results from each probe
        :rtype: dict

        """
        results = {}
        for p, d in probes_data:
            d = p.stop(d)
            results[p.name] = d
        return results

    @contextlib.contextmanager
    def run(self, probe_data):
        """Run the start/stop probe operations within a context manager.

        :return: start/stop information for each enabled probe.
        :rtype: dict with keys compatible for make_result function.

        """
        if self._enabled is None:
            raise RuntimeError('probes API has not been configured')

        results = None
        start_data = self.start()
        try:
            yield
        finally:
            results = self.stop(start_data)
            probe_data.update(results)


probes = ProbeManager()
