# Ubuntu Testing Automation Harness
# Copyright 2013 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Helpers for battery measurements."""

import os
import re
import logging

from utah.client.exceptions import UTAHClientError


class _SWBatteryImpl(object):

    """Gathers battery information via the kernel's "power_supply" class.

    .. seealso::
       https://www.kernel.org/doc/Documentation/power/power_supply_class.txt

    """

    POWER_SUPPLY_DIR = '/sys/class/power_supply'
    BATTERY_DIR_REGEX = re.compile(r'bat.*', re.IGNORECASE)
    BATTERY_FILE_REGEX = re.compile(r'(.*_now|capacity|status)')

    def __init__(self):
        self.filenames = None

    @staticmethod
    def get_name():
        """return the implementation name displayed to the user."""
        return 'software'

    def get_data(self):
        """Get data from the battery information files.

        Every file is read and the contents of the file is written to a
        dictionary using as key the basename of the file.

        :returns: Data as read from the battery information files
        :rtype: dict

        """
        def read_file(filename):
            """Read contents of a file and try to convert data to integer

            :param filename: Path to file to be read
            :type filename: str
            :returns: File contents converted to integer if possible
            :rtype: str | int
            """
            with open(filename, 'r') as f:
                data_str = f.read().strip()
                try:
                    data_int = int(data_str)
                    return data_int
                except ValueError:
                    return data_str

        # Cache file names for later use
        if self.filenames is None:
            self.filenames = self._get_files()

        return {os.path.basename(filename): read_file(filename)
                for filename in self.filenames}

    def _get_files(self):
        """Figure out which files to look at.

        :returns: Paths to files with battery information.
        :rtype: list(str)

        """
        if not os.path.isdir(self.POWER_SUPPLY_DIR):
            logging.warning('Power supply directory not found: {!r}'
                            .format(self.POWER_SUPPLY_DIR))
            return []

        entries = [os.path.join(self.POWER_SUPPLY_DIR, entry)
                   for entry in os.listdir(self.POWER_SUPPLY_DIR)
                   if self.BATTERY_DIR_REGEX.match(entry)]
        subdirectories = [entry for entry in entries
                          if os.path.isdir(entry)]

        if not subdirectories:
            logging.warning('Battery information directory not found')
            return []

        if len(subdirectories) > 1:
            logging.error('Multiple battery information directories found: '
                          '{}'.format(', '.join(subdirectories)))
            return []

        subdirectory = subdirectories[0]
        entries = [os.path.join(subdirectory, filename)
                   for filename in os.listdir(subdirectory)
                   if self.BATTERY_FILE_REGEX.match(filename)]
        filenames = [filename for filename in entries
                     if os.path.isfile(filename)]

        logging.debug('Battery information files found: {}'
                      .format(', '.join(filenames)))
        return filenames


class _Battery(object):
    def __init__(self):
        self._impl = _SWBatteryImpl()
        self._impls = {}

        for name, cls in globals().iteritems():
            if name.endswith("BatteryImpl"):
                self._impls[cls.get_name()] = cls

    def implementations(self):
        """return the battery implementation names."""
        return self._impls.keys()

    def get_data(self):
        """return the battery data from the implementation."""
        if not self._impl:
            raise UTAHClientError('No battery implementation set')
        return self._impl.get_data()

    def start(self, outputdir):
        """Capture battery information at start.

        :return: battery info at start
        :rtype: dict

        """
        return self.get_data()

    def stop(self, start_data):
        """Capture battery information at end.

        :return: battery info at start and end of probe
        :rtype: dict

        """
        return {'start': start_data, 'end': self.get_data()}

    def _get_impl(self):
        if not self._impl:
            return None
        return self._impl.get_name()

    def _set_impl(self, name):
        self._impl = self._impls[name]()

    implementation = property(_get_impl, _set_impl)

battery = _Battery()
