# Ubuntu Testing Automation Harness
# Copyright 2012 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Provide common functions and content for self tests."""


import os
import shutil

from utah.client.common import UTAH_DIR, MASTER_RUNLIST
from utah.client.probe import probes


def get_module_path():
    """Return the path for the 'utah' directory.

    This is used in tests to refer to the location where the branch resides.
    It is used in conjunction with running the self tests via:

    sudo nosetests utah/client/tests

    from the top-level directory in the repo.

    """
    import utah

    module_path = os.path.dirname(utah.__file__)

    del utah

    return module_path


master_runlist_content = """# utah/self_test.py master runlist
# needed for utah/self_test.py runs
---
testsuites:
  - name: examples
    fetch_method: bzr-export
    fetch_location: {REPO_PATH}/client/examples/examples
  - name: utah_tests
    fetch_method: bzr-export
    fetch_location: {REPO_PATH}/client/examples/utah_tests
  - name: utah_tests_sample
    fetch_method: bzr-export
    fetch_location: {REPO_PATH}/client/examples/utah_tests_sample
""".format(REPO_PATH=get_module_path())

master_runlist = os.path.join(UTAH_DIR, MASTER_RUNLIST)
master_runlist_bak = master_runlist + ".bak"
ts_dir = os.path.join(UTAH_DIR, 'testsuites')
examples_source = os.path.join(get_module_path(), 'client', 'examples')


def setUp():
    """Create a sample master.run file.

    This file will copy our 'examples' directory to the testing directory
    (usually /var/lib/utah).

    """
    # If we're not root don't bother
    if os.geteuid() != 0 and os.getuid() != 0:
        raise RuntimeError('These tests must be run as root')

    probes.enable_probes([])

    # some tests will destroy a directory it was in, so we need to start
    # in a directory that exists to make run_cmd safe in any context.
    os.chdir('/tmp')

    # save a copy of the master runlist
    if os.path.exists(master_runlist):
        os.rename(master_runlist, master_runlist_bak)

    if os.path.exists(ts_dir):
        shutil.rmtree(ts_dir)
    for x in os.listdir(examples_source):
        p = os.path.join(examples_source, x)
        if os.path.isdir(p):
            # NOTE: the runner expects the directory tree for test suite to
            # look like:
            #   /var/lib/utah/testsuites/<tsname>/<tsname>/tslist.run
            shutil.copytree(p, os.path.join(ts_dir, x, x))

    with open(master_runlist, 'w') as fp:
        fp.write(master_runlist_content)


def tearDown():
    """Clean up after ourselves."""
    # restore the copy of the master runlist
    if os.path.exists(master_runlist_bak):
        os.rename(master_runlist_bak, master_runlist)
    elif os.path.exists(master_runlist):
        os.remove(master_runlist)

    if os.path.exists(ts_dir):
        shutil.rmtree(ts_dir)


def _get_partial_state_file(filename):
    """Read a partial state file from disk.

    :returns: A pre-defined partially run test suite state.
    :rtype: dict

    """
    state = None

    with open(filename, 'r') as fp:
        state = fp.read()

    return state

partial_state_file_content = _get_partial_state_file(
    os.path.join(os.path.dirname(__file__), "..", "examples",
                 "test_state_partial_inprogress.yaml")
)

partial_state_file_content_run_all = _get_partial_state_file(
    os.path.join(os.path.dirname(__file__), "..", "examples",
                 "test_state_partial_run_all.yaml")
)

partial_state_file_content_done_all_failed = _get_partial_state_file(
    os.path.join(os.path.dirname(__file__), "..", "examples",
                 "test_state_partial_all_failed.yaml")
)
