# Ubuntu Testing Automation Harness
# Copyright 2012 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test utah client common functions."""

import os
import signal
import unittest

# REQUIRES that the top level utah package be in the Python path.
from utah.client.common import (
    CONFIG,
    debug_print,
    get_media_info,
    get_host_info,
    get_product_uuid,
    MEDIA_INFO,
    run_cmd,
)

from utah.process import _get_process_children

from utah.client.tests.common import (  # NOQA
    setUp,     # Used by nosetests
    tearDown,  # Used by nosetests
)


class TestCommon(unittest.TestCase):

    """Tests for utah.client.common."""

    def test_get_product_uuid(self):
        """Test that product_uuid is returned when possible."""

        product_uuid = get_product_uuid()

        if os.access('/sys/class/dmi/id/product_uuid', os.R_OK):
            self.assertIsNotNone(product_uuid)

    def test_get_media_info(self):
        """Test that if there is a media-info file that it is parsed."""
        media_info = get_media_info()

        print("media_info: {}".format(media_info))

        if os.path.exists(MEDIA_INFO):
            self.assertFalse(media_info == '' or media_info is None)
        else:
            self.assertTrue(media_info == 'unknown')

    def test_get_host_info(self):
        """Test that get_host_info returns results."""
        host_info = get_host_info()

        print("host_info: {}".format(host_info))

        self.assertTrue(host_info is not None)

    def test_run_cmd(self):
        """Test running a command."""
        result = run_cmd('echo 3')
        self.assertEqual(result['stdout'], '3\n')

        result = run_cmd('echo blah >&2')
        self.assertEqual(result['stderr'], 'blah\n')

        result = run_cmd('/bin/true')
        self.assertEqual(result['returncode'], 0)

        result = run_cmd('/bin/false')
        self.assertEqual(result['returncode'], 1)

    def test_run_cmd_utf8(self):
        """Test running a command with utf-8 specific output.

        Prints skull and crossbones character which our code will convert to
        a ascii. Since this character isn't in the ASCII set, it will convert
        it as a "?"

        """
        result = run_cmd('/bin/echo -e "\xE2\x98\xA0"')
        self.assertEqual(result['stdout'], '?\n')
        from utah.client.result import ResultYAML
        res = ResultYAML('/dev/null', False)
        res.status = 'PASS'
        res.add_result(result)
        res.result()

    def test_run_cmd_cp1252(self):
        r"""Test running a command with cp1252 specific output.

        '\x97' is the "EM DASH" (long dash) character that's invalid in utf-8.
        This asserts we can handle any random output a runlist might produce.
        The code won't know how to properly convert the character and it will
        end up as a "?"

        """
        result = run_cmd('/bin/echo -e "\x97"')
        self.assertEqual(result['stdout'], '?\n')

    def test_debug_print(self):
        """Test that debug_print prints at the appropriate times.

        It shouldn't print when DEBUG is False unless 'debug' is passed in and
        is True.

        """
        old_debug = CONFIG['DEBUG']

        CONFIG['DEBUG'] = True
        if not debug_print("testing debug_print"):
            raise Exception("didn't print with CONFIG['DEBUG'] set to True")

        CONFIG['DEBUG'] = False

        debug = debug_print("shouldn't be printed "
                            "unless common.CONFIG['DEBUG'] is True")
        self.assertEqual(debug,
                         CONFIG['DEBUG'],
                         "printed with CONFIG['DEBUG'] set to False")
        debug = debug_print("should be printed (using 'force=True')",
                            force=True)
        self.assertEqual(debug, True)

        CONFIG['DEBUG'] = old_debug

    def test_run_as(self):
        """Test that run_as functions correctly."""
        user = "nobody"
        result = run_cmd("touch /tmp/run_as_test.tmp", run_as=user)
        print("result: {}".format(result))

        self.assertEqual(result['returncode'], 0)
        self.assertEqual(result['user'], user)

        result = run_cmd("rm /tmp/run_as_test.tmp", run_as=user)
        print("result: {}".format(result))

        self.assertEqual(result['returncode'], 0)
        self.assertEqual(result['user'], user)

    def test_run_as_default(self):
        """Test that the default user is correctly identified."""
        result = run_cmd("touch /tmp/run_as_test.tmp")
        print("result: {}".format(result))

        self.assertEqual(result['returncode'], 0)
        self.assertNotEqual(result['user'], "unknown")

        result = run_cmd("rm /tmp/run_as_test.tmp")
        print("result: {}".format(result))

        self.assertEqual(result['returncode'], 0)
        self.assertNotEqual(result['user'], "unknown")

    def test_get_process_children(self):
        """Test that the get_process_children function works."""
        children = []
        for _ in xrange(3):
            pid = os.fork()
            if pid == 0:
                # child to wait
                os.execv('/bin/sleep', ['/bin/sleep', '5'])
            children.append(pid)

        ret = _get_process_children(str(os.getpid()))
        # the last PID from childre will be the "ps" process we ran, so
        # we can ignore that one
        ret.pop()
        for pid in children:
            os.kill(pid, signal.SIGTERM)
            os.waitpid(pid, 0)
        self.assertEqual(children, ret)
