# Ubuntu Testing Automation Harness
# Copyright 2012 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test schema parsing functionality."""


import jsonschema
import os
import unittest
import yaml

from utah.client.runner import Runner


yaml_content = """---
timeout: 101
repeat_count: 99
testsuites:
    - name: testsuite1
      fetch_method: bzr
      fetch_location: lp:utah
    - name: testsuite2
      fetch_method: bzr
      fetch_location: lp:utah
"""
yaml_content_include = """---
testsuites:
    - include: master.run.other
    - name: testsuite1
      fetch_method: bzr
      fetch_location: lp:utah
    - name: testsuite2
      fetch_method: bzr
      fetch_location: lp:utah
"""

yaml_content_bad = [
    """---
testsuties:
    - name: testsuite1
      fetch_method: bzr
      fetch_location: lp:utah
""",  # typo
    """---
timeout: 101
repeat_count: 99
""",  # missing testsuites
    """---
- name: blah
- name: blip
""",  # missing fetch options
    """---
testsuites:
  - name: blah
  - name: blip
""",  # missing fetch options
    """---
testsuites:
""",  # no testsuites
]


class TestJSONSchema(unittest.TestCase):

    """Test schema validation."""

    def test_multi_schema(self):
        """Test that the new master.run validates correctly."""
        data_new = yaml.load(yaml_content)
        print("data_new: {}".format(data_new))
        jsonschema.validate(data_new, Runner.MASTER_RUNLIST_SCHEMA)

    def test_example_runlist(self):
        """Test that the example runlist validates correctly."""
        example_runlist = os.path.join(os.path.dirname(__file__),
                                       "../examples/master.run")

        self.assertTrue(os.path.exists(example_runlist))

        with open(example_runlist, "r") as fp:
            data = yaml.load(fp)
            print("data: {}".format(data))
            jsonschema.validate(data, Runner.MASTER_RUNLIST_SCHEMA)
            fp.close()

    def test_include_schema(self):
        """Test that include works correctly."""
        data = yaml.load(yaml_content_include)
        print("data: {}".format(data))
        jsonschema.validate(data, Runner.MASTER_RUNLIST_SCHEMA)

    def test_bad_schemas(self):
        """Test that required fields are enforced by the schema."""
        for content in yaml_content_bad:
            data = yaml.load(content)
            print("data: {}".format(data))
            self.assertRaises(jsonschema.ValidationError,
                              jsonschema.validate,
                              data, Runner.MASTER_RUNLIST_SCHEMA)
