# Ubuntu Testing Automation Harness
# Copyright 2012 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test package information classes."""

import functools
import unittest

from mock import patch, Mock

from utah.client.package_info import (
    DeltaPackages,
    InstalledPackages,
)


class TestInstalledPackages(unittest.TestCase):

    """Verify installed packages from the apt cache are properly retrieved."""

    def setUp(self):
        """Mock apt cache to simulate only a few packages are available."""
        self.patcher = patch('utah.client.package_info.apt')
        self.apt = self.patcher.start()

        package_data = [
            ('a', True, '1'),
            ('b', False, '2'),
            ('c', True, '3')]
        mock_data = []
        for name, installed, version in package_data:
            mock_obj = Mock()
            mock_obj.name = name
            mock_obj.is_installed = installed
            mock_obj.installed.version = version
            mock_data.append(mock_obj)
        self.apt.cache.Cache.return_value = mock_data

    def tearDown(self):
        """Undo apt cache mocking."""
        self.patcher.stop()

    def test_only_installed(self):
        """Verify that only installed packages from the cache are included."""
        packages = InstalledPackages()
        self.assertDictEqual({'a': '1', 'c': '3'}, dict(packages))


@functools.total_ordering
class HashedString(object):

    """Helper class to define string objects and their hash code.

    This is useful to manipulate the ordering of keys in dictionaries.

    """

    def __init__(self, string_value, hash_value=None):
        self.string_value = string_value
        self.hash_value = hash_value

    def __repr__(self):
        return repr(self.string_value)

    def __hash__(self):
        return self.hash_value

    def __eq__(self, other):
        if isinstance(other, HashedString):
            return self.string_value == other.string_value
        if isinstance(other, str):
            return self.string_value == other

        return False

    def __lt__(self, other):
        if isinstance(other, HashedString):
            return self.string_value < other.string_value
        if isinstance(other, str):
            return self.string_value < other

        return False


class TestDeltaPackages(unittest.TestCase):

    """Verify delta values are calculated properly."""

    def setUp(self):
        """Create dictionaries that are used as InstalledPacakges objects."""
        self.start = {
            HashedString('a', 2): '1',
            HashedString('b', 1): '1',
            HashedString('c', 2): '1',
            'd': '1',
            HashedString('e', 1): '1',
        }
        self.end = {
            HashedString('c', 2): '2',
            'd': '1',
            HashedString('e', 1): '2',
            HashedString('f', 2): '1',
            HashedString('g', 1): '1',
        }

    def test_installed(self):
        """Verify that installed packages are detected."""
        delta = DeltaPackages(self.start, self.end)
        self.assertListEqual(delta.installed,
                             ['f', 'g'])

    def test_uninstalled(self):
        """Verify that uninstalled packages are detected."""
        delta = DeltaPackages(self.start, self.end)
        self.assertListEqual(delta.uninstalled,
                             ['a', 'b'])

    def test_updated(self):
        """Verify that updated packages are detected."""
        delta = DeltaPackages(self.start, self.end)
        self.assertListEqual(
            delta.updated,
            [{'name': 'c',
              'old_version': '1',
              'new_version': '2'},
             {'name': 'e',
              'old_version': '1',
              'new_version': '2'}])

    def test_to_dict(self):
        """Verify it can be easily converted to a dictionary."""
        delta = DeltaPackages(self.start, self.end)
        self.assertDictEqual(
            dict(delta),
            {'uninstalled': ['a', 'b'],
             'installed': ['f', 'g'],
             'updated': [{'name': 'c',
                          'old_version': '1',
                          'new_version': '2'},
                         {'name': 'e',
                          'old_version': '1',
                          'new_version': '2'}]
             })
