# Ubuntu Testing Automation Harness
# Copyright 2012 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test phoenix metadata generation tool."""


import os
import shutil
import unittest

from utah.client.phoenix import Phoenix

directory_name = "/tmp/phoenix"


def setUp():
    """Set up test prerequisite."""
    if os.path.exists(directory_name):
        raise Exception("directory '{}' already exists".format(directory_name))


def tearDown():
    """Remove test resources."""
    if os.path.exists(directory_name):
        shutil.rmtree(directory_name)


class TestPhoenix(unittest.TestCase):

    """Test phoenix test suite creation utility."""

    testsuite_name = "testsuite_name"
    testcases = ['test_one', 'test_two']
    directory_name = directory_name

    def setUp(self):
        """Set up test suite creation information."""
        self.phoenix = Phoenix(
            self.testsuite_name,
            directory=self.directory_name,
            testcases=self.testcases,
        )

    def test_missing_required_args(self):
        """Test that __init__ fails without the required arguments."""
        self.assertRaises(
            TypeError,
            Phoenix,
        )

    def test_providing_required_args(self):
        """Test that __init__ succeeds with the required arguments."""
        self.assertTrue(self.phoenix is not None)

    def test_suite_name(self):
        """Test that the testsuite name is correctly set."""
        self.assertEqual(self.phoenix.testsuite, self.testsuite_name)

    def test_directory_name(self):
        """Test that the directory name is correctly set."""
        self.assertEqual(self.phoenix.directory, self.directory_name)

    def test_default_directory_name(self):
        """Test that the default directory name is correctly set."""
        phoenix = Phoenix(self.testsuite_name)

        self.assertEqual(phoenix.directory, '.')

    def test_testcases(self):
        """Test that the testcases are correctly set."""
        self.assertListEqual(self.phoenix.testcases, self.testcases)

    def test_testcases_default(self):
        """Test that the default testcase is correctly set."""

        phoenix = Phoenix(self.testsuite_name)
        # there should be only one default testcase.
        self.assertEqual(len(phoenix.testcases), 1)

    def test_build_suite(self):
        """Test building a test suite."""
        self.phoenix.build_suite()

        dir_path = self.directory_name
        ts_path = os.path.join(dir_path, self.testsuite_name)
        tc_paths = []
        for tc in self.testcases:
            tc_paths.append(os.path.join(ts_path, tc))

        self.assertTrue(os.path.exists(dir_path),
                        "{} doesn't exist".format(dir_path))
        self.assertTrue(os.path.exists(ts_path),
                        "{} doesn't exist".format(ts_path))

        for path in tc_paths:
            self.assertTrue(os.path.exists(path),
                            "{} doesn't exist".format(path))

    def test_create_file(self):
        """Test that Phoenix create file works correctly."""
        fn = '/tmp/phoenix_create_file_test.file'
        fc = 'a test file'

        # file shouldn't be there to start
        self.assertFalse(os.path.exists(fn))

        self.phoenix.create_file(fn, fc)

        self.assertTrue(os.path.exists(fn))

        with open(fn, 'r') as fp:
            fc_read = fp.read()
        self.assertEqual(fc, fc_read)

        os.remove(fn)
