# Ubuntu Testing Automation Harness
# Copyright 2012 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test our probe interface."""

import os
import shutil
import tempfile
import time
import unittest

from utah.client.probe import probes


class TestProbes(unittest.TestCase):

    """Basic testing of the probes api."""

    def setUp(self):
        """Create temporary directory for probe output."""
        probes.output_dir = tempfile.mkdtemp()

    def tearDown(self):
        """Cleanup artifacts from setUp."""
        shutil.rmtree(probes.output_dir)

    def test_needs_config(self):
        """Ensure the probes api requires configuration."""
        probes._enabled = None
        with self.assertRaises(RuntimeError):
            with probes.run({}):
                pass

    def test_run_no_probes(self):
        """Ensure testing with nothing enabled works."""
        data = {}
        probes.enable_probes([])
        with probes.run(data):
            pass
        self.assertEqual(0, len(data))

    def test_run_battery(self):
        """Ensure testing with just battery works."""
        data = {}
        probes.enable_probes(['battery'])
        with probes.run(data):
            pass
        self.assertTrue('battery' in data)
        self.assertTrue('start' in data['battery'])
        self.assertTrue('end' in data['battery'])

    @unittest.skipUnless(os.path.exists('/usr/bin/eventstat'), 'need evenstat')
    def test_run_eventstat(self):
        """Ensure testing with just eventstat works."""
        data = {}
        probes.enable_probes(['eventstat'])
        with probes.run(data):
            time.sleep(0.5)
        self.assertIn('eventstat', data)
        self.assertIn('results', data['eventstat'])
        self.assertTrue(os.path.exists(data['eventstat']['results']))

    def test_run_contextswitch(self):
        """Ensure contextswitch probe works."""
        data = {}
        probes.enable_probes(['contextswitch'])
        with probes.run(data):
            time.sleep(0.5)
        self.assertIn('contextswitch', data)
        self.assertIn('total', data['contextswitch'])
        self.assertIn('duration', data['contextswitch'])
        self.assertIn('rate', data['contextswitch'])

    def test_run_interrupts(self):
        """Ensure interrupt probe works."""
        data = {}
        probes.enable_probes(['interrupts'])
        with probes.run(data):
            time.sleep(0.5)
        self.assertIn('interrupts', data)
        self.assertIn('total', data['interrupts'])
        self.assertIn('duration', data['interrupts'])
        self.assertIn('rate', data['interrupts'])
        self.assertIn('interrupts_start', data['interrupts'])
        self.assertIn('interrupts_end', data['interrupts'])
        self.assertTrue(os.path.exists(data['interrupts']['interrupts_start']))
        self.assertTrue(os.path.exists(data['interrupts']['interrupts_end']))

    @unittest.skipUnless(os.path.exists('/usr/bin/eventstat'), 'need evenstat')
    def test_run_multiple(self):
        """Ensure we can run more than one probe."""
        data = {}
        probes.enable_probes(['battery', 'eventstat'])
        with probes.run(data):
            time.sleep(0.5)
        self.assertIn('eventstat', data)
        self.assertIn('results', data['eventstat'])
        self.assertTrue(os.path.exists(data['eventstat']['results']))

        self.assertTrue('battery' in data)
        self.assertTrue('start' in data['battery'])
        self.assertTrue('end' in data['battery'])
