# Ubuntu Testing Automation Harness
# Copyright 2012 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test our ability to report results."""


import unittest
import yaml

from utah.client.result import (
    Result,
    ResultYAML,
)


class TestResult(unittest.TestCase):

    """Test result processing and display functionality."""

    def setUp(self):
        """Create needed data structures."""
        self.result_yaml = ResultYAML(None, False, "result_yaml")
        self.result_text = Result(None, False, "result")

        self.result = {
            'command': 'echo hi',
            'returncode': 0,
            'stdout': 'hi',
            'stderr': '',
            'start_time': '20121107 12:00:00',
        }

        self.result_fail = {
            'command': 'echo failure',
            'returncode': 2,
            'stdout': 'hi',
            'stderr': '',
            'start_time': '20121107 12:00:00',
        }

    def test_add_result(self):
        """Test adding test results."""
        self.result_yaml.add_result(self.result)

        self.assertEqual(self.result_yaml._count_results(), 1)

        self.result_yaml.result()

    def test_result(self):
        """Test generating results."""
        self.result_yaml.add_result(self.result)

        # check that the result was added.
        self.assertEqual(self.result_yaml._count_results(), 1)

        self.result_yaml.result()

        # check that the result was processed and removed.
        self.assertEqual(self.result_yaml._count_results(), 0)

    def test_result_pass(self):
        """Test that adding a pass result sets the status to 'PASS'."""
        self.result_yaml.add_result(self.result)

        self.assertEqual(self.result_yaml.status, 'PASS')

    def test_result_failure(self):
        """Test that adding a failure result sets the status to 'FAIL'."""
        self.result_yaml.add_result(self.result_fail)

        self.assertEqual(self.result_yaml.status, 'FAIL')

    def test_result_clear_result(self):
        """Test that the result is actually cleared."""
        self.result_yaml.add_result(self.result_fail)
        self.assertEqual(self.result_yaml.status, 'FAIL')

        self.result_yaml._clear_results()

        self.assertEqual(self.result_yaml.status, 'PASS')
        self.assertEqual(self.result_yaml._count_results(), 0)

    def test_result_result(self):
        """Test that ERROR status is retained after adding pass results."""
        self.result_yaml.add_result(self.result_fail)

        self.assertEqual(self.result_yaml.status, 'FAIL')

        self.result_yaml.add_result(self.result)

        self.assertEqual(self.result_yaml.status, 'FAIL')

    def test_result_yaml(self):
        """Test generating YAML results."""
        self.result_yaml.add_result(self.result)

        import sys
        import StringIO

        str_output = StringIO.StringIO()

        # redirect output to a string
        old_stdout = sys.stdout
        sys.stdout = str_output

        self.result_yaml.result()

        sys.stdout = old_stdout

        result_str = str_output.getvalue()

        print('result_str: {}'.format(result_str))

        data = yaml.load(result_str)

        self.assertEqual(len(data['commands']), 1)

        result = data['commands'][0]

        print('data: {}'.format(data))

        self.assertEqual(result['command'], self.result['command'])
        self.assertEqual(result['returncode'], self.result['returncode'])
        self.assertEqual(result['stdout'], self.result['stdout'])
        self.assertEqual(result['stderr'], self.result['stderr'])
