# Ubuntu Testing Automation Harness
# Copyright 2012 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Who tests the test runners?"""


import os
import shutil
import tempfile
import unittest
from mock import patch

import jsonschema

from utah.client.result import ResultYAML
from utah.client.state_agent import StateAgentYAML
from utah.client.runner import Runner
from utah.client.common import (
    DefaultValidator,
    ReturnCodes,
    CMD_TC_CLEANUP,
    CMD_TC_TEST,
    CMD_TC_SETUP,
    CMD_TS_FETCH,
)
from utah.client import exceptions

from utah.client.tests.common import (  # NOQA
    setUp,  # Used by nosetests
    tearDown,  # Used by nosetests
    master_runlist_content,
    get_module_path,
)


class TestRunner(unittest.TestCase):

    """Test the TestRunner class."""

    def setUp(self):
        """Set up test prerequisites."""
        self.state_file = '/tmp/state.yaml'
        self.state_agent = StateAgentYAML(state_file=self.state_file)
        self.output_file = '/tmp/utah.out'
        self.runner = Runner(install_type='desktop',
                             result=ResultYAML(filename=None,
                                               append_to_file=False),
                             state_agent=self.state_agent,
                             output=self.output_file)
        self.bad_dir = "/tmp/does_not_exist"
        _, self.tmpfile = tempfile.mkstemp()

        self.assertFalse(os.path.exists(self.bad_dir),
                         "{} shouldn't exist".format(self.bad_dir))

    def tearDown(self):
        """Clean up test resources."""
        os.unlink(self.tmpfile)
        if os.path.exists(self.bad_dir):
            shutil.rmtree(self.bad_dir)

    def test_fetch_failed(self):
        """Test handling of failed fetch commands."""
        bad_runlist_content = """---
testsuites:
  - name: fake_tests
    fetch_method: bzr
    fetch_location: /dev/null
"""
        runlist = '/tmp/master_fetch_fails.run'

        with open(runlist, 'w') as fp:
            fp.write(bad_runlist_content)

        runner = Runner(install_type='desktop',
                        result=ResultYAML(filename=None,
                                          append_to_file=False),
                        state_agent=self.state_agent,
                        runlist=runlist)

        os.remove(runlist)

        runner.run()

        self.assertEqual(runner.fetch_errors, 1)

    def test_rc_local(self):
        """Test rc.local functionality for reboots."""
        tmp_rc_local = '/tmp/rc.local'
        self.runner.setup_rc_local(tmp_rc_local)

        self.assertTrue(os.path.exists(tmp_rc_local))

    def test_rc_local_contents(self):
        """Test rc.local output setting."""
        tmp_rc_local = '/tmp/rc.local'
        self.runner.setup_rc_local(tmp_rc_local)

        with open(tmp_rc_local) as rc_local:
            self.assertIn(self.output_file, rc_local.read(),
                          'Output directory not written to rc.local')

    def test_run(self):
        """Test running all test suites."""
        self.runner.run()

        for suite in self.runner.suites:
            self.assertTrue(suite.is_done())

        # ensure the proper fetch stuff happens
        results = self.runner.result.results
        for i in xrange(len(self.runner.suites) * 2):
            self.assertEqual(results[i]['cmd_type'], CMD_TS_FETCH)
            if i % 2:
                self.assertTrue(results[i]['command'].startswith('bzr revno'))
            else:
                self.assertTrue(results[i]['command'].startswith('bzr export'))

        retcode = self.runner.process_results()
        self.assertEqual(retcode, ReturnCodes.FAIL)

    def test_rerun(self):
        """Ensure rerun logic works at the runner level."""
        self.runner.repeat_count = 1
        self.runner.run()
        results = self.runner.result.results
        offset = len(results)
        self.runner.run(1)
        i = offset
        for suite in self.runner.suites:
            for test in suite.tests:
                self.assertEqual(results[i + 0]['testcase'], test.name)
                self.assertEqual(results[i + 0]['cmd_type'], CMD_TC_SETUP)
                self.assertEqual(results[i + 1]['cmd_type'], CMD_TC_TEST)
                self.assertEqual(results[i + 2]['cmd_type'], CMD_TC_CLEANUP)
                i += 3
        self.assertEqual(len(results), i)

    def test_missing_testdir(self):
        """Test that the runner fails gracefully on missing testdir."""
        self.assertRaises(exceptions.BadDir,
                          Runner,
                          install_type="desktop",
                          result=ResultYAML(filename=None,
                                            append_to_file=False),
                          testdir=self.bad_dir)

    def test_missing_master_runlist(self):
        """Test that the runner fails gracefully on missing 'master.run'."""
        tmp_master_runlist = os.path.join(self.bad_dir, 'master.run')

        if not os.path.exists(self.bad_dir):
            os.mkdir(self.bad_dir)

        # Make sure there isn't a master runlist
        if os.path.exists(tmp_master_runlist):
            os.remove(tmp_master_runlist)

        self.assertRaises(exceptions.MissingFile,
                          Runner,
                          install_type="desktop",
                          result=ResultYAML(filename=None,
                                            append_to_file=False),
                          runlist=tmp_master_runlist,
                          testdir=self.bad_dir)

    def test_runlist(self):
        """Test that passing a runlist to the Runner works properly."""
        runlist = '/tmp/master_runlist_test'

        with open(runlist, 'w') as fp:
            fp.write(master_runlist_content)

        runner = Runner(install_type='desktop',
                        result=ResultYAML(filename=None,
                                          append_to_file=False),
                        runlist=runlist)
        runner.save_state()

        print("tests: {}".format(runner.count_tests()))

        self.assertTrue(runner.count_tests() > 0)

        os.remove(runlist)

    def test_testdir(self):
        """Test that the default test directory is created."""
        testdir = self.runner.testdir

        self.assertTrue(os.path.exists(testdir))

    @patch('utah.client.runner.TestSuite')
    def test_include_exclude(self, mock_class):
        """Test that includes/excludes are passed to the test suite."""
        include_runlist_content = """---
testsuites:
  - name: utah_tests
    fetch_method: bzr-export
    fetch_location: {REPO_PATH}/client/examples/utah_tests
    include_tests:
      - test_a
      - test_b
      - test_c
    exclude_tests:
      - test_1
      - test_2
      - test_3
""".format(REPO_PATH=get_module_path())

        runlist = '/tmp/master_runlist_test'
        with open(runlist, 'w') as fp:
            fp.write(include_runlist_content)

        Runner(install_type='desktop',
               result=ResultYAML(filename=None,
                                 append_to_file=False),
               runlist=runlist)

        args, kwargs = mock_class.call_args
        self.assertListEqual(args[3],
                             ['test_a', 'test_b', 'test_c'])
        self.assertListEqual(args[4],
                             ['test_1', 'test_2', 'test_3'])

    def test_fetch_runlist(self):
        """Ensure proper error handling of runlist fetching logic."""
        with self.assertRaises(exceptions.MissingFile):
            Runner._fetch_and_parse('this is a bad url')

        with self.assertRaises(exceptions.ValidationError):
            with open(self.tmpfile, 'w') as f:
                f.write('andy: blah\n')
            Runner._fetch_and_parse(self.tmpfile)

    def test_reset_rc(self):
        """Test reset_rc_local is only called when the argument is True."""
        with patch('utah.client.runner.Runner.reset_rc_local') as reset:
            Runner(install_type='desktop',
                   result=ResultYAML(filename=None,
                                     append_to_file=False),
                   state_agent=self.state_agent,
                   output=self.output_file,
                   reset_rc=False)
            assert not reset.called
            Runner(install_type='desktop',
                   result=ResultYAML(filename=None,
                                     append_to_file=False),
                   state_agent=self.state_agent,
                   output=self.output_file,
                   reset_rc=True)
            assert reset.called


class TestRunnerMasterRunlistSchema(unittest.TestCase):

    """Provide master runlist schema test cases."""

    def setUp(self):
        """Setup needed resources."""
        schema = Runner.MASTER_RUNLIST_SCHEMA
        DefaultValidator.check_schema(schema)
        self.validator = DefaultValidator(schema)

    def validate(self, data):
        """Validate a schema.

        :returns: validates schema
        :rtype: object

        """
        return self.validator.validate(data)

    def test_empty_string_invalid(self):
        """Verify an empty string doesn't validate."""
        with self.assertRaises(jsonschema.ValidationError):
            self.validate('')

    def test_empty_dict_invalid(self):
        """Verify an empty dictionary is invalid."""
        with self.assertRaises(jsonschema.ValidationError):
            self.validate({})

    def test_valid(self):
        """Verify valid data works."""
        self.validate({
            'testsuites': [{
                'name': 'name',
                'fetch_method': 'dev',
                'fetch_location': 'location',
            }]
        })

    def test_testsuites_required(self):
        """Verify a list of test suites is required."""
        with self.assertRaises(jsonschema.ValidationError):
            self.validate({
                'battery_measurements': True,
            })

    def test_testsuites_empty_invalid(self):
        """Verify an empty list of test suites is invalid."""
        with self.assertRaises(jsonschema.ValidationError):
            self.validate({
                'testsuites': [],
            })

    def test_testsuite_name_required(self):
        """Verify test suite name is required."""
        with self.assertRaises(jsonschema.ValidationError):
            self.validate({
                'testsuites': [{
                    'fetch_method': 'dev',
                    'fetch_location': 'fetch location',
                }],
            })

    def test_testsuite_fetch_method_required(self):
        """Verify test suite fetch method is required."""
        with self.assertRaises(jsonschema.ValidationError):
            self.validate({
                'testsuites': [{
                    'name': 'name',
                    'fetch_location': 'fetch location',
                }],
            })

    def test_testsuite_fetch_method_known_values(self):
        """Verify test suite fetch method known values are valid."""
        testsuite = {
            'name': 'name',
            'fetch_location': 'fetch location',
        }

        data = {'testsuites': [testsuite]}

        for fetch_method in ['dev', 'bzr', 'bzr-export', 'git']:
            testsuite['fetch_method'] = fetch_method
            self.validate(data)

    def test_testsuite_fetch_method_unknown_value(self):
        """Verify test suite fetch method unknown values are valid."""
        with self.assertRaises(jsonschema.ValidationError):
            self.validate({
                'testsuites': [{
                    'name': 'name',
                    'fetch_method': 'unknown',
                    'fetch_location': 'fetch location',
                }],
            })

    def test_testsuite_fetch_location_required(self):
        """Verify test suite fetch location is required."""
        with self.assertRaises(jsonschema.ValidationError):
            self.validate({
                'testsuites': [{
                    'name': 'name',
                    'fetch_method': 'dev',
                }],
            })

    def test_testsuite_include_tests_empty_invalid(self):
        """Verify an empty list of included tests is invalid."""
        with self.assertRaises(jsonschema.ValidationError):
            self.validate({
                'testsuites': [{
                    'name': 'name',
                    'fetch_method': 'dev',
                    'fetch_location': 'fetch location',
                    'include_tests': [],
                }],
            })

    def test_testsuite_include_tests_string(self):
        """Verify a list of strings as included tests is valid."""
        self.validate({
            'testsuites': [{
                'name': 'name',
                'fetch_method': 'dev',
                'fetch_location': 'fetch location',
                'include_tests': ['test_1', 'test_2', 'test_3'],
            }],
        })

    def test_testsuite_exclude_tests_empty_invalid(self):
        """Verify an empty list of excluded tests is invalid."""
        with self.assertRaises(jsonschema.ValidationError):
            self.validate({
                'testsuites': [{
                    'name': 'name',
                    'fetch_method': 'dev',
                    'fetch_location': 'fetch location',
                    'exclude_tests': [],
                }],
            })

    def test_testsuite_exclude_tests_string(self):
        """Verify a list of strings as excluded tests is valid."""
        self.validate({
            'testsuites': [{
                'name': 'name',
                'fetch_method': 'dev',
                'fetch_location': 'fetch location',
                'exclude_tests': ['test_1', 'test_2', 'test_3'],
            }],
        })

    def test_battery_measurements_false_by_default(self):
        """Verify battery measurements is false by default."""
        data = {
            'testsuites': [{
                'name': 'name',
                'fetch_method': 'dev',
                'fetch_location': 'fetch location',
            }],
        }
        self.validate(data)
        self.assertFalse(data['battery_measurements'])

    def test_battery_measurements_string_invalid(self):
        """Verify battery measurements string value is invalid."""
        with self.assertRaises(jsonschema.ValidationError):
            self.validate({
                'testsuites': [{
                    'name': 'name',
                    'fetch_method': 'dev',
                    'fetch_location': 'fetch location',
                }],
                'battery_measurements': 'value',
            })

    def test_battery_measurements_boolean_valid(self):
        """Verify a battery measurements boolean value is valid."""
        self.validate({
            'testsuites': [{
                'name': 'name',
                'fetch_method': 'dev',
                'fetch_location': 'fetch location',
            }],
            'battery_measurements': True,
        })

    def test_zero_timeout_valid(self):
        """Verify a zero timeout is valid."""
        self.validate({
            'testsuites': [{
                'name': 'name',
                'fetch_method': 'dev',
                'fetch_location': 'fetch location',
            }],
            'timeout': 0,
        })

    def test_negative_timeout_invalid(self):
        """Verify a negative timeout is invalid."""
        with self.assertRaises(jsonschema.ValidationError):
            self.validate({
                'testsuites': [{
                    'name': 'name',
                    'fetch_method': 'dev',
                    'fetch_location': 'fetch location',
                }],
                'timeout': -1,
            })

    def test_repeat_count_zero_by_default(self):
        """Verify repeat count is zero by default."""
        data = {
            'testsuites': [{
                'name': 'name',
                'fetch_method': 'dev',
                'fetch_location': 'fetch location',
            }],
        }

        self.validate(data)
        self.assertEqual(data['repeat_count'], 0)

    def test_negative_repeat_count_invalid(self):
        """Verift. negative repeat count is invalid."""
        with self.assertRaises(jsonschema.ValidationError):
            self.validate({
                'testsuites': [{
                    'name': 'name',
                    'fetch_method': 'dev',
                    'fetch_location': 'fetch location',
                }],
                'repeat_count': -1,
            })

    def test_probe_default(self):
        """Verify probes default to nothing."""
        data = {
            'testsuites': [{
                'name': 'name',
                'fetch_method': 'dev',
                'fetch_location': 'fetch location',
            }],
        }

        self.validate(data)
        self.assertEqual(0, len(data['probes']))

    def test_probe_invalid(self):
        """Verify bad probe name is rejected."""
        with self.assertRaises(jsonschema.ValidationError):
            self.validate({
                'testsuites': [{
                    'name': 'name',
                    'fetch_method': 'dev',
                    'fetch_location': 'fetch location',
                }],
                'probes': ['this is not a probe'],
            })

    def test_probe_battery(self):
        """Verify probes work with a good value."""
        data = {
            'testsuites': [{
                'name': 'name',
                'fetch_method': 'dev',
                'fetch_location': 'fetch location',
            }],
            'probes': ['battery'],
        }

        self.validate(data)
        self.assertEqual(1, len(data['probes']))
        self.assertEqual(data['probes'][0], 'battery')
