# Ubuntu Testing Automation Harness
# Copyright 2012 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test our ability to save and restore state information."""


import os
import unittest
import yaml

from utah.client.state_agent import StateAgentYAML
from utah.client.runner import Runner
from utah.client.result import ResultYAML

from utah.client.tests.common import (  # NOQA
    setUp,  # Used by nosetests
    tearDown,  # Used by nosetests
    partial_state_file_content,
    partial_state_file_content_run_all,
    partial_state_file_content_done_all_failed,
)


class TestStateAgentYAML(unittest.TestCase):

    """Test the YAML state agent."""

    def setUp(self):
        """Create a state agent for testing."""
        self.state_file = '/tmp/state.yaml'

        if os.path.exists(self.state_file):
            os.unlink(self.state_file)

        self.state_agent = StateAgentYAML(state_file=self.state_file)
        self.runner = Runner(
            install_type="desktop",
            result=ResultYAML(filename=None, append_to_file=False),
            state_agent=self.state_agent)

    def tearDown(self):
        """Clean up the test state file."""
        if os.path.exists(self.state_file):
            os.remove(self.state_file)

    def test_creation(self):
        """Test that state file is correctly created."""
        self.runner.run()

        self.assertTrue(os.path.exists(self.state_file))

    def test_cleanup(self):
        """Test that cleanup removes state file."""
        self.runner.save_state()

        self.assertTrue(os.path.exists(self.state_file),
                        "Runner failed to save state")

        self.state_agent.clean()

        self.assertFalse(os.path.exists(self.state_file),
                         'Failed to remove state file ({})'
                         .format(self.state_file))

    def test_status(self):
        """Test that a runner properly saves its state during test runs."""
        suite_count = self.runner.count_suites()
        test_count = self.runner.count_tests()

        self.runner.run()

        self.assertTrue(os.path.exists(self.state_file))

        self.runner.load_state()

        self.assertEqual(self.runner.count_suites(), suite_count)
        self.assertEqual(self.runner.count_tests(), test_count)

        with open(self.state_file, 'r') as f:
            state_data = yaml.load(f)

        self.assertEqual(state_data['status'], 'DONE')

    def _test_partial(self, content, next_test, passes, fails, errors):
        with open(self.state_file, 'w') as f:
            f.write(content)

        # we must call the common setUp again since our object's setUp()
        # creates a Runner object which will delete everything under
        # /var/lib/utah/testsuites
        setUp()
        state_agent = StateAgentYAML(state_file=self.state_file)
        runner = Runner(install_type="desktop",
                        state_agent=state_agent,
                        result=ResultYAML(filename=None,
                                          append_to_file=False),
                        resume=True)

        self.assertEqual(runner.status, 'RUN')
        self.assertEqual(runner.count_suites(), 3)
        self.assertEqual(runner.count_tests(), 5)

        if next_test:
            next_test_obj = runner.get_next_test()
            self.assertEqual(next_test_obj.name, next_test)

        runner.run()

        self.assertEqual(runner.passes, passes)
        self.assertEqual(runner.failures, fails)
        self.assertEqual(runner.errors, errors)
        t = passes + fails + errors
        self.assertEqual(runner.passes + runner.failures + runner.errors, t)

    def test_load_state_partial(self):
        """Test that a partially run state file can be resumed."""
        self._test_partial(partial_state_file_content, 'test_two', 4, 1, 0)

    def test_load_state_partial_run_all(self):
        """Test partially run state file - run all jobs."""
        self._test_partial(
            partial_state_file_content_run_all, 'test_one', 3, 2, 0)

    def test_load_state_partial_done_all_failed(self):
        """Test partially run state file - all jobs failed."""
        self._test_partial(
            partial_state_file_content_done_all_failed, None, 0, 5, 0)
