# Ubuntu Testing Automation Harness
# Copyright 2012 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test handling of test cases."""


import os
import shutil
import tempfile
import unittest

import mock
import jsonschema

from utah.client import testcase
from utah.client.common import (
    DefaultValidator,
    CMD_TC_CLEANUP,
    CMD_TC_SETUP,
    CMD_TC_TEST,
)
from utah.client.result import ResultYAML
from utah.client.tests.common import (  # NOQA
    setUp,  # Used by nosetests
    tearDown,  # Used by nosetests
)
import utah.client.testsuite as testsuite


class TestTestCase(unittest.TestCase):

    """Tests for the TestCase class."""

    def setUp(self):
        """Initialize basic test case data."""

        # basically construct an empty testsuite that's needed to drive
        # our testcase objects
        self.runner = mock.Mock()
        self.runner.result = ResultYAML(filename=None, append_to_file=False)
        self.runner.testsuitedir = tempfile.mkdtemp()
        self.runner.timeout = None
        client_dir = os.path.dirname(os.path.dirname(__file__))
        examples = os.path.join(client_dir, 'examples')
        dst = os.path.join(self.runner.testsuitedir, 'examples')
        shutil.copytree(examples, dst)
        self.suite = testsuite.TestSuite('examples', self.runner,
                                         excludes=['test_one', 'test_two'])

        self.name = "test_one"
        self.command = 'echo "command"'
        self.case = testcase.TestCase('test_one', self.suite, self.command)

    def tearDown(self):
        """Remove tempfiles created by test."""
        shutil.rmtree(self.runner.testsuitedir)

    def test_run(self):
        """Test that a TestCase can be run."""
        curdir = os.getcwd()
        self.case.run()

        self.assertEqual(os.getcwd(), curdir)
        self.assertTrue(self.case.is_done())

    def test_rerun(self):
        """Test that a TestCase can be run."""
        curdir = os.getcwd()
        self.case.run()

        # make sure after a testcase run that the cwd hasn't changed.
        self.assertEqual(os.getcwd(), curdir)
        self.assertTrue(self.case.is_done())

        for x in xrange(3):
            self.case.run(rerun=True)
            results = self.case.suite.runner.result.results
            self.assertEqual(len(results), 6 + (3 * x))

            # ensure the last rerun's results do a setup, test, and cleanup
            self.assertEqual(results[(3 * x)]['cmd_type'], CMD_TC_SETUP)
            self.assertEqual(results[(3 * x) + 1]['cmd_type'], CMD_TC_TEST)
            self.assertEqual(results[(3 * x) + 2]['cmd_type'], CMD_TC_CLEANUP)

    def test_setup_fail(self):
        """Ensure a failed setup works properly."""
        self.case.tc_setup = '/bin/false'
        self.assertEqual(self.case.run(), 'PASS')
        self.assertEqual(self.case.summary, 'ERROR')
        self.assertEqual(self.case.status, 'SETUP')

    def test_build_fail(self):
        """Ensure a failed build works properly."""
        self.case.build_cmd = '/bin/false'
        self.assertEqual(self.case.run(), 'PASS')
        self.assertEqual(self.case.summary, 'ERROR')
        self.assertEqual(self.case.status, 'BUILD')

    def test_timeout(self):
        """Test that timeouts are parsed correctly."""
        self.assertEqual(self.case.timeout, 0)

    def test_command(self):
        """Test that commands are parsed correctly."""
        self.assertEqual(self.case.command, self.command)

    def test_save_state(self):
        """Test that state can be saved."""
        state = self.case.save_state()

        self.assertEqual(state['status'], 'NOTRUN')
        self.assertEqual(state['command'], self.command)
        self.assertEqual(state['timeout'], 0)

    def test_save_and_load_state(self):
        """Test that state can be saved and loaded."""
        path = self.case.path
        state = self.case.save_state()
        case = testcase.TestCase(self.name, self.suite)
        case.load_state(state)

        self.assertFalse(case.is_done())

        self.assertEqual(case.name, self.name)
        self.assertEqual(case.path, path)
        self.assertEqual(case.command, self.command)
        self.assertEqual(case.timeout, 0)

        case.run()

        self.assertTrue(case.is_done())


class TestTestCaseControlSchema(unittest.TestCase):

    """Test case control schema test cases."""

    def setUp(self):
        """Setup needed resources."""
        schema = testcase.TestCase.CONTROL_SCHEMA
        DefaultValidator.check_schema(schema)
        self.validator = DefaultValidator(schema)

    def validate(self, data):
        """Validate a schema.

        :returns: validates schema
        :rtype: object

        """
        schema = testcase.TestCase.CONTROL_SCHEMA
        return jsonschema.validate(data, schema)

    def test_empty_string_invalid(self):
        """An empty string doesn't validate."""
        with self.assertRaises(jsonschema.ValidationError):
            self.validate('')

    def test_empty_dict_invalid(self):
        """An empty dictionary is valid."""
        with self.assertRaises(jsonschema.ValidationError):
            self.validate({})

    def test_valid(self):
        """Verify valid data works."""
        self.validate({
            'description': 'description',
            'dependencies': 'dependencies',
            'action': 'action',
            'expected_results': 'expected_results',
            'command': 'command',
            'run_as': 'run_as',
        })

    def test_description_required(self):
        """Verify description is a required property."""
        with self.assertRaises(jsonschema.ValidationError):
            self.validate({
                'dependencies': 'dependencies',
                'action': 'action',
                'expected_results': 'expected_results',
                'command': 'command',
                'run_as': 'run_as',
            })

    def test_dependencies_required(self):
        """Verify dependencies is a required property."""
        with self.assertRaises(jsonschema.ValidationError):
            self.validate({
                'description': 'description',
                'action': 'action',
                'expected_results': 'expected_results',
                'command': 'command',
                'run_as': 'run_as',
            })

    def test_action_required(self):
        """Verify action is a required property."""
        with self.assertRaises(jsonschema.ValidationError):
            self.validate({
                'description': 'description',
                'dependencies': 'dependencies',
                'expected_results': 'expected_results',
                'command': 'command',
                'run_as': 'run_as',
            })

    def test_expected_results_required(self):
        """Verify expected results is a required property."""
        with self.assertRaises(jsonschema.ValidationError):
            self.validate({
                'description': 'description',
                'dependencies': 'dependencies',
                'action': 'action',
                'command': 'command',
                'run_as': 'run_as',
            })

    def test_command_required(self):
        """Verify command is a required property."""
        with self.assertRaises(jsonschema.ValidationError):
            self.validate({
                'description': 'description',
                'dependencies': 'dependencies',
                'action': 'action',
                'expected_results': 'expected_results',
                'run_as': 'run_as',
            })

    def test_zero_timeout_valid(self):
        """Verify a zero timeout is valid."""
        self.validate({
            'description': 'description',
            'dependencies': 'dependencies',
            'action': 'action',
            'expected_results': 'expected_results',
            'command': 'command',
            'run_as': 'run_as',
            'timeout': 0,
        })

    def test_negative_timeout_invalid(self):
        """Verify a negative timeout is invalid."""
        with self.assertRaises(jsonschema.ValidationError):
            self.validate({
                'description': 'description',
                'dependencies': 'dependencies',
                'action': 'action',
                'expected_results': 'expected_results',
                'command': 'command',
                'run_as': 'run_as',
                'timeout': -1
            })

    def test_type_userland_valid(self):
        """Verify userland type is valid."""
        self.validate({
            'description': 'description',
            'dependencies': 'dependencies',
            'action': 'action',
            'expected_results': 'expected_results',
            'command': 'command',
            'run_as': 'run_as',
            'type': 'userland',
        })

    def test_type_unknown_invalid(self):
        """Verify unknown type is valid."""
        with self.assertRaises(jsonschema.ValidationError):
            self.validate({
                'description': 'description',
                'dependencies': 'dependencies',
                'action': 'action',
                'expected_results': 'expected_results',
                'command': 'command',
                'run_as': 'run_as',
                'type': 'unknown',
            })

    def test_reboot_known_valid(self):
        """Verify reboot known values are valid."""
        data = {
            'description': 'description',
            'dependencies': 'dependencies',
            'action': 'action',
            'expected_results': 'expected_results',
            'command': 'command',
            'run_as': 'run_as',
        }

        for reboot in ['always', 'pass', 'never']:
            data['reboot'] = reboot
            self.validate(data)

    def test_unknown_reboot_invalid(self):
        """Verify reboot unknown value is invalid."""
        with self.assertRaises(jsonschema.ValidationError):
            self.validate({
                'description': 'description',
                'dependencies': 'dependencies',
                'action': 'action',
                'expected_results': 'expected_results',
                'command': 'command',
                'run_as': 'run_as',
                'reboot': 'unknown',
            })

    def test_reboot_default(self):
        """Reboot defaults to 'never'."""
        data = {
            'description': 'description',
            'dependencies': 'dependencies',
            'action': 'action',
            'expected_results': 'expected_results',
            'command': 'command',
            'run_as': 'run_as',
        }
        self.validator.validate(data)
        self.assertEqual(data['reboot'], 'never')
