# Ubuntu Testing Automation Harness
# Copyright 2012 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test handling of test suites."""

import os
import shutil
import tempfile
import unittest

import jsonschema
import mock

from utah.client import testsuite
from utah.client.common import (
    debug_print,
    DefaultValidator,
    CMD_TC_BUILD,
    CMD_TC_CLEANUP,
    CMD_TC_SETUP,
    CMD_TC_TEST,
    CMD_TS_SETUP,
)
from utah.client.exceptions import ValidationError
from utah.client.result import ResultYAML

from utah.client.tests.common import (  # NOQA
    setUp,  # Used by nosetests
    tearDown,  # Used by nosetests
)


def state_saver():
    """Provide state saving functionality for testing."""
    debug_print("Saving state", force=True)


class TestTestSuite(unittest.TestCase):

    """Tests for the TestSuite class."""

    def setUp(self):
        """Use the 'examples' test suite that's part of the UTAH package."""
        self.name = 'examples'
        self.runner = mock.Mock()
        self.runner.result = ResultYAML(filename=None, append_to_file=False)
        self.runner.testsuitedir = tempfile.mkdtemp()
        self.runner.timeout = None
        client_dir = os.path.dirname(os.path.dirname(__file__))
        examples = os.path.join(client_dir, 'examples')
        dst = os.path.join(self.runner.testsuitedir, 'examples')
        shutil.copytree(examples, dst)

        self.suite = testsuite.TestSuite(self.name, self.runner)

    def tearDown(self):
        """Remove the temporary files we created."""
        shutil.rmtree(self.runner.testsuitedir)

    def test_run(self):
        """Test running a test suite."""
        self.suite.run()
        self.assertEqual(self.suite.runner.result.status, 'PASS')

    def test_rerun(self):
        """Validate the rerun parameter works on a testsuite.run."""
        result = self.suite.runner.result
        self.suite.run()
        self.assertEqual(result.status, 'PASS')

        results = result.results
        offset = len(results)

        self.suite.run(rerun=True)
        self.assertEqual(result.status, 'PASS')

        self.assertEqual(results[1]['testcase'], 'test_one')
        self.assertEqual(results[1]['cmd_type'], CMD_TC_BUILD)
        self.assertEqual(results[2]['cmd_type'], CMD_TC_SETUP)
        self.assertEqual(results[3]['cmd_type'], CMD_TC_TEST)
        self.assertEqual(results[4]['cmd_type'], CMD_TC_CLEANUP)

        self.assertEqual(results[5]['testcase'], 'test_two')
        self.assertEqual(results[5]['cmd_type'], CMD_TC_BUILD)
        self.assertEqual(results[6]['cmd_type'], CMD_TC_SETUP)
        self.assertEqual(results[7]['cmd_type'], CMD_TC_TEST)
        self.assertEqual(results[8]['cmd_type'], CMD_TC_CLEANUP)

        # 3 = each rerun test should have a setup, test, clean
        newresults = len(self.suite.tests) * 3
        self.assertEqual(len(results), newresults + offset)
        self.assertEqual(results[0]['cmd_type'], CMD_TS_SETUP)

        for x in xrange(len(self.suite.tests)):
            name = self.suite.tests[x].name
            runoff = x * 3 + offset
            self.assertEqual(results[runoff + 0]['testcase'], name)
            self.assertEqual(results[runoff + 0]['cmd_type'], CMD_TC_SETUP)
            self.assertEqual(results[runoff + 1]['cmd_type'], CMD_TC_TEST)
            self.assertEqual(results[runoff + 2]['cmd_type'], CMD_TC_CLEANUP)

    def test_count_tests(self):
        """Test ability of test suite to count tests."""
        self.assertEquals(self.suite.count_tests(), 2)

    def test_next_test(self):
        """Test ability of test suite to indicate next test to run."""
        next_test = self.suite.get_next_test()

        self.assertTrue(hasattr(next_test, 'name'))
        self.assertEquals(self.suite.get_next_test().name, 'test_one')

    def test_is_done(self):
        """Test test suite completion reporting."""
        self.assertFalse(self.suite.is_done())

    def test_save_state(self):
        """Test saving test suite state."""
        state = self.suite.save_state()
        self.assertEqual(state['status'], 'NOTRUN')

        tests = 0
        for test in state['tests']:
            self.assertEqual(test['status'], 'NOTRUN')
            tests += 1

        self.assertEqual(tests, self.suite.count_tests())

    def test_overrides(self):
        """Test that overrides from tslist.run are used."""

        self.assertIsNone(self.suite.timeout)
        overrides = {
            'test_one': [3, 'python test_one.py'],
            'test_two': [300, 'python test_two.py'],
        }
        for test in self.suite.tests:
            self.assertEqual(overrides[test.name][0], test.timeout)
            self.assertEqual(overrides[test.name][1], test.command)

    def test_include(self):
        """Verify that included tests are executed."""
        included_test_name = 'test_one'

        suite = testsuite.TestSuite(self.name, self.runner,
                                    includes=[included_test_name])

        # Make sure just one test case has been selected
        self.assertEqual(len(suite.tests), 1)

        # Make sure that the included test case name matches
        case = suite.tests[0]
        self.assertEqual(case.name, included_test_name)

    def test_exclude(self):
        """Verify that excluded tests are not executed."""
        excluded_test_name = 'test_two'

        suite = testsuite.TestSuite(self.name, self.runner,
                                    excludes=[excluded_test_name])

        # Make sure just one test case has been selected
        self.assertEqual(len(suite.tests), 1)

        # Make sure that the selected test case name matches
        case = suite.tests[0]
        self.assertNotEqual(case.name, excluded_test_name)

    def test_include_exclude(self):
        """Verify that include/exclude combination works fine."""
        included_test_name = 'test_one'
        excluded_test_name = 'test_two'

        suite = testsuite.TestSuite(self.name, self.runner,
                                    includes=[included_test_name],
                                    excludes=[excluded_test_name])
        # Make sure just one test case has been selected
        self.assertEqual(len(suite.tests), 1)

        # Make sure that the selected test case name matches
        case = suite.tests[0]
        self.assertEqual(case.name, included_test_name)
        self.assertNotEqual(case.name, excluded_test_name)


class TestTestSuiteRunlistSchema(unittest.TestCase):

    """Test suite run list schema test cases."""

    def setUp(self):
        """Initialize basic test case data."""
        schema = testsuite.TestSuite.RUNLIST_SCHEMA
        DefaultValidator.check_schema(schema)
        self.validator = DefaultValidator(schema)

    def validate(self, data):
        """Validate a schema.

        :returns: validates schema
        :rtype: object

        """
        return self.validator.validate(data)

    def test_empty_string_invalid(self):
        """Verify an empty string doesn't validate."""
        with self.assertRaises(jsonschema.ValidationError):
            self.validate('')

    def test_empty_dict_invalid(self):
        """Verify an empty dictionary doesn't validate."""
        with self.assertRaises(jsonschema.ValidationError):
            self.validate({})

    def test_empty_list_invalid(self):
        """Verify an empty list doesn't validate."""
        with self.assertRaises(jsonschema.ValidationError):
            self.validate([])

    def test_string_list_invalid(self):
        """Verify a list of strings is invalid."""
        with self.assertRaises(jsonschema.ValidationError):
            self.validate(['a', 'b', 'c'])

    def test_empty_dict_list_invalid(self):
        """Verify a list of empty dictionaries is invalid."""
        with self.assertRaises(jsonschema.ValidationError):
            self.validate([{}, {}, {}])

    def test_unknown_test_property_invalid(self):
        """Verify an unknown property in a test object invalid."""
        with self.assertRaises(jsonschema.ValidationError):
            self.validate(
                [{'test': 'test name',
                  'unknown-property': 'value',
                  }])

    def test_all_overrides_at_test_level(self):
        """Verify that all overrides at the same level as test are valid."""
        self.validate(
            [{'test': 'test name',
              'timeout': 1,
              'build_cmd': '<command>',
              'command': '<command>',
              'run_as': '<user>',
              }])

    def test_empty_overrides_invalid(self):
        """Verify an empty overrides section is invalid."""
        with self.assertRaises(jsonschema.ValidationError):
            self.validate([{'test': 'test name',
                            'overrides': {}}])

    def test_unknown_property_in_overrides_invalid(self):
        """Verify an unknown property in overrides section is invalid."""
        with self.assertRaises(jsonschema.ValidationError):
            self.validate(
                [{'test': 'test name',
                  'overrides': {'unknown-property': 'value'}
                  }])

    def test_zero_timeout_valid(self):
        """Verify a zero timeout is valid."""
        self.validate([{'test': 'test name',
                        'overrides': {'timeout': 0}}])

    def test_negative_timeout_invalid(self):
        """Verify a negative timeout is invalid."""
        with self.assertRaises(jsonschema.ValidationError):
            self.validate([{'test': 'test name',
                            'overrides': {'timeout': -1}}])

    def test_all_overrides(self):
        """Verify all properties in the overrides section are valid."""
        self.validate(
            [{'test': 'test name',
              'overrides': {
                  'timeout': 1,
                  'build_cmd': '<command>',
                  'command': '<command>',
                  'run_as': '<user>',
              },
              }])

    def test_list_valid(self):
        """Verify a list of test cases is valid."""
        self.validate(
            [{'test': 'test name'},
             {'test': 'test name 2'}])

    def test_with_overrides_list_valid(self):
        """Verify a list of test case names with overrides is valid."""
        self.validate([
            {'test': 'test name',
             'overrides': {'timeout': 1},
             },
            {'test': 'test name 2',
             'overrides': {'timeout': 1}
             },
        ])


class TestTestSuiteControlSchema(unittest.TestCase):

    """Test suite control schema test cases."""

    def setUp(self):
        """Initialize basic test case data."""
        schema = testsuite.TestSuite.CONTROL_SCHEMA
        DefaultValidator.check_schema(schema)
        self.validator = DefaultValidator(schema)

    def validate(self, data):
        """Validate a schema.

        :returns: validates schema
        :rtype: object

        """
        return self.validator.validate(data)

    def test_empty_string_invalid(self):
        """An empty string doesn't validate."""
        with self.assertRaises(jsonschema.ValidationError):
            self.validate('')

    def test_empty_dict_valid(self):
        """An empty dictionary is valid."""
        self.validate({})

    def test_unknown_property_invalid(self):
        """An unknown control property is invalid."""
        with self.assertRaises(jsonschema.ValidationError):
            self.validate({'unknown-property': 'value'})

    def test_zero_timeout_valid(self):
        """A zero timeout is valid."""
        self.validate({'timeout': 0})

    def test_negative_timeout_invalid(self):
        """A negative timeout is invalid."""
        with self.assertRaises(jsonschema.ValidationError):
            self.validate({'timeout': -1})

    def test_valid(self):
        """An example of valid control file contents."""
        self.validate({'build_cmd': 'command',
                       'ts_setup': 'command',
                       'ts_cleanup': 'command',
                       'timeout': 1,
                       })


class TestDynamicTestSuite(unittest.TestCase):

    """Ensures the functionality of tslist.auto works."""

    def setUp(self):
        """Create temporary artifacts needed for testing."""
        self.runner = mock.Mock()
        self.runner.result = ResultYAML(filename=None, append_to_file=False)
        self.runner.timeout = None
        self.runner.reboot = 'never'

        self.tmpdir = tempfile.mkdtemp()
        self.runner.testsuitedir = self.tmpdir
        self.orig_dir = os.getcwd()
        os.chdir(self.tmpdir)

        self.name = 'testsuite_name'
        os.mkdir(os.path.join(self.runner.testsuitedir, self.name))
        os.mkdir(os.path.join(self.runner.testsuitedir, self.name, self.name))

        self.goodlist = os.path.join(self.tmpdir, 'tslist.auto-good')
        with open(self.goodlist, 'w') as f:
            f.write('-\n')
            f.write('  discovery_cmd: /usr/bin/seq 3\n')
            f.write('  test_cmd: /bin/echo {}\n')

        self.faillist = os.path.join(self.tmpdir, 'tslist.auto-fail')
        with open(self.faillist, 'w') as f:
            f.write('-\n')
            f.write('  discovery_cmd: /usr/bin/seq 3\n')
            f.write('  test_cmd: /bin/false\n')

        self.discovery_err = os.path.join(self.tmpdir, 'tslist.discovery-fail')
        with open(self.discovery_err, 'w') as f:
            f.write('-\n')
            f.write('  discovery_cmd: /usr/bin/seq 3; exit 123\n')
            f.write('  test_cmd: /bin/false\n')

        self.namedlist = os.path.join(self.tmpdir, 'tslist.auto-good2')
        with open(self.namedlist, 'w') as f:
            f.write('-\n')
            f.write('  discovery_cmd: /usr/bin/seq 3\n')
            f.write('  test_cmd: /bin/echo {test} {}\n')

    def tearDown(self):
        """Remove temporary artifacts created for testing."""
        os.chdir(self.orig_dir)
        shutil.rmtree(self.tmpdir)

    def _create_ts_control(self, name, contents):
        path = os.path.join(self.tmpdir, name, name)
        os.makedirs(path)
        with open(os.path.join(path, 'ts_control'), 'w') as f:
            f.write(contents)

    def test_bad_runlist(self):
        """Ensure bad runlist throws proper error."""
        autolist = os.path.join(self.tmpdir, 'tslist.auto')
        with open(autolist, 'w') as f:
            f.write('not a valid file')
        with self.assertRaises(ValidationError):
            testsuite.DynamicTestSuite(self.name, self.runner, autolist)

    def test_good_runlist(self):
        """Ensure a good runlist works."""
        testsuite.DynamicTestSuite(self.name, self.runner, self.goodlist)

    def test_setup_passed(self):
        """Ensure the discovery commands are executed after setup."""
        # make sure it works without a ts_control file
        ts = testsuite.DynamicTestSuite(self.name, self.runner, self.goodlist)
        ts.status = 'BUILD'
        ts.setup()
        self.assertListEqual(['1', '2', '3'], [x.name for x in ts.tests])

        # now add a ts_control file with a setup command
        name = 'blah'
        self._create_ts_control(name, 'ts_setup: /bin/true')
        ts = testsuite.DynamicTestSuite(name, self.runner, self.goodlist)
        ts.status = 'BUILD'
        ts.setup()
        self.assertListEqual(['1', '2', '3'], [x.name for x in ts.tests])

    def test_setup_failed(self):
        """Ensure the discovery commands are not executed after setup."""
        name = 'tsname'
        self._create_ts_control(name, 'ts_setup: /bin/false')
        ts = testsuite.DynamicTestSuite(name, self.runner, self.goodlist)
        ts.status = 'BUILD'
        ts.setup()
        self.assertEqual(0, len(ts.tests))

    def test_run_passed(self):
        """Ensure we properly execute a runlist with passing tests."""
        ts = testsuite.DynamicTestSuite(self.name, self.runner, self.goodlist)
        ts.run()
        res = self.runner.result
        self.assertEqual(3, ts.passes)
        self.assertEqual(0, ts.failures)
        self.assertEqual(0, ts.errors)
        self.assertEqual('PASS', res.status)
        self.assertEqual(res.results[0]['stdout'], '1\n2\n3\n')
        for i in range(1, 4):
            self.assertEqual(res.results[i]['stdout'], '{}\n'.format(i))

    def test_run_passed_named(self):
        """Ensure we properly execute a runlist that repeats the tcname"""
        ts = testsuite.DynamicTestSuite(self.name, self.runner, self.namedlist)
        ts.run()
        res = self.runner.result
        self.assertEqual(3, ts.passes)
        self.assertEqual(0, ts.failures)
        self.assertEqual(0, ts.errors)
        self.assertEqual('PASS', res.status)
        self.assertEqual(res.results[0]['stdout'], '1\n2\n3\n')
        for i in range(1, 4):
            self.assertEqual(res.results[i]['stdout'], '{} {}\n'.format(i, i))

    def test_bad_discovery_cmd(self):
        """Make sure we handle a discovery command failing properly."""
        ts = testsuite.DynamicTestSuite(
            self.name, self.runner, self.discovery_err)
        ts.run()
        res = self.runner.result
        self.assertEqual(res.results[0]['returncode'], 123)
        self.assertEqual(len(res.results), 1)

    def test_run_failed(self):
        """Ensure we properly execute a runlist with failing tests."""
        ts = testsuite.DynamicTestSuite(self.name, self.runner, self.faillist)
        ts.run()
        self.assertEqual(0, ts.passes)
        self.assertEqual(3, ts.failures)
        self.assertEqual(0, ts.errors)
        self.assertEqual('PASS', ts.runner.result.status)
