# Ubuntu Testing Automation Harness
# Copyright 2012 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test bzr interface."""

import mock
import os
import shutil
import tempfile
import unittest

from utah.client.common import run_cmd
from utah.client.vcs import BzrHandler
from utah.retry import retry

BZR_TEST_REPO = "/tmp/utahbzrtest"


def setUp():
    """Create a local bzr repository for testing."""
    # should fail if the repo already exists.
    os.mkdir(BZR_TEST_REPO)

    run_cmd("bzr init", cwd=BZR_TEST_REPO)
    run_cmd("touch test.py", cwd=BZR_TEST_REPO)
    run_cmd("bzr add test.py", cwd=BZR_TEST_REPO)
    run_cmd("bzr commit -m 'Initial import' test.py", cwd=BZR_TEST_REPO)


def tearDown():
    """Remove the test repo."""
    # We should only ever get here if the repo was created in setUp().
    shutil.rmtree(BZR_TEST_REPO)


class TestBzrHandler(unittest.TestCase):

    """Test the bazaar VCS handler."""

    def setUp(self):
        """Create a temporary directory for handler."""
        self.dst_dir = tempfile.mkdtemp()
        self.dst_name = os.path.basename(self.dst_dir)
        self.dst_base = os.path.dirname(self.dst_dir)
        # now delete the directory since .get needs it to not exist
        shutil.rmtree(self.dst_dir)

    def cleanUp(self):
        """Cleanup directories we've created."""
        shutil.rmtree(self.dst_dir)

    def test_init(self):
        """Test the initial setup."""
        repo = BZR_TEST_REPO
        bzr = BzrHandler(repo=repo)

        self.assertIsNotNone(bzr)
        self.assertEqual(bzr.repo, repo)
        self.assertEqual(bzr.get_command, "bzr branch  {} ".format(repo))
        self.assertEqual(bzr.rev_command, "bzr revno")

    def test_bzr_get(self):
        """Test retrieving a bzr repo."""
        bzr = BzrHandler(repo=BZR_TEST_REPO, destination=self.dst_name)

        result = bzr.get(directory=self.dst_base)
        self.assertEqual(result['returncode'], 0)
        self.assertTrue(os.path.exists(self.dst_dir))

    def test_bzr_revision(self):
        """Test getting revision information from a bzr repo."""
        bzr = BzrHandler(repo=BZR_TEST_REPO, destination=self.dst_name)

        result = bzr.get(directory=self.dst_base)
        self.assertEqual(result['returncode'], 0)
        self.assertTrue(os.path.exists(self.dst_dir))

        result = bzr.revision(directory=self.dst_dir)
        self.assertEqual(result['returncode'], 0)

    @mock.patch('utah.client.vcs.VCSHandler.get')
    def test_bzr_retriable(self, vcs_get):
        """Ensure bzr handler uses the proper retry logic."""
        vcs_get.side_effect = [
            {'returncode': 3, 'stderr': 'http code 502'},
            {'returncode': 0, 'stderr': ''},
        ]
        bzr = BzrHandler(repo=BZR_TEST_REPO, destination=self.dst_name)
        res = retry(bzr.get, self.dst_base, retry_timeout=.1)
        self.assertEqual(0, res['returncode'])
