# Ubuntu Testing Automation Harness
# Copyright 2012 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test the dev pseudo-VCS interface."""


import os
import shutil
import tempfile
import unittest

from utah.client.vcs import DevHandler

from utah.client.tests.common import (  # NOQA
    setUp,     # Used by nosetests
    tearDown,  # Used by nosetests
)


class TestDevHandler(unittest.TestCase):

    """Test the dev pseudo-VCS handler."""

    def setUp(self):
        """Create a local directory for testing."""
        self.repo = tempfile.mkdtemp()
        with open(os.path.join(self.repo, 'test.py'), 'w') as f:
            f.write('')
        self.dest = tempfile.mkdtemp()

    def tearDown(self):
        """Remove the test repo."""
        shutil.rmtree(self.repo)
        shutil.rmtree(self.dest)

    def test_init(self):
        """Test the initial setup."""
        dev = DevHandler(repo=self.repo)

        self.assertIsNotNone(dev)
        self.assertEqual(dev.repo, self.repo)
        self.assertEqual(dev.get_command, "cp -r {} .".format(self.repo))
        self.assertEqual(dev.rev_command, "echo 'DEVELOPMENT'")

    def test_dev_get_absolute(self):
        """Test copying files."""
        dev = DevHandler(repo=self.repo, destination=self.dest)
        self.assertEqual(0, len(os.listdir(self.dest)))
        result = dev.get(directory=self.dest)
        self.assertEqual(result['returncode'], 0)
        bname = os.path.basename(self.repo)
        self.assertListEqual([bname], os.listdir(self.dest))
        d = os.path.join(self.dest, bname)
        self.assertListEqual(['test.py'], os.listdir(d))

    def test_dev_get_relative(self):
        """Test we can copy when passed a relative path."""
        runlist = os.path.join(self.repo, 'master.run')
        with open(runlist, 'w') as f:
            f.write('')

        dev = DevHandler(repo='./', destination=self.dest, runlist=runlist)
        result = dev.get(directory=self.dest)
        self.assertEqual(result['returncode'], 0)
        self.assertListEqual(['test.py', 'master.run'], os.listdir(self.dest))

    def test_dev_revision(self):
        """Test fake revision gathering support."""
        dev = DevHandler(repo=self.repo, destination=self.dest)
        result = dev.get(directory=self.dest)
        self.assertEqual(result['returncode'], 0)
        result = dev.revision(directory=self.dest)
        self.assertEqual(result['returncode'], 0)
