# Ubuntu Testing Automation Harness
# Copyright 2012 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test the git interface."""


import os
import shutil
import unittest

from utah.client.common import run_cmd
from utah.client.vcs import GitHandler


GIT_TEST_REPO = "/tmp/utahgittest"


def setUp():
    """Create a local git repository for testing."""
    # should fail if the repo already exists.
    os.mkdir(GIT_TEST_REPO)

    run_cmd("git init", cwd=GIT_TEST_REPO)
    run_cmd('git config user.email "utah@launchpad.net"', cwd=GIT_TEST_REPO)
    run_cmd("touch test.py", cwd=GIT_TEST_REPO)
    run_cmd("git add test.py", cwd=GIT_TEST_REPO)
    run_cmd("git commit -m'Initial import' test.py", cwd=GIT_TEST_REPO)


def tearDown():
    """Remove the test repo."""
    # We should only ever get here if the repo was created in setUp().
    shutil.rmtree(GIT_TEST_REPO)


class TestGitHandler(unittest.TestCase):

    """Test the bazaar VCS handler."""

    def test_init(self):
        """Test the initial setup."""
        repo = GIT_TEST_REPO
        git = GitHandler(repo=repo)

        self.assertIsNotNone(git)
        self.assertEqual(git.repo, repo)
        self.assertEqual(git.get_command, "git clone  {} ".format(repo))
        self.assertEqual(git.rev_command, "git rev-parse HEAD")

    def test_git_get(self):
        """Test retrieving a git repo."""
        repo = GIT_TEST_REPO
        dest = "utah-dev"
        git = GitHandler(repo=repo, destination=dest)
        directory = "/tmp"
        path = os.path.join(directory, dest)

        self.assertFalse(os.path.exists(path), path)

        result = git.get(directory=directory)
        self.assertEqual(result['returncode'], 0)

        self.assertTrue(os.path.exists(path), path)
        shutil.rmtree(path)

    def test_git_revision(self):
        """Test getting revision information from a git repo."""
        repo = GIT_TEST_REPO
        dest = "utah-dev"
        git = GitHandler(repo=repo, destination=dest)
        directory = "/tmp"
        path = os.path.join(directory, dest)

        self.assertFalse(os.path.exists(path))
        result = git.get(directory=directory)
        self.assertEqual(result['returncode'], 0)
        self.assertTrue(os.path.exists(path))

        result = git.revision(directory=path)

        shutil.rmtree(path)
        self.assertEqual(result['returncode'], 0)
