# Ubuntu Testing Automation Harness
# Copyright 2012 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""UTAH client VCS support."""

import logging
import os
import subprocess

from utah.client.common import (
    CMD_TS_FETCH,
    run_cmd,
)
from utah.exceptions import UTAHException


REPO_DEFAULT_DIR = "."


class VCSHandler(object):

    """Base class for Version Control System support.

    :param repo: Repository location
    :type repo: str
    :param destination:
        Local directory where the repository should be made available
    :type destination: str

    """

    def __init__(self, repo, destination='', runlist=None):
        self.repo = repo
        self.destination = destination
        self.runlist = runlist

    def get(self, directory=REPO_DEFAULT_DIR):
        """Execute VCS get command.

        The get command will make a copy of a given repository to the directory
        specified as destination.

        :param directory: Working directory
        :type directory: str
        :returns: Get command execution result
        :rtype: dict

        .. seealso:: :func:`run_cmd`

        """
        return run_cmd(self.get_command, cwd=directory, cmd_type=CMD_TS_FETCH)

    def revision(self, directory=REPO_DEFAULT_DIR):
        """Execute revision command.

        :param directory: Working directory
        :type directory: str
        :returns: Revision command execution result.
        :rtype: dict

        """
        return run_cmd(self.rev_command, cwd=directory, cmd_type=CMD_TS_FETCH)


class BzrHandler(VCSHandler):

    """Bazaar VCS handler."""

    def __init__(self, repo, branch=True, options="", destination="",
                 **kwargs):
        super(BzrHandler, self).__init__(repo, destination, **kwargs)
        self.options = options

        if branch:
            self.get_command = "bzr branch {} {} {}".format(
                options,
                self.repo,
                self.destination,
            )

            self.rev_command = "bzr revno"
        else:
            self.get_command = "bzr export {} {} {}".format(
                options,
                self.destination,
                self.repo,
            )

            self.rev_command = "bzr revno {}".format(self.repo)

    @staticmethod
    def _retriable(result):
        errors = [
            'bzr: ERROR: Not a branch:',
        ]
        for e in errors:
            if e in result['stderr']:
                return False
        return True

    def get(self, directory=REPO_DEFAULT_DIR):
        """Override the super-class get to properly deal with retry logic.

        :returns: Get command execution result
        :rtype: dict

        """
        res = super(BzrHandler, self).get(directory)
        if res['returncode'] == 3 and self._retriable(res):
            msg = 'Launchpad temporary error detected: {}'.format(res)
            raise UTAHException(msg, retry=True)
        return res


class GitHandler(VCSHandler):

    """Git VCS handler."""

    def __init__(self, repo, options="", destination="", **kwargs):
        super(GitHandler, self).__init__(repo, destination, **kwargs)
        self.options = options

        self.get_command = "git clone {} {} {}".format(
            options,
            self.repo,
            self.destination,
        )

        self.rev_command = "git rev-parse HEAD"

    def sparse_checkout(self, paths):
        """
        Do a sparse checkout in the destination folder
        :param paths: the repository path to be added in the checkout
        """
        current_dir = os.getcwd()
        os.chdir(self.destination)
        try:
            subprocess.check_call(['git', 'init'])
            subprocess.check_call(['git', 'remote', 'add', '-f', 'origin',
                                   self.repo])
            subprocess.check_call(['git', 'config', 'core.sparseCheckout',
                                   'true'])
            with open('.git/info/sparse-checkout', 'w') as f:
                for path in paths:
                    f.write(path)
            self.pull()
        except subprocess.CalledProcessError:
            raise
        finally:
            os.chdir(current_dir)

    def pull(self, repo='origin', branch='master'):
        current_dir = os.getcwd()
        os.chdir(self.destination)
        try:
            subprocess.check_call(['git', 'pull', repo, branch])
        except subprocess.CalledProcessError:
            raise
        finally:
            os.chdir(current_dir)


class DevHandler(VCSHandler):

    """Development VCS handler.

    This isn't a handler for any VCS, but a helper to ease development and let
    runlists point to a path that can be copied locally.

    """

    def __init__(self, repo, destination=".", runlist=None):
        super(DevHandler, self).__init__(repo, destination, runlist)

        if self.repo[0] != '/':
            logging.debug('using relative path for dev handler')
            path = os.path.abspath(runlist)
            path = os.path.dirname(path)
            self.repo = os.path.join(path, self.repo)

        self.get_command = "cp -r {} {}".format(
            self.repo,
            self.destination,
        )
        self.rev_command = "echo 'DEVELOPMENT'"


def create_vcs_handler(method, location, destination, runlist):
    """Find an instantiate the proper VCS handler object.

    :return: VCSHandler object
    :rtype: VCSHandler

    """
    kwargs = {
        'repo': location,
        'destination': destination,
        'runlist': runlist,
    }

    handlers = {
        'bzr': (BzrHandler, {}),
        'bzr-export': (BzrHandler, {'branch': False}),
        'dev': (DevHandler, {}),
        'git': (GitHandler, {}),
    }

    cls, args = handlers[method]
    kwargs.update(args)
    return cls(**kwargs)
