#!/usr/bin/env python

# Ubuntu Testing Automation Harness
# Copyright 2012 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Provide config variables via utah.config.

When run directly as part of the package build, output machine-independent
defaults as json for writing a config file.

"""

import getpass
import json
import logging
import os
import socket
import subprocess
import sys
import traceback

from argparse import ArgumentTypeError

import jsonschema

from utah.url import url_argument


class _Config(object):

    """Configuration handling class."""

    SCHEMA = {
        "$schema": "http://json-schema.org/draft-04/schema#",
        'title': 'UTAH configuration',
        'type': 'object',
        'additionalProperties': True,
        'properties': {
            'arch': {
                'description': 'Default architecture',
                'type': 'string',
            },
            'boot': {
                'description': 'Default kernel parameters',
                'type': ['string', 'null'],
                'default': None,
            },
            'boottimeout': {
                'description': 'Time to wait for installed system to boot',
                'type': 'integer',
                'minimum': 0,
                'default': 900,
            },
            'bridge': {
                'description': 'Interface for VMs bridget network',
                'type': 'string',
            },
            'checktimeout': {
                'description': (
                    'Time to wait between checking if system '
                    'is available over ssh'),
                'type': 'integer',
                'minimum': 0,
                'default': 180,
            },
            'clean': {
                'description': (
                    'Setting for whether to clean up after a test run'),
                'type': 'boolean',
                'default': True,
            },
            'poweroff': {
                'description': (
                    'Setting for whether to power off VM after a test run'),
                'type': 'boolean',
                'default': False,
            },
            'client_install_timeout': {
                'description': (
                    'Maximum amount of time to wait '
                    'for the client package to install'),
                'type': 'integer',
                'minimum': 0,
                'default': 120,
            },
            'consoleloglevel': {
                'description': 'Log level to print to the console',
                'type': 'integer',
                'minimum': 0,
                'default': logging.WARNING,
            },
            'debug': {
                'description': 'Default debug setting',
                'type': 'boolean',
                'default': False,
            },
            'debuglog': {
                'description': 'Path to debug log file',
                'type': 'string',
                'format': 'path',
            },
            'dlcommand': {
                'description': 'Command to download test images',
                'type': 'string',
                'format': 'path',
                'default': 'dl-ubuntu-test-iso',
            },
            'dlpercentincrement': {
                'description': (
                    'Percentage increment '
                    'at which to report download updates'),
                'type': 'integer',
                'minimum': 0,
                'maximum': 100,
                'default': 10,
            },
            'dlretries': {
                'description': 'Number of times to retry image download',
                'type': 'integer',
                'minimum': 1,
                'default': 10,
            },
            'diskbus': {
                'description': 'Diskbus for VMs',
                'type': 'string',
                'default': 'virtio',
            },
            'disksizes': {
                'description': 'Default disk size list for VMs (in GB)',
                'type': 'array',
                'items': {'type': 'integer'},
                'minItems': 1,
                'default': [8],
            },
            'emulator': {
                'description': (
                    'Emulator for VMs. '
                    'If None, kvm will be used if available, otherwise qemu'),
                'type': ['string', 'null'],
                'default': None,
            },
            'fileloglevel': {
                'description': 'Default log level to write to log file',
                'type': 'integer',
                'minimum': 0,
                'default': logging.INFO,
            },
            'group': {
                'description': 'Group for file ownership and permissions',
                'type': 'string',
                'default': 'utah',
            },
            'hostname': {
                'description': (
                    'Hostname (used for logfile among other things)'),
                'type': 'string',
            },
            'image': {
                'description': 'Default image to use for system installation',
                'type': ['string', 'null'],
                'format': 'path',
                'default': None,
            },
            'initrd': {
                'description': 'Default initrd to use for system installation',
                'type': ['string', 'null'],
                'format': 'path',
                'default': None,
            },
            'installpackages': {
                'description': 'Packages to install on system',
                'type': 'array',
                'items': {'type': 'string'},
                'default': ['openssh-server'],
            },
            'installtimeout': {
                'description': 'Time limit for overall install process',
                'type': 'integer',
                'default': 9000,
            },
            'installtype': {
                'description': (
                    'Install type '
                    '(mini is default to save bandwidth and space)'),
                'type': 'string',
                'default': 'mini',
            },
            'isodir': {
                'description': 'Directory to store local images',
                'type': 'string',
                'format': 'path',
                'default': '/var/cache/utah/iso',
            },
            'jobtimeout': {
                'description': 'Overall timeout for run_utah_tests',
                'type': ['integer', 'null'],
                'default': None,
            },
            'kernel': {
                'description': 'Kernel for installation',
                'type': ['string', 'null'],
                'format': 'path',
                'default': None,
            },
            'launchpad_user': {
                'description': 'Launchpad whose keys should be imported on a live-server install',
                'type': 'string',
                'default': 'server-team-bot',
            },
            'logfile': {
                'description': 'Path to log file',
                'type': 'string',
                'format': 'path',
            },
            'logpath': {
                'description': 'Location of log directory',
                'type': 'string',
                'format': 'path',
                'default': '/var/log/utah',
            },
            'logrotation_when': {
                'description': (
                    'Server files rotation using '
                    'logging.handlers.TimedRotatingLogHandler '
                    '(when parameter)'),
                'type': 'string',
                'default': 'W0',
            },
            'logrotation_interval': {
                'description': (
                    'Server files rotation using '
                    'logging.handlers.TimedRotatingLogHandler '
                    '(interval parameter)'),
                'type': 'integer',
                'default': 1,
            },
            'logrotation_backupCount': {
                'description': (
                    'Server files rotation using '
                    'logging.handlers.TimedRotatingLogHandler '
                    '(backupCount parameter)'),
                'type': 'integer',
                'default': 10,
            },
            'machinedir': {
                'description': (
                    'Directory to hold non-temporary files (mostly for VM)'),
                'type': ['string', 'null'],
                'default': None,
            },
            'machineid': {
                'description': 'Machine ID',
                'type': ['string', 'null'],
                'default': None,
            },
            'machineinfo': {
                'description': (
                    'Machine information object for physical machines'),
                'type': ['string', 'null'],
                'default': None,
            },
            'machinetype': {
                'description': 'Machine type',
                'type': 'string',
                'default': 'virtual',
            },
            'machineuuid': {
                'description': 'UUID for machine',
                'type': ['string', 'null'],
                'default': None,
            },
            'macs': {
                'description': 'List of mac addresses for VMs',
                'type': 'array',
                'items': {'type': 'string'},
                'default': [],
            },
            'name': {
                'description': 'Machine name',
                'type': ['string', 'null'],
                'default': None,
            },
            'new': {
                'description': 'Install a new machine or use an existing one',
                'type': 'boolean',
                'default': False,
            },
            # NFS options only used for cobbler-based desktop installs
            'nfscommand': {
                'description': 'NFS service command',
                'type': 'array',
                'items': {'type': 'string'},
                'default': ['sudo', 'service', 'nfs-kernel-server'],
            },
            'nfsconfigfile': {
                'description': 'Path to NFS config file',
                'type': 'string',
                'format': 'path',
                'default': '/etc/exports',
            },
            'nfsiface': {
                'description': 'nfs interface',
                'type': 'string',
            },
            'nfsoptions': {
                'description': 'Options for NFS shares',
                'type': 'string',
                'default': '*(ro,async,no_root_squash,no_subtree_check)',
            },
            'outputpreseed': {
                'description': (
                    'Whether to output the final preseed file as a log file '
                    'during the main Provision/Run Tests/Gather Results '
                    'process'
                ),
                'type': 'boolean',
                'default': False,
            },
            'packagedir': {
                'description': (
                    'Directory where utah client '
                    'and other needed packages reside'),
                'type': 'string',
                'default': '/usr/share/utah',
            },
            'powercmd': {
                'description': 'Command to use to power cycle machine',
                'type': ['string', 'null'],
                'default': None,
            },
            'powertimeout': {
                'description': (
                    'Time to wait between power off '
                    'and power on for physical systems'),
                'type': 'integer',
                'default': 15,
            },
            'preboot': {
                'description': 'Preboot for bamboo-feeder panda boards',
                'type': ['string', 'null'],
                'default': None,
            },
            'prefix': {
                'description': 'Prefix for machine names',
                'type': 'string',
                'default': 'utah',
            },
            'preseed': {
                'description': 'Preseed',
                'type': 'string',
                'format': 'path',
                'default': '/etc/utah/default-preseed.cfg',
            },
            'pxedir': {
                'description': 'Location of PXE configuration files',
                'type': 'string',
                'format': 'path',
                'default': '/var/lib/tftpboot/pxelinux.cfg',
            },
            'qemupath': {
                'description': 'libvirt URL for virtual machine creation',
                'type': 'string',
                'default': 'qemu:///system',
            },
            'retry_timeout': {
                'description': 'Timeout between retry attempts',
                'type': 'integer',
                'default': 3,
            },
            'rewrite': {
                'description': 'Setting for configuration rewriting',
                'type': 'string',
                'default': 'all',
            },
            'runtimeout': {
                'description': (
                    'Time to wait for client process to finish in run.py. '
                    'Default setting is to wait indefinitely'),
                'type': ['integer', 'null'],
                'default': None,
            },
            'series': {
                'description': 'Default series',
                'type': 'string',
            },
            'serieschoices': {
                'description': 'Available releases',
                'type': 'array',
                'items': {'type': 'string'},
                'default': [
                    'lucid',
                    'precise',
                    'quantal',
                    'raring',
                    'saucy',
                    'trusty',
                    'utopic'
                ],
            },
            'sqlite3_db_connection_timeout': {
                'description': (
                    'sqlite database connection timeout. '
                    'A value higher than the default is used '
                    'to avoid db locking problems'),
                'type': 'integer',
                'default': 30,
            },
            'ssh_logfile': {
                'description': 'Path to ssh log file',
                'type': 'string',
            },
            'sshknownhosts': {
                'description': 'Path to .ssh/known_hosts file',
                'type': 'string',
                'format': 'path',
            },
            'sshprivatekey': {
                'description': 'Path to ssh private key file',
                'type': 'string',
                'format': 'path',
            },
            'sshpublickey': {
                'description': 'Path to ssh public key file',
                'type': 'string',
                'format': 'path',
            },
            'template': {
                'description': 'Machine template',
                'type': ['string', 'null'],
                'default': None,
            },
            'template_dir': {
                'description':
                ('location of the jinja2 templates '
                 'used by the provisionig module'),
                'type': 'string',
                'default': '/usr/share/utah/templates',
            },
            'user': {
                'description': (
                    'User to install and use on provisioned machines'),
                'type': 'string',
            },
            'variant': {
                'description': (
                    'Default architecture variant '
                    '(armel, armhd, non-pae, etc.)'),
                'type': ['string', 'null'],
                'default': None,
            },
            'vmpath': {
                'description': 'Path for VM files',
                'type': 'string',
                'format': 'path',
            },
            'wwwdir': {
                'description': 'Directory to hold web-accessible files',
                'type': 'string',
                'format': 'path',
                'default': '/var/www/utah',
            },
            'wwwiface': {
                'description': 'www interface',
                'type': 'string',
            },
            'xml': {
                'description': 'Path to VM XML file',
                'type': 'string',
                'format': 'path',
                'default': '/etc/utah/default-vm.xml',
            },
        },
    }

    # Default values from the schema
    # Note: dict comprehension not used to avoid static analysis warning
    # regarding SCHEMA being undefined
    DEFAULTS = dict(
        [(key, value['default'])
         for key, value in SCHEMA['properties'].iteritems()
         if 'default' in value])

    def __init__(self):
        # A mapping from variable names to file names
        # that were used to set its value
        self._set_by = {}

        self._set_local_defaults()
        self._set_log_files()
        self._process_config_files()
        self._expand_user_paths()

    def __iter__(self):
        """Iterate through all configuration keys/values."""
        data = self.DEFAULTS.copy()
        # Filter out private variables
        public_vars = {
            key: value
            for key, value in vars(self).iteritems()
            if not key.startswith('_')}
        data.update(public_vars)
        return data.iteritems()

    def __getattr__(self, key):
        """Use DEFAULTS as fallback when looking for a value."""
        if key in self.DEFAULTS:
            return self.DEFAULTS[key]

        raise AttributeError(key)

    def _set_local_defaults(self):
        """Set local default values."""
        # Default network bridge for machines
        self.bridge = self._getbridge()
        self.nfsiface = self.bridge
        self.wwwiface = self.bridge

        # Hostname (used for logfile among other things)
        self.hostname = socket.gethostname()
        # SSH private key to use for provisioned machine logins
        self.sshprivatekey = os.path.join('~', '.ssh', 'utah')
        # SSH public key to install on provisioned machines
        self.sshpublickey = os.path.join('~', '.ssh', 'utah.pub')
        # SSH known hosts file
        self.sshknownhosts = os.path.join('~', '.ssh', 'known_hosts')
        # User to install and use on provisioned machines
        self.user = getpass.getuser()
        # Path for VM files
        self.vmpath = os.path.join('~', 'vm')

        self._set_arch()
        self._set_series()

    def _set_arch(self):
        """Set arch default value based on dpkg-architecture output."""
        command = ['dpkg-architecture', '-qDEB_HOST_ARCH']
        self.arch = self._check_binary_and_run_command(command) or 'i386'

    def _set_series(self):
        """Set series default based on lb_release output."""
        command = ['lsb_release', '-sc']
        self.series = self._check_binary_and_run_command(command) or 'precise'

    def _check_binary_and_run_command(self, command):
        """Check binary exists and run command."""
        with open(os.devnull, 'w') as fnull:
            try:
                if subprocess.call(['which', command[0]],
                                   stdout=fnull, stderr=fnull) == 0:
                    # If command is available, return its output
                    return subprocess.check_output(command).strip()
            except (OSError, subprocess.CalledProcessError):
                return None

    @staticmethod
    def _getbridge():
        """Get virbr0 if it is set in the qemu config or the utah config.

        Use this (or the default virbr0) as the bridge for bridged network VMs.

        """
        dns = os.path.join('/', 'etc', 'utah', 'dns')
        if os.path.isfile(dns):
            try:
                process = subprocess.Popen([dns], stdout=subprocess.PIPE)
                output = process.communicate()[0].strip()
                return output
            except OSError:
                return 'virbr0'
        else:
            return 'virbr0'

    def _set_log_files(self):
        """Set log files according to path."""
        # These depend on other config options, so they're added last.
        # Default logfile is /var/log/utah/{hostname}.log
        self.logfile = os.path.join(
            self.logpath,
            '{}.log'.format(self.hostname))
        # Default logfile is /var/log/utah/{hostname}-debug.log
        # Set to None to disable separate debug log
        self.debuglog = os.path.join(
            self.logpath,
            '{}-debug.log'.format(self.hostname))

        # Log file used to log SSH related information
        self.ssh_logfile = os.path.join(
            self.logpath,
            '{}-ssh.log'.format(self.hostname))

    def _process_config_files(self):
        """Process all configuration files."""
        files = self._get_config_files()

        for conffile in files:
            self._process_config_file(conffile)

    def _get_config_files(self):
        """Get a list of configuration files."""
        # First, the standard /etc/utah/config file is processed
        # Second, anything in /etc/utah/conf.d
        # Third, any existing ~/.utah/config file
        # Fourth, anything in ~/.utah/conf.d
        # Fifth, a file referred to by UTAH_CONFIG_FILE
        # Sixth, anything in UTAH_CONFIG_DIR
        # Technically, all files will process as directories, and vice versa
        files = []
        paths = ['/etc/utah/config', '/etc/utah/conf.d', '~/.utah/config',
                 '~/.utah/conf.d', os.getenv('UTAH_CONFIG_FILE'),
                 os.getenv('UTAH_CONFIG_DIR')]
        for path in paths:
            if path is not None:
                abspath = os.path.expanduser(path)
                if os.path.isdir(abspath):
                    for walkdir, _dirs, walkfiles in os.walk(abspath):
                        for walkfile in walkfiles:
                            filename = os.path.join(abspath, walkdir, walkfile)
                            files.append(filename)
                elif os.path.isfile(abspath):
                    files.append(abspath)
                elif abspath.startswith('lp:'):
                    try:
                        files.append(url_argument(abspath))
                    except ArgumentTypeError:
                        traceback.print_exc()
        return files

    def _process_config_file(self, conffile):
        """Read configuration file and update values."""
        if os.path.isfile(conffile) and os.path.getsize(conffile) > 0:
            with open(conffile) as fp:
                try:
                    data = json.load(fp)
                except ValueError:
                    sys.stderr.write(
                        'ERROR: {} is not a valid JSON file\n'
                        .format(conffile))
                    return
                try:
                    jsonschema.validate(data, self.SCHEMA)
                except jsonschema.ValidationError as exception:
                    sys.stderr.write(
                        'ERROR: {} is not a valid configuration file\n'
                        'Detailed information: {}\n'
                        .format(conffile, exception.message))
                    return

                for key, value in data.iteritems():
                    if key in self._set_by:
                        sys.stderr.write(
                            'INFO: {} was already set by {}\n'
                            .format(key, self._set_by[key]))
                    setattr(self, key, value)
                    self._set_by[key] = conffile

    def _expand_user_paths(self):
        """Support ~-based paths on variables that use paths."""
        # Options that are paths, are marked with "format: 'path'"
        options = [key for key, value in self.SCHEMA['properties'].iteritems()
                   if value.get('format') == 'path']
        for option in options:
            value = getattr(self, option)
            if value is not None:
                setattr(self, option, os.path.expanduser(value))

    def dump_defaults(self):
        """Dump the default values to create a configuration file."""
        return json.dumps(self.DEFAULTS, sort_keys=True, indent=4)

    def dump(self):
        """Dump all config values."""
        data = dict(self)
        return json.dumps(data, sort_keys=True, indent=4)


config = _Config()

if __name__ == '__main__':
    print config.dump_defaults()
