#!/usr/bin/env python

# Ubuntu Testing Automation Harness
# Copyright 2012 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

#
# Copyright (C) 2012, Canonical Ltd (http://www.canonical.com/)
#
# This file is part of utah.
#
# iso_static_validation is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# iso_static_validation is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ubuntu-server-iso-testing.  If not, see
# <http://www.gnu.org/licenses/>.
#

"""ISO static validation.

The test cases in this scripts are used to make sure that an ISO image isn't
broken and can be used for smoke testing.

"""
# TODO: support for non-daily images


import os
import argparse
import sys
import logging
import urllib2
import subprocess
import hashlib
import unittest
import re
import tempfile
import shutil
import time
import platform
import distro_info
from traceback import format_exception

import yaml

lib_path = os.path.abspath('../')
sys.path.append(lib_path)
from utah.iso import ISO
from utah.client.probe import probes
from utah.client.vcs import GitHandler

# Defaults
DEFAULT_ISOROOT = os.path.expanduser('~/Downloads')
DEFAULT_URL = 'http://cdimage.ubuntu.com/'
DEFAULT_LOG_DIR = '/var/log/utah/'

#constants
ONE_MB_BLOCK = 2 ** 20

#Read the iso name and initialise global variables
parser = argparse.ArgumentParser()
parser.add_argument('--name', help='Path to the ISO file to be tested')
args = parser.parse_args()


def configure_logging(log_level=logging.DEBUG):
    """Configure logging to stderr using the passed log_level.

    :param log_level: Log level to be used
    :type log_level: int

    """
    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(levelname)s: %(message)s')

    log_handler = logging.StreamHandler()
    log_handler.setFormatter(formatter)
    log_handler.setLevel(log_level)

    logger.addHandler(log_handler)


configure_logging()

if args.name is None or '.iso' not in args.name:
    print 'The usage:', sys.argv[0],  '--name release-variant-arch.iso'
    sys.exit(1)
else:
    # Set default path
    iso_location = args.name
    # Iterate over possible paths, starting with old default behavior
    for path in (os.path.join(DEFAULT_ISOROOT, args.name), args.name):
        if os.path.isfile(path):
            iso_location = path
            break
    iso = ISO(image=iso_location)
    st_release = iso.series
    st_variant = iso.installtype
    st_arch = iso.arch


#Set a flag for ubiquity based images
if st_variant == 'desktop' or st_variant == 'dvd':
    ubiquity_image = True
else:
    ubiquity_image = False

#Set a flag for subiqutiy based images
if st_variant == 'live-server':
    subiquity_image = True
else:
    subiquity_image = False

#Set a flag for UEFI&SecureBoot
if (st_variant in ('desktop', 'server', 'live-server', 'dvd')
        and st_arch == 'amd64'):
    uefi_image = True
else:
    uefi_image = False

#Setup the path for data files
DATA_PATH = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data')


class TestValidateISO(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        probes.enable_probes([])
        try:
            cls.temp_dir = tempfile.mkdtemp()
        except OSError:
            logging.error("Creating temp directory failed")
            unittest.fail("Setup error")

    def setUp(self):
        logging.info('\n{separator}\n{id}\n{description}\n{separator}'
                     .format(separator='-' * 80,
                             id=self.id(),
                             description=self.shortDescription() or ''))
        self.block_size = ONE_MB_BLOCK
        self.iso_location = iso_location
        logging.debug('Using iso at: %s', self.iso_location)
        self.iso = ISO(image=self.iso_location, logger=logging)
        self.st_release = self.iso.series
        self.st_variant = self.iso.installtype
        self.st_arch = self.iso.arch
        self.iso_name = ('{}.iso'
                         .format('-'.join([self.st_release,
                                           self.st_variant,
                                           self.st_arch])))
        logging.debug('Standard name for this iso is: %s', self.iso_name)

        try:
            ubuntu_distro_devel = distro_info.UbuntuDistroInfo().devel()
        except:
            self.fail("Can not find distro info details. "
                      "Distro info data could be out of date")

        if self.st_release != ubuntu_distro_devel:
            self.url = os.path.join(DEFAULT_URL, self.st_release)
        else:
            self.url = DEFAULT_URL

        if self.st_variant == 'desktop':
            self.url = os.path.join(self.url, 'daily-live')
        elif self.st_variant == 'dvd':
            self.url = os.path.join(self.url, 'dvd')
        elif self.st_variant == 'alternate':
            self.url = os.path.join(self.url, 'daily')
        elif self.st_variant == 'live-server':
            if self.st_release != ubuntu_distro_devel:
                self.url = os.path.join(DEFAULT_URL, 'ubuntu-server',
                                        self.st_release, 'daily-live')
            else: #current dev release
                self.url = os.path.join(DEFAULT_URL, 'ubuntu-server',
                                        'daily-live')
        elif self.st_variant == 'server':
            if self.st_release != ubuntu_distro_devel:
                self.url = os.path.join(DEFAULT_URL, 'ubuntu-server',
                                        self.st_release, 'daily')
            else:  # current dev release
                self.url = os.path.join(DEFAULT_URL, 'ubuntu-server', 'daily')
        else:
            self.fail("Image name be in the form of release-variant-arch.iso")

    def test_sha256_checksum(self):
        """Test if the ISO SHA checksum matches what is given in the server."""
        current_iso_digest = {}
        sha256sum_url = os.path.join(self.url, 'pending', 'SHA256SUMS')
        try:
            fh = urllib2.urlopen(sha256sum_url)
        except urllib2.HTTPError, e:
            logging.error("Failed to fetch URL '%s': %s. Aborting!",
                          sha256sum_url, e)
            self.fail("Failed to fetch sha checksum from the server")

        # Process the file (contains a number of entries containing iso
        # names and the corresponding checksums)
        for digest in fh:
            (chksum, sep,  iso_image) = digest.partition(' *')
            current_iso_digest[iso_image.rstrip()] = chksum
        fh.close()

        try:
            iso_sha256 = current_iso_digest[self.iso_name]
        except KeyError:
            logging.error("No digest found for image '%s'", self.iso_name)
            self.fail("The iso does not have a matching digest")

        # Validate checksum of file fname against sha256 hash
        logging.debug("Calculating hash for file '%s'", self.iso_name)
        sha256 = hashlib.sha256()
        with open(self.iso_location) as f:
            while True:
                data = f.read(self.block_size)
                if not data:
                    break
                sha256.update(data)
        sha256 = sha256.hexdigest()
        logging.debug("Expected Checksum: '%s'", iso_sha256)
        logging.debug("Local File Checksum: '%s'", sha256)
        self.assertEqual(iso_sha256, sha256)

    def test_files_list(self):
        """Test if the list repository file matches the content of the ISO."""
        list_url = os.path.join(
            self.url, 'pending', '{}.list'
            .format('-'.join([self.st_release, st_variant, self.st_arch])))

        try:
            list_repository = urllib2.urlopen(list_url)
        except urllib2.HTTPError, e:
            logging.error("Failed to fetch URL '%s': %s . Aborting!",
                          list_url, e)
            self.fail("Failed to fetch files list from the server")

        logging.debug('Extracting list file info from ISO')
        (stdout, stderr) = self.iso.listfiles()
        logging.debug('Checking for error in extracting the file list')
        self.assertEqual(stderr, '')

        logging.debug('Checking that each list entry is in the iso')
        iso_files = set(os.path.normpath(path)
                        for path in stdout.splitlines())
        for list_entry in list_repository:
            fname = os.path.normpath(list_entry[1:].rstrip())
            self.assertIn(fname, iso_files,
                          '{!r} not found in:\n{}'
                          .format(fname, stdout))

    @unittest.skipUnless(ubiquity_image,
                         "manifest only for ubiquity based images")
    def test_manifest(self):
        """Test if the ISO manifest matches the one in the server."""
        manifest_url = os.path.join(
            self.url, 'pending', '{}.manifest'
            .format('-'.join([self.st_release, st_variant, self.st_arch])))

        try:
            manifest = urllib2.urlopen(manifest_url)
        except urllib2.HTTPError, e:
            logging.error("Failed to fetch URL '%s': %s . Aborting!",
                          manifest_url, e)
            self.fail("Failed to fetch manifest file from the server")

        logging.debug('Extracting manifest from the iso')
        manifest_path = self.iso.extract('casper/filesystem.manifest',
                                         self.temp_dir)
        manifest_local = open(manifest_path)
        logging.debug('Checking manifest are the same in iso and repo')
        for manifest_entry in manifest:
            self.assertIn(manifest_entry, manifest_local)

    def test_build_id(self):
        """Test if the ISO buildid matches the one in the server.

        The buildid is extracted from the distro info file.

        """
        logging.debug('Extract distro information of the image from the iso')
        buildid_path = self.iso.extract("./.disk/info", self.temp_dir)
        buildinfo = open(buildid_path)

        try:
            buildid_page = urllib2.urlopen(self.url).read()
        except urllib2.HTTPError, e:
            logging.error("Failed to fetch URL '%s': %s. Abort!", self.url, e)
            self.fail("Failed to fetch build page info from the server")

        logging.debug('Look for the current buildid from the downloaded page')
        versions = sorted(re.findall(r'"([0-9]{8}(\.[0-9])?)/"', buildid_page))
        id_server = '(' + versions[-1][0] + ')'

        for build_info_line in buildinfo:
            logging.debug('Checking buildids are the same in iso and repo')
            self.assertIn(id_server, build_info_line)

    @unittest.skipUnless(st_variant == 'desktop' and
                         st_arch != 'powerpc',
                         "Skipping images that aren't ubiquity-based.")
    @unittest.skipIf(st_release != 'trusty', "Skipping images without wubi.")
    def test_wubi(self):
        """Test if  wubi is present for desktop or dvd and it is a valid PE."""
        logging.debug('Extracting wubi from the desktop images')
        wubi_path = self.iso.extract('./wubi.exe', self.temp_dir)
        cmd = ["file", wubi_path]
        output = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                                  stderr=subprocess.PIPE)
        (stdout, stderr) = output.communicate()
        logging.debug('Check that wubi has a valid PE header')
        self.assertEqual(stderr, '')
        self.assertIn(" PE32 executable (GUI)", stdout)

    @unittest.skipUnless(st_variant == 'desktop' and
                         st_arch != 'powerpc', "Skipping for non desktop iso")
    def test_files_ubiquity(self):
        """Test if the relevant files are present in the ISO for desktop."""
        (stdout, stderr) = self.iso.listfiles()
        logging.debug('Check for error in extracting file list from the iso')
        self.assertEqual(stderr, '')
        files_list = open(os.path.join(DATA_PATH, 'file_list_ubiquity'))
        exclude_files = []
        logging.debug('Check if relevant files are present in the iso')
        for list_server in files_list:
            path = list_server.rstrip()
            if path in exclude_files:
                logging.debug('%s excluded based on release', path)
            else:
                self.assertIn(path, stdout)

    # Test is skipped for [s]ubiquity based images
    @unittest.skipIf(ubiquity_image, "Skipping for ubiquity images")
    @unittest.skipIf(subiquity_image, "Skipping for subiquity images")
    def test_files_di(self):
        """Test if the relevant files are present for d-i based images ISO.

        Two list of files are used:
        file_list_di for i386 and amd64
        file_list_di_powerpc for powerpc and amd64+mac

        """
        (stdout, stderr) = self.iso.listfiles()
        logging.debug('Check for error in extracting file list from the iso')
        self.assertEqual(stderr, '')

        if self.st_arch in ['powerpc', 'ppc64el', 'amd64+mac']:
            files_list = open(os.path.join(DATA_PATH, 'file_list_di_powerpc'))
        else:  # i386 or amd64
            files_list = open(os.path.join(DATA_PATH, 'file_list_di'))

        for list_server in files_list:
            logging.debug('check if important d-i files are present in iso')
            path = list_server.rstrip()
            self.assertIn(path, stdout)

    @unittest.skipUnless(subiquity_image, "Skipping for non-subiquity images")
    def test_files_subiquity(self):
        """Test if the relevant files are present for subiquity isos."""
        (stdout, stderr) = self.iso.listfiles()
        logging.debug('Check for error in extracting file list from the iso')
        self.assertEqual(stderr, '')

        if self.st_arch == 'ppc64el':
            files_list = open(os.path.join(DATA_PATH,
                              'file_list_subiquity_powerpc'))
        else:
            files_list = open(os.path.join(DATA_PATH, 'file_list_subiquity'))

        for list_server in files_list:
            logging.debug('check if important d-i files are present in iso')
            path = list_server.rstrip()
            self.assertIn(path, stdout)

    @unittest.skipIf(st_arch in ['powerpc', 'ppc64el'],
                         "vmlinuz is not present only for powerpc images")
    def test_vmlinuz(self):
        """Test if vmlinuz is a valid executable."""
        logging.debug('Extract kernel image from the iso')
        path = iso.kernelpath()
        vmlinuz_path = self.iso.extract(path, self.temp_dir)

        # In Trusty the type returned by file for the kernel is different from
        # other releases
        host_release = subprocess.check_output(['lsb_release', '-sc'])
        kernel_string = 'Linux kernel'
        if 'trusty' in host_release:
            kernel_string = 'x86 boot sector'

        cmd = ['file', vmlinuz_path]
        output = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                                  stderr=subprocess.PIPE)
        (stdout, stderr) = output.communicate()
        logging.debug('Check if vmlinuz present in the iso is a Linux kernel')
        self.assertEqual(stderr, '')
        self.assertIn(kernel_string, stdout)

    @unittest.skipUnless(ubiquity_image or
                         (st_variant == 'server' and st_release > 'precise'),
                         "Skipping for d-i images without squashfs.")
    def test_filesystem_squashfs(self):
        """Test if filesystem.squashfs is present and its contents.

        The squash fs must be for ubiquity based images.

        """
        logging.debug('Extracting the filesystem.squashfs')
        if ubiquity_image:
            squashfs_name = 'casper/filesystem.squashfs'
        else:
            squashfs_name = 'install/filesystem.squashfs'

        squashfs_path = self.iso.extract(squashfs_name,
                                         self.temp_dir)
        logging.debug('Checking the file is an actual .squashfs')
        cmd = ["file", squashfs_path]
        output = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                                  stderr=subprocess.PIPE)
        (stdout, stderr) = output.communicate()
        logging.debug('Check if signature contains "Squashfs filesystem"')
        self.assertEqual(stderr, '')
        self.assertIn("Squashfs filesystem", stdout)

        # Skipping vmlinuz image extraction on server images
        if not ubiquity_image:
            return

        #unsquashfs extracts only in pwd
        cwd = os.getcwd()
        try:
            os.chdir(self.temp_dir)

            cmd = ["unsquashfs", "-f",  squashfs_path, "-l", "vmlinuz"]
            output = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                                      stderr=subprocess.PIPE)
            (stdout, stderr) = output.communicate()
        finally:
            os.chdir(cwd)

        logging.debug('Check that vmlinuz could be extracted')
        self.assertEqual(stderr, '')
        self.assertIn("created 1 directories", stdout)

    @unittest.skipUnless(uefi_image,
                         "Test EFI files on ubuntu amd64 images")
    def test_efi_files_present(self):
        """Test if the relevant files are present in the ISO for UEFI."""
        (stdout, stderr) = self.iso.listfiles()
        logging.debug('Check for error in extracting file list from the iso')
        self.assertEqual(stderr, '')
        files_list = open(os.path.join(DATA_PATH, 'file_list_uefi'))
        logging.debug('Check if relevant files are present in the iso')
        for list_server in files_list:
            path = list_server.rstrip()
            self.assertIn(path, stdout)

    @unittest.skipUnless(uefi_image,
                         "Test Secure Boot UEFI signatures")
    def test_efi_secure_boot_signatures(self):
        logging.debug('Generating verification certificates')
        repo_dir = os.path.expanduser('~/repo_dir')
        keys_repo_path = 'notes_testing/secure-boot/keys'
        keys = os.path.join(repo_dir, keys_repo_path)
        if not os.path.isdir(repo_dir):
            os.makedirs(repo_dir)
        git_keys = GitHandler(
            repo='https://git.launchpad.net/qa-regression-testing',
            destination=repo_dir)
        # We check if the keys exists to avoid the sparse checkout
        if os.path.isdir(keys) and os.listdir(keys):
            # It means that the keys are already checked out,
            # so we just pull to ensure it is up to date
            git_keys.pull()
        else:
            git_keys.sparse_checkout(paths=[keys_repo_path])
        cmd = ["make", "-C", keys]
        output = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                                  stderr=subprocess.PIPE)
        (stdout, stderr) = output.communicate()
        self.assertEqual(stderr, '')
        microsoft = os.path.join(keys, 'microsoft-uefica-public.pem')
        canonical = os.path.join(keys, 'canonical-master-public.pem')
        canonical_chain = os.path.join(
            keys, 'canonical-master-signing-public-chain.pem')

        logging.debug('Extracting UEFI boot and kernel images')
        shim = self.iso.extract('./EFI/BOOT/BOOTx64.EFI', self.temp_dir)
        grub = self.iso.extract('./EFI/BOOT/grubx64.efi', self.temp_dir)
        kernel = self.iso.extract(iso.kernelpath(), self.temp_dir)

        logging.debug('Verifying UEFI shim')
        cmd = ['sbverify', '--cert', microsoft, shim]
        output = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                                  stderr=subprocess.PIPE)
        (stdout, stderr) = output.communicate()
        self.assertEqual(stdout, 'Signature verification OK\n')

        logging.debug('Verifying UEFI grub')
        cmd = ['sbverify', '--cert', canonical, grub]
        output = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                                  stderr=subprocess.PIPE)
        (stdout, stderr) = output.communicate()
        self.assertEqual(stdout, 'Signature verification OK\n')

        logging.debug('Detaching kernel signature')
        kernel_sig = os.path.join(self.temp_dir, 'vmlinuz.efi.signature')
        cmd = ['sbattach', '--detach', kernel_sig, kernel]
        output = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                                  stderr=subprocess.PIPE)
        (stdout, stderr) = output.communicate()
        self.assertEqual(stdout, '')
        self.assertEqual(stderr, '')

        logging.debug('Verifying kernel signature')
        cmd = ['sbverify', '--cert', canonical_chain,
               '--detached', kernel_sig, kernel]
        output = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                                  stderr=subprocess.PIPE)
        (stdout, stderr) = output.communicate()
        self.assertEqual(stdout, 'Signature verification OK\n')

    # clean up
    @classmethod
    def tearDownClass(cls):
        try:
            shutil.rmtree(cls.temp_dir, ignore_errors=False)
        except OSError:
            logging.error("Removing the temp directory failed")
            unittest.fail("Cleanup error")


def get_report_data(runlist, results):
    errors = len(results.errors)
    failures = len(results.failures)
    skipped = len(results.skipped)
    passes = results.testsRun - skipped - errors - failures
    report_timestamp = time.strftime('%Y-%m-%d %H:%M:%S', time.gmtime())

    # Note: Ideally 'name' is the runlist name and 'runlist' the path to the
    # runlist file. Given that there isn't a real runlist file here, the name
    # is being used for both fields.
    report_data = {
        'name': runlist,
        'runlist': runlist,
        'media-info': iso.media_info,
        'build_number': iso.buildnumber,
        'release': iso.series,
        'ran_at': report_timestamp,
        'install_type': iso.installtype,
        'arch': iso.arch,
        'commands': [],
        'fetch_errors': 0,
        'errors': errors,
        'failures': failures,
        'passes': passes,
        'uname': [],
    }

    return report_data


class ISOResult(unittest.TestResult):
    def _print_error(self, test, error):
        logging.error(test)
        logging.error(''.join(format_exception(*error)))

    def addError(self, test, error):
        super(ISOResult, self).addError(test, error)
        self._print_error(test, error)

    def addFailure(self, test, error):
        super(ISOResult, self).addError(test, error)
        self._print_error(test, error)

    def addSkip(self, test, reason):
        logging.info('\n{separator}\n{id}\n{separator}'
                     .format(separator='-' * 80, id=test.id()))
        logging.info('Skipped: {}'.format(reason))


def main():
    """Run test cases and write qa dashboard friendly report."""
    suite = unittest.TestLoader().loadTestsFromTestCase(TestValidateISO)
    results = ISOResult()
    results.buffer = True
    suite.run(results)

    fake_runlist = 'iso_static_validation'
    report_data = get_report_data(fake_runlist, results)
    report = yaml.dump(report_data,
                       explicit_start='---',
                       default_flow_style=False)
    filename_timestamp = time.strftime('%Y-%m-%d_%H-%M-%S', time.gmtime())
    log_filename = ('{machine}_{runlist}_{timestamp}.{suffix}'
                    .format(machine=platform.node(),
                            runlist=fake_runlist,
                            timestamp=filename_timestamp,
                            suffix='yaml'))
    absolute_log_filename = os.path.join(DEFAULT_LOG_DIR, log_filename)
    with open(absolute_log_filename, 'w') as f:
        f.write(report)
    print('Test log copied to\n\t{}'.format(absolute_log_filename))

    return 0 if results.wasSuccessful() else 1

if __name__ == '__main__':
    sys.exit(main())
