# Ubuntu Testing Automation Harness
# Copyright 2012 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Provide functions for a machine to interface with a power management system.
Currently only the Sentry CDU in use in the Canonical labs is supported.
"""

import time

from utah.config import config
from utah.process import ProcessRunner
from utah.provisioning.baremetal.exceptions import UTAHBMProvisioningException


class PowerMixin(object):

    """Provide power cycle commands for physical machines.

    Only the Sentry CDU is well supported, but arbitrary commands can work.

    """

    def __init__(self, machineinfo=config.machineinfo,
                 powercmd=config.powercmd, *args, **kw):
        """Store power control info for later use."""
        if machineinfo is not None:
            self.power = {}
            for item in machineinfo:
                if 'power' in item:
                    self.power[item] = machineinfo[item]
        self.powercmd = powercmd
        super(PowerMixin, self).__init__(*args, **kw)

    def powercommand(self):
        """Return the command used to control power for this machine."""
        try:
            cmd = self.powercmd
        except AttributeError:
            try:
                if self.power['power-type'] == 'sentryswitch_cdu':
                    cmd = ['/usr/sbin/fence_cdu',
                           '-a', self.power['power-address'],
                           '-n', self.power['power-id'],
                           '-l', self.power['power-user'],
                           '-p', self.power['power-pass'],
                           '-o']
                else:
                    raise UTAHBMProvisioningException(
                        'Power type {power_type!r} not supported'
                        .format(power_type=self.power['power-type']))
            except AttributeError:
                raise UTAHBMProvisioningException('Power commands not defined'
                                                  ' for this machine')
        return cmd

    def _start(self):
        """Use the machine's powercommand to start the machine."""
        self.logger.debug('Starting system')
        # TODO: check return code here before setting active
        ProcessRunner(self.powercommand() + ['on'])
        self.active = True

    def stop(self):
        """Use the machine's powercommand to stop the machine."""
        self.logger.debug('Stopping system')
        # TODO: check return code here before setting active
        ProcessRunner(self.powercommand() + ['off'])
        self.active = False

    def restart(self):
        """Stop the machine, wait powertimeout, then start it."""
        self.logger.info('Restarting system')
        self.stop()
        self.logger.debug('Waiting %s seconds', str(config.powertimeout))
        time.sleep(config.powertimeout)
        self._start()
