# Ubuntu Testing Automation Harness
# Copyright 2012 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Data structure to help simplify provisioning logic."""

import os
import shutil


class ProvisionData(object):

    """A structure to hold information about image customizations needed.

    This can be used to inform our provisioning logic about extra files and
    auto-start commands it should set up for the target

    """

    def __init__(self):
        self._files = []
        self._cmds = []

    def add_files(self, files, target_dir):
        """Add a file to get copied to the target system."""
        if isinstance(files, basestring):
            files = [files]
        self._files.append({'files': files, 'target_dir': target_dir})

    def add_autostart_cmd(self, name, cmd):
        """Add a command to be run at startup."""
        self._cmds.append({'name': name, 'cmd': cmd})

    def _update_initrd_files(self, path):
        copy_dir = os.path.join(path, 'utah-copy-files')
        os.mkdir(copy_dir)

        with open(os.path.join(path, 'utah-copy-files.sh'), 'w') as f:
            f.write('#!/bin/sh\n')
            for fileobj in self._files:
                tgt_dst = '/target{}'.format(fileobj['target_dir'])
                f.write('[ -d {d} ] || mkdir -p {d}\n'.format(d=tgt_dst))
                for path in fileobj['files']:
                    fname = os.path.basename(path)
                    hst_cpy = os.path.join(copy_dir, fname)
                    tgt_cpy = os.path.join('/utah-copy-files', fname)
                    shutil.copyfile(path, hst_cpy)
                    f.write('cp {} {}/\n'.format(tgt_cpy, tgt_dst))

    def _update_initrd_cmds(self, path):
        path = os.path.join(path, 'utah-autorun')
        os.mkdir(path)

        for cmd in self._cmds:
            with open(os.path.join(path, cmd['name']), 'w') as f:
                f.write('#!/bin/sh\n')
                f.write('{}\n'.format(cmd['cmd']))

    def update_initrd(self, path):
        """Build the initrd file to include needed files commands.

        This copies the files to the initrd, and builds a script that can be
        run to put them in the proper directory on the installed system. It
        also adds a set of auto run files that can be copied over to the
        installed system during the late command.

        """
        self._update_initrd_files(path)
        self._update_initrd_cmds(path)
