
# Support for testing "live-server" (aka subiquity) images.  Missing
# functionality: extracting syslog, installer, curtin logs from the
# image file. Not sure how to hook that in.

import os
import re
import select
import string
import shutil
import tempfile
import time

import yaml

import libvirt

from xml.etree import ElementTree

from utah.config import config

from utah.process import ProcessRunner

from utah.timeout import UTAHTimeout

from utah.provisioning.exceptions import UTAHProvisioningException
from utah.provisioning.ssh import SSHMixin
from utah.provisioning.vm import (
    LibvirtVM,
    random_mac_address,
    UTAHVMProvisioningException,
    )

import utah.timeout
from utah.retry import retry


class LiveServerVM(SSHMixin, LibvirtVM):
    def __init__(self, machineid, diskbus, disksizes, emulator, name, **kw):
        self.diskbus = diskbus
        self.disksizes = disksizes
        self.emulator = emulator
        self.macs = config.macs
        self.prefix = config.prefix
        self.disks = []
        if name is None:
            name = '-'.join([str(config.prefix), str(machineid)])
            autoname = True
        super(LiveServerVM, self).__init__(
            machineid=machineid, name=name, **kw)
        if autoname:
            self._namesetup()
            self._loggerunsetup()
            self._loggersetup()
        if self.inventory is not None:
            self.cleanfunction(self.inventory.destroy,
                               machineid=self.machineid)
        self.directory = os.path.join(config.vmpath, self.name)
        self.cleanfile(self.directory)
        self.logger.info('Checking if machine directory {} exists'
                         .format(self.directory))
        if not os.path.isdir(self.directory):
            self.logger.debug('Creating {}'.format(self.directory))
            try:
                os.makedirs(self.directory)
            except OSError as err:
                raise UTAHVMProvisioningException(err)
        self.logger.debug('LiveServerVM init finished')

    def _tmpimage(self, image=None, tmpdir=None):
        """Create a temporary copy of the image so libvirt will lock that one.

        This allows other simultaneous processes to update the cached image.

        :returns: The path to the temporary image
        :rtype: str

        """
        if image is None:
            image = self.image.image
        if tmpdir is None:
            tmpdir = self.tmpdir
        self.logger.info('Making temp copy of install image')
        tmpimage = os.path.join(tmpdir, os.path.basename(image))
        self.logger.debug('Copying %s to %s', image, tmpimage)
        shutil.copyfile(image, tmpimage)
        return tmpimage

    def _setuppreseed(self, tmpdir):
        # Read the provided "answers file" for subiquity, overwrite
        # the Identity section so that we can log in how SSHMixin
        # expects to be able to and prepare a disk image in the form
        # subiquity expects that contains it.
        self.answers_image_path = os.path.join(tmpdir, "answers.img")
        mount_dir = os.path.join(tmpdir, "answers")
        os.makedirs(mount_dir)
        cmds = [
            ['truncate', '-s', '1M', self.answers_image_path],
            ['mkfs.ext2', '-U', '00c629d6-06ab-4dfd-b21e-c3186f34105d', self.answers_image_path],
            ['sudo', 'mount', self.answers_image_path, mount_dir],
            ['sudo', 'chown', '{}:{}'.format(os.getuid(), os.getgid()), mount_dir],
            ]
        for cmd in cmds:
            if ProcessRunner(cmd).returncode != 0:
                raise UTAHVMProvisioningException(
                    '{} failed'.format(cmd))
        yaml_file = os.path.join(mount_dir, "answers.yaml")
        answers = yaml.safe_load(open(self.preseed))
        answers['Identity'] = {
            # This is "ubuntu", crypted:
            'password': '$6$wdAcoXrU039hKYPd$508Qvbe7ObUnxoj15DRCkzC3qO7edjH0VV7BPNRDYK4QR8ofJaEEF2heacn0QgD.f8pO8SNp83XNdWG6tocBM1',
            'hostname': self.name,
            'realname': 'Utah',
            'ssh-import-id': 'lp:' + config.launchpad_user.encode('utf-8'),
            'username': config.user,
            }

        with open(yaml_file, 'w') as f:
            f.write(yaml.dump(answers))
        cmd = ['sudo', 'umount', mount_dir]
        if ProcessRunner(cmd).returncode != 0:
            raise UTAHVMProvisioningException(
                '{} failed'.format(cmd))
        disk = {'bus': self.diskbus,
                'file': self.answers_image_path,
                'size': 1024*1024,
                'type': 'raw'}
        self.disks.append(disk)

    def _createdisks(self, disksizes=None):
        """Create disk files if needed and build a list of them."""
        self.logger.info('Creating disks')
        if disksizes is None:
            disksizes = self.disksizes
        for index, size in enumerate(disksizes):
            disksize = '{}G'.format(size)
            basename = 'disk{}.qcow2'.format(index)
            diskfile = os.path.join(self.directory, basename)
            if not os.path.isfile(diskfile):
                cmd = ['qemu-img', 'create', '-f', 'qcow2', diskfile,
                       disksize]
                self.logger.debug('Creating %s disk using: %s',
                                  disksize, ' '.join(cmd))
                if ProcessRunner(cmd).returncode != 0:
                    raise UTAHVMProvisioningException(
                        'Could not create disk image at {}'.format(diskfile))
            disk = {'bus': self.diskbus,
                    'file': diskfile,
                    'size': disksize,
                    'type': 'qcow2'}
            self.disks.append(disk)
            self.cleanfile(diskfile)
            self.logger.debug('Adding disk to list')

    def _installxml_rewrite_boot(self, xmlt):
        # Set the arch appropriately, remove any configuration of
        # kernel/initrd/commandline and configure the VM to boot from
        # CD.
        xmlt.getroot().set('type', 'kvm')
        ose = xmlt.find('os')
        if self.image.arch == 'i386':
            ose.find('type').set('arch', 'i686')
        elif self.image.arch == 'amd64':
            ose.find('type').set('arch', 'x86_64')
        elif self.image.arch == 'ppc64el':
            ose.find('type').set('arch', 'ppc64le')
            ose.find('type').set('machine', 'pseries')
        else:
            ose.find('type').set('arch', self.image.arch)
        self.logger.debug('Setting up boot info')
        for kernele in list(ose.iterfind('kernel')):
            ose.remove(kernele)
        for initrde in list(ose.iterfind('initrd')):
            ose.remove(initrde)
        for cmdlinee in list(ose.iterfind('cmdline')):
            ose.remove(cmdlinee)
        ose.find('boot').set('dev', 'cdrom')

    def _installxml_rewrite_disks(self, devices, image):
        # 1. Remove all existing disks
        # 2. Hook image up to cdrom
        # 3. Create all required disks.
        self.logger.debug('Setting up disks')
        for disk in list(devices.iterfind('disk')):
            if disk.get('device') == 'disk':
                devices.remove(disk)
                self.logger.debug('Removed existing disk')
            # TODO: Add a cdrom if none exists
            if disk.get('device') == 'cdrom':
                if self.image.arch == 'ppc64el':
                    disk.find('target').set('bus', 'virtio')
                else:
                    disk.find('target').set('bus', 'ide')
                for address in list(disk.iterfind('address')):
                    disk.remove(address)
                if disk.find('source') is not None:
                    disk.find('source').set('file', image)
                    self.logger.debug('Rewrote existing CD-ROM')
                else:
                    source = ElementTree.Element('source')
                    source.set('file', image)
                    disk.append(source)
                    self.logger.debug(
                        'Added source to existing CD-ROM')
        for disk in self.disks:
            diske = ElementTree.Element('disk')
            diske.set('type', 'file')
            diske.set('device', 'disk')
            driver = ElementTree.Element('driver')
            driver.set('name', 'qemu')
            driver.set('type', disk['type'])
            diske.append(driver)
            serial = ElementTree.Element('serial')
            serial.text = "disk-{}".format(self.disks.index(disk))
            diske.append(serial)
            source = ElementTree.Element('source')
            source.set('file', disk['file'])
            diske.append(source)
            target = ElementTree.Element('target')
            dev = ("vd{}"
                   .format(string.ascii_lowercase[self.disks.index(disk)]))
            target.set('dev', dev)
            target.set('bus', disk['bus'])
            diske.append(target)
            devices.append(diske)
            self.logger.debug('Added %s disk', str(disk['size']))

    def _installxml_rewrite_nics(self, devices):
        macs = list(self.macs)
        for interface in devices.iterfind('interface'):
            if interface.get('type') in ['network', 'bridge']:
                if len(macs) > 0:
                    mac = macs.pop(0)
                    self.logger.debug('Using specified mac address: %s', mac)
                else:
                    mac = random_mac_address()
                    self.logger.debug('Using random mac address: %s', mac)
                    self.macs.append(mac)
                interface.find('mac').set('address', mac)
            if interface.get('type') == 'bridge':
                interface.find('source').set('bridge', config.bridge)

    def _installxml_add_serial(self, devices):
        serial = ElementTree.Element('serial')
        serial.set('type', 'file')
        source = ElementTree.Element('source')
        source.set('path', self.syslog)
        #TODO: reintegrate this when the required qemu version lands
#        source.set('append', 'on')  # lp#1548499
        serial.append(source)
        target = ElementTree.Element('target')
        target.set('port', '0')
        serial.append(target)
        devices.append(serial)

    def _installxml(self, image, tmpdir, xml):
        """Return the XML tree to be passed to libvirt for VM installation."""
        self.logger.info('Creating installation XML')
        xmlt = ElementTree.ElementTree(file=xml)
        xmlt.find('on_reboot').text = 'destroy'
        self.logger.debug('Rewriting basic info')
        xmlt.find('name').text = self.name
        xmlt.find('uuid').text = self.uuid
        self._installxml_rewrite_boot(xmlt)
        devices = xmlt.find('devices')
        self._installxml_rewrite_disks(devices, image)
        self._installxml_rewrite_nics(devices)
        self._installxml_add_serial(devices)
        xmlt.write(os.path.join(tmpdir, 'install.xml'))
        self.logger.info('Installation XML ready')
        return xmlt

    def _installvm(self, lv=None, tmpdir=None, xml=None):
        """Install a VM, then undefine it in libvirt.

        The final installation will recreate the VM using the existing disks.

        """
        self.logger.info('Creating VM')
        if lv is None:
            lv = self.lv
        if xml is None:
            xml = self.xml
        self.logger.info(ElementTree.tostring(xml.getroot()))
        if tmpdir is None:
            tmpdir = self.tmpdir
        os.chmod(tmpdir, 0o755)
        self.logger.info(
            'Installing system on VM (may take over an hour)')
        self.logger.info('You can watch the progress with virt-viewer')
        log_filename = os.path.join(config.logpath,
                                    '{}.syslog.log'.format(self.name))
        self.logger.info('Logs will be written to %s', log_filename)

        before = time.time()
        try:
            vm = lv.defineXML(ElementTree.tostring(xml.getroot()))
            vm.create()
            while vm.isActive() is not 0:
                time.sleep(1)
        finally: # This stops us from leaking VMs on exceptions
            try:
                vm.destroy()
            except libvirt.libvirtError:
                pass
            finally:
                vm.undefine()

        after = time.time()
        self.logger.info('Installation complete in %s seconds' %
                         int(after-before))

    def _finalxml(self, tmpdir, xml):
        """Create the XML to be used for the post-installation VM.

        This "ejects" the CD and removes the answers image file.
        """
        self.logger.info('Creating final VM XML')
        self.logger.debug('Setting VM to reboot normally on reboot')
        xml.find('on_reboot').text = 'restart'
        self.logger.debug('Removing VM install parameters')
        devices = xml.find('devices')
        for disk in list(devices.iterfind('disk')):
            if disk.get('device') == 'cdrom':
                disk.remove(disk.find('source'))
            source = disk.find('source')
            if source is not None and source.get('file') == self.answers_image_path:
                devices.remove(disk)
        xml.find('os').find('boot').set('dev', 'hd')
        xml.write(os.path.join(tmpdir, 'final.xml'))
        return xml

    def _create(self, provision_data):
        """Create the VM, install the system, and prepare it to boot.
        """
        self.logger.info('Creating custom virtual machine')

        tmpdir = tempfile.mkdtemp(prefix='{}/tmp'.format(self.directory))
        self.cleanfile(tmpdir)
        self.logger.debug('Working dir: %s', tmpdir)
        os.chdir(tmpdir)

        if provision_data:
            1/0

        self._setuppreseed(tmpdir=tmpdir)

        #self._setuplogging(tmpdir=tmpdir)

        self._createdisks()

        image = self._tmpimage(image=self.image.image, tmpdir=tmpdir)

        xml = self._installxml(image=image, tmpdir=tmpdir, xml=self.xml)

        self._installvm(lv=self.lv, tmpdir=tmpdir, xml=xml)

        xml = self._finalxml(tmpdir=tmpdir, xml=xml)

        self.logger.info('Setting up final VM')
        self.logger.info(ElementTree.tostring(xml.getroot()))
        self.vm = self.lv.defineXML(ElementTree.tostring(xml.getroot()))
        self.cleanfunction(self.vm.destroy)
        self.cleanfunction(self.vm.undefine)
        if self.poweroff:
            self.cleanfunction(self.vm.shutdown, force=True)

    def _start(self):
        """Start the VM."""
        self.logger.info('Starting LiveServerVM')
        if self.vm is not None:
            if self.vm.isActive() == 0:
                self.vm.create()
        else:
            raise UTAHVMProvisioningException('Failed to provision VM')

        def check():
            # Not sure if it's appropriate to assume default network here...
            network = self.lv.networkLookupByName('default')
            for d in network.DHCPLeases():
                for m in self.macs:
                    if m == d['mac']:
                        self._ssh_address = d['ipaddr']
                        self.logger.debug("found IP address %s for mac %s", self._ssh_address, m)
                        return
            self.logger.debug("No DHCP lease seen yet")
            raise UTAHVMProvisioningException(retry=True)
        utah.timeout.timeout(300, retry, check,
                             logmethod=self.logger.debug, retry_timeout=10)
        try:
            self.logger.info(
                'Waiting %d seconds for ping response', config.boottimeout)
            self.pingpoll(timeout=config.boottimeout)
        except UTAHTimeout:
            # Ignore timeout for ping, since depending on the network
            # configuration ssh might still work despite of the ping failure.
            self.logger.warning('Network connectivity (ping) failure')
        self.sshpoll(timeout=config.boottimeout)
        self._enable_root_login()
        self.active = True

    def pingcheck(self):
        """Check network connectivity using ping.

        :raises UTAHProvisioningException:
            When ping fails.

        .. seealso:: :func:`utah.retry.retry`, :meth:`pingpoll`

        """
        self.logger.info('Checking network connectivity (ping)')
        returncode = \
            ProcessRunner(['ping', '-c1', '-w5', self.ssh_address]).returncode
        if returncode != 0:
            err = 'Ping returned {0}'.format(returncode)
            raise UTAHProvisioningException(err, retry=True)

    @property
    def ssh_address(self):
        return self._ssh_address

    def _enable_root_login(self):
        try:
            self.ssh_client.connect(self.ssh_address,
                                    username=config.user,
                                    key_filename=config.sshprivatekey)
            chan = self.ssh_client.get_transport().open_session()
            chan.get_pty()
            chan.set_combine_stderr(True)
            chan.exec_command("sudo su -c 'cp -aT .ssh /root/.ssh && chown root:root -R /root/.ssh && echo ok'")
            def expect(pat):
                prog = re.compile(pat)
                buf = ''
                while True:
                    match = prog.match(buf)
                    if match:
                        return match
                    r, _, _ = select.select([chan], [], [], 10)
                    if not r:
                        raise Exception("no output for 10s after %r", buf)
                    c = chan.recv(1)
                    buf += c
            expect('\[sudo\] password for ([a-z]+): ')
            chan.send('ubuntu\n')
            expect('\r\n')
            m = expect("Sorry|ok")
            if m.string != "ok":
                raise UTAHVMProvisioningException("_enable_root_login failed")
        finally:
            self.ssh_logger.debug('Closing SSH connection')
            self.ssh_client.close()
