# Ubuntu Testing Automation Harness
# Copyright 2017 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import subprocess
import time

AUTOPKGTEST_NAME = 'autopkgtest.service'
AUTOPKGTEST_SERVICE = os.path.join('/lib/systemd/system', AUTOPKGTEST_NAME)
AUTOPKGTEST_CFG = '/etc/default/grub.d/90-autopkgtest.cfg'

""" This script is run on first boot of the qemu after installation, by the
    qemu-setup.service. It will prepare the system by:
    - Creating a serial console on ttyS1 that will be used in future by
      autopkgtest. The service will be started automatically on all subsequent
      boots by autopkgtest.service.
    - Disable qemu-setup.service to stop this script from running on subsequent
      boots.
    - Power down the system to complete the setup process.
"""


def wait_for_network():
    """Wait until network connection is active."""
    max_count = 60
    for count in range(max_count):
        status = subprocess.call(['wget', '-q', '--spider', 'launchpad.net'])
        if status == 0:
            break
        elif count >= max_count - 1:
            raise Exception('No network connection available')
        else:
            print('wget attempt {} failed, trying again...'.format(count + 1))
            time.sleep(1)


def wait_for_locks():
    """Wait for file locks to be released for doing install operations."""
    max_count = 60
    lock_files = [
        '/var/lib/dpkg/lock',
        '/var/cache/apt/archives/lock',
        '/var/lib/apt/lists/lock',
    ]
    for count in range(max_count):
        status = subprocess.call(['fuser'] + lock_files)
        if status == 1:
            break
        elif count >= max_count - 1:
            raise RuntimeError('File locks not released')
        else:
            print('file locks active, trying again...'.format(count + 1))
            time.sleep(1)


def get_architecture():
    """Return architecture of running system."""
    return subprocess.check_output(
        ['dpkg', '--print-architecture']).decode().strip()


def install_packages(packages):
    """Install specfied list of packages."""
    wait_for_locks()
    subprocess.check_call(['dpkg', '--configure', '-a'])
    subprocess.check_call(['apt-get', 'install', '-y'] + packages)


def create_file(path, content):
    """Create a file with specified path and content."""
    with open(path, 'w') as f:
        f.write(content)


def setup_root_shell_service():
    """Setup root shell service on ttyS1 for use by autopkgtest."""
    create_file(
        AUTOPKGTEST_SERVICE,
        '[Unit]\n'
        'Description=autopkgtest root shell on ttyS1\n'
        'ConditionPathExists=/dev/ttyS1\n\n'
        '[Service]\n'
        'ExecStart=/bin/sh\n'
        'StandardInput=tty-fail\n'
        'StandardOutput=tty\n'
        'StandardError=tty\n'
        'TTYPath=/dev/ttyS1\n'
        'SendSIGHUP=yes\n'
        'SuccessExitStatus=0 208 SIGHUP SIGINT SIGTERM SIGPIPE\n\n'
        '[Install]\n'
        'WantedBy=multi-user.target\n'
    )
    os.chmod(AUTOPKGTEST_SERVICE, 0o644)
    subprocess.check_call(['systemctl', 'enable', AUTOPKGTEST_NAME])


def _grubd_autopkgtest_setup(architecture):
    """Add autopkgtest config to grub.d directory.
    :param architecture: Required architecture.
    :return: True if changes made, False otherwise.
    """
    if architecture == 'i386':
        create_file(
            AUTOPKGTEST_CFG,
            'GRUB_CMDLINE_LINUX_DEFAULT="console=ttyS0 vmalloc=512M"'
        )
        return True
    elif architecture == 'amd64':
        create_file(
            AUTOPKGTEST_CFG,
            'GRUB_CMDLINE_LINUX_DEFAULT="console=ttyS0"'
        )
        return True
    return False


def _grub_autopkgtest_setup(architecture):
    """Add autopkgtest config to grub directory.
    :param architecture: Required architecture.
    :return: True if changes made, False otherwise.
    """
    changed = False
    if architecture == 'i386':
        subprocess.check_call(
            ['sed', '-i',
             '/CMDLINE_LINUX_DEFAULT/ s/"$/ console=ttyS0 '
             'vmalloc=512M"/', '/etc/default/grub']
        )
        changed = True
    elif architecture == 'amd64':
        subprocess.check_call(
            ['sed', '-i',
             '/CMDLINE_LINUX_DEFAULT/ s/"$/ console=ttyS0"/',
             '/etc/default/grub']
        )
        changed = True

    zero_timeout = subprocess.check_output(
        ['grep', '-q', 'GRUB_HIDDEN_TIMEOUT=0', '/etc/default/grub']
    ) == 0

    if zero_timeout:
        subprocess.check_output(
            ['sed', '-i', '/^GRUB_TIMEOUT=/ s/=.*$/=1/',
             '/etc/default/grub']
        )
        changed = True

    return changed


def setup_serial_console():
    """Setup a serial console for autopkgtest to use."""
    # This method is derived from:
    # /usr/share/autopkgtest/setup-commands/setup-testbed
    setup_root_shell_service()
    grub = subprocess.call(['which', 'update-grub']) == 0
    if not os.path.isfile(AUTOPKGTEST_CFG) and grub:
        arch = get_architecture()
        if os.path.isdir('/etc/default/grub.d'):
            changed = _grubd_autopkgtest_setup(arch)
        else:
            changed = _grub_autopkgtest_setup(arch)
        if changed:
            subprocess.check_call(['update-grub'])


def disable_qemu_setup_on_boot():
    """Disable this setup script from running again on future boot up."""
    subprocess.call(['systemctl', 'disable', 'qemu-setup.service'])


def system_shutdown():
    """Shutdown the system to complete first boot setup."""
    subprocess.call(['shutdown', '--poweroff', '0'])


def do_setup():
    """Complete first boot setup actions."""
    wait_for_network()
    setup_serial_console()
    install_packages(['openssh-server'])


def do_teardown():
    """Cleanup actions to disable this from running after first boot."""
    disable_qemu_setup_on_boot()
    system_shutdown()


def main():
    try:
        do_setup()
    finally:
        do_teardown()


if __name__ == '__main__':
    main()
