# Ubuntu Testing Automation Harness
# Copyright 2013 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Provides helpers for dealing with Ubuntu Touch devices."""

import atexit
import contextlib
import sys
import subprocess
import tempfile
import time


import utah.process as process

from utah.exceptions import UTAHException


class ADBShellError(UTAHException):

    """An error occurred running an adb shell command."""

    def __init__(self, cmd, stdout, retry=True):
        Exception.__init__(self, 'ERROR: running {}'.format(cmd))
        self.cmd = cmd
        self.stdout = stdout
        self.retry = retry
        self.adb_error = 'ADB_RC=' not in stdout

    def long_message(self):
        """return a long descriptive message about the error."""
        return '{} stdout was:\n{}'.format(self.message, self.stdout)


class Device(object):

    """Provides a simple API for dealing with a touch device."""

    def __init__(self, logger_func, serial=None):
        self._log = logger_func
        self._serial = serial

    def _adb_operation(self, opargs, exit_on_fail=True):
        """Wrapper for things like adb wait-for-device etc."""
        cmd = ['adb']
        if self._serial:
            cmd.extend(['-s', self._serial])
        cmd.extend(opargs)
        try:
            subprocess.check_call(cmd)
        except subprocess.CalledProcessError as e:
            self._log('Error executing: {}'.format(e.cmd))
            if exit_on_fail:
                sys.exit(1)

    def _adb_shell(self, cmd, stream=True):
        """Run a file under adb-shell with hack to determine its exit code.

        :return: stdout of the command
        :rtype: string

        """
        if self._serial:
            adb = 'adb -s {} shell'.format(self._serial)
        else:
            adb = 'adb shell'

        cmd = '{} {} \\; echo ADB_RC=\\$?'.format(adb, cmd)
        res = process.run(cmd, None, 3600, stream)
        stdout = res[2]
        if 'ADB_RC=0' not in stdout:
            raise ADBShellError(cmd, stdout)
        if stdout.endswith('ADB_RC=0\r\n'):
            # strip it off:
            stdout = stdout[:-10]
        return stdout

    def adb_root(self):
        """Ensure ADB daemon is running as root."""
        self._adb_operation(['root'])

    def reboot_and_wait(self):
        """Reboot the device and wait for it to become active."""
        # kill adb to make sure wait-for-device doesn't return before the
        # system has rebooted
        try:
            self._adb_operation(['reboot'])
            self.run_ubuntu('stop android-tools-adbd', False)
        except ADBShellError as e:
            # the command will fail because ADB will die
            if not e.adb_error:
                raise e
        self.wait_for_device()

    def wait_for_device(self):
        """Wait for the device to be available."""
        self._adb_operation(['wait-for-device'])

    def wait_for_network(self):
        """Wait for networking to be up and running on the target"""
        args = ['phablet-network', '--skip-setup']
        if self._serial:
            args.extend(['-s', self._serial])
        try:
            subprocess.check_call(args)
        except subprocess.CalledProcessError:
            self._log('wait for network failed, rebooting and retrying...')
            self.reboot_and_wait()
            subprocess.check_call(args)

    def run_ubuntu(self, cmd, stream=True):
        """Run a command from the Ubuntu environment.

        :return: stdout of the command
        :rtype: string

        """
        return self._adb_shell(cmd, stream)

    def detect_type(self):
        """determine the device type.

        :return: device type like "mako"
        :rtype: string

        """
        device = self._adb_shell('getprop ro.cm.device', stream=False).strip()
        # Check Android property
        if not device:
            device = self._adb_shell(
                'getprop ro.product.device', stream=False).strip()
        return device

    def create_file_ubuntu(self, pathname, contents, mode=None):
        """Create a file in the ubuntu file system."""
        with tempfile.NamedTemporaryFile() as f:
            f.write(contents)
            f.flush()
            self._adb_operation(['push', f.name, pathname])
        if mode:
            self._adb_operation(['shell', 'chmod', mode, pathname])

    def get_file_ubuntu(self, tgt_file, dst_dir, exit_on_fail=True):
        """Copy a file from the ubuntu file system to host directory."""
        self._adb_operation(['pull', tgt_file, dst_dir], exit_on_fail)

    @contextlib.contextmanager
    def tail_syslog(self):
        """Tail the syslog while in the context."""
        cmd = ['adb']
        if self._serial:
            cmd.extend(['-s', self._serial])
        cmd.extend(['shell', 'tail', '-f', '/var/log/syslog'])
        p = subprocess.Popen(cmd)
        atexit.register(p.terminate)  # incase _chrooted ends in an error
        try:
            yield
        finally:
            # there's a window where initctl will have ended but our tailing of
            # syslog won't have read everything, so sleep a bit to and get it
            time.sleep(2)
            p.terminate()
