#!/usr/bin/python
# UESQA Metrics
# Copyright (C) 2017 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from argparse import ArgumentParser
from junit_xml import TestSuite, TestCase
import os
import yaml

from utah.client.const import (
    ERROR, FAIL, TESTSUITE, TESTCASE, STATUS, STDERR, STDOUT
)

CMD_TYPE = 'cmd_type'
COMMANDS = 'commands'
TESTCASE_CMD = 'testcase_test'


class YAML2XUnitConverter:
    """ Convert a results yaml file to xunit format """

    def __init__(self, yaml_file, output_file):
        self.yaml_file = yaml_file
        self.output_file = output_file

        self.suites = []
        self.all_docs = []

        self._check_files()

    def _check_files(self):
        """ Check the yaml and output files are ok and load the input data """
        if not os.path.exists(self.yaml_file):
            print('File to convert does not exist: {}'.format(self.yaml_file))
        else:
            self._read_yaml()

        if os.path.exists(self.output_file):
            os.remove(self.output_file)

    def _get_suite_name(self, suite_path):
        """ Get the test suite name based on the path """
        return suite_path.split('/')[-1]

    def _is_last_suite(self, name):
        """ Indicate if the last test suite is the passed as parameter """
        return len(self.suites) > 0 and self.suites[-1].name == name

    def _create_test_suite(self, name):
        """ Create a test suite """
        ts = TestSuite(name, [])
        self.suites.append(ts)
        return ts

    def _is_test_command_wellformed(self, command):
        mandatory_fields = [STDOUT, STATUS, STDERR, TESTCASE, TESTSUITE]
        for field in mandatory_fields:
            if field not in command:
                print('Field {} not available in command {}'.format(
                    field, command))
            return False
        return True

    def _create_test_case(self, command):
        """ Create a test case and its suite in case it was not created """
        if self._is_test_command_wellformed(command):
            print('Error found in command, skipping')
            return
        suite_name = self._get_suite_name(command[TESTSUITE])
        if not self._is_last_suite(suite_name):
            ts = self._create_test_suite(suite_name)
        else:
            ts = self.suites[-1]

        tc = TestCase(command[TESTCASE], stdout=command[STDOUT])
        if command[STATUS] == ERROR:
            tc.add_error_info(command[STDERR])
        if command[STATUS] == FAIL:
            tc.add_failure_info(command[STDERR])

        ts.test_cases.append(tc)

    def _read_yaml(self):
        """ Read the yaml content and store it """
        with open(self.yaml_file, "r") as yf:
            docs = yaml.load_all(yf)
            for doc in docs:
                for k, v in doc.items():
                    self.all_docs.append((k, v))

    def _process(self):
        """ Process the data got from yaml and create the test suites """
        for (k, v) in self.all_docs:
            if k == COMMANDS:
                for command in v:
                    if command[CMD_TYPE] == TESTCASE_CMD:
                        self._create_test_case(command)

    def _write_xunit(self):
        """ Write the xunit data to file """
        with open(self.output_file, 'w') as f:
            TestSuite.to_file(f, self.suites, prettyprint=False)

    def convert(self):
        """ Convert from the yaml data to xunit format """
        self._process()
        self._write_xunit()


def check_args():
    """ Validate input args """
    args_parser = ArgumentParser("Parse input parameters")
    args_parser.add_argument('yamlfile',
                             help='Path to the yaml file to convert')
    args_parser.add_argument('outputfile',
                             help='Path to the output file to create')
    return args_parser.parse_args()


if __name__ == '__main__':
    args = check_args()
    converter = YAML2XUnitConverter(args.yamlfile, args.outputfile)
    converter.convert()
